/* Copyright (c) 2005 PrimeBase Technologies GmbH, Germany
 *
 * PrimeBase XT
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * 2005-05-24	Paul McCullagh
 *
 * H&G2JCtL
 */

#include "xt_config.h"
#include <mysql.h>
#include <m_ctype.h>
#include <my_handler.h>

#include <stdio.h>
#include <time.h>

#include "thread_xt.h"
#include "filesys_xt.h"
#include "cache_xt.h"
#include "table_xt.h"
#include "trace_xt.h"
#include "util_xt.h"
#include "index_xt.h"

#define XT_TIME_DIFF(start, now) (\
	((xtWord4) (now) < (xtWord4) (start)) ? \
	((xtWord4) 0XFFFFFFFF - ((xtWord4) (start) - (xtWord4) (now))) : \
	((xtWord4) (now) - (xtWord4) (start)))

/*
 * -----------------------------------------------------------------------
 * D I S K   C A C H E
 */

#define IDX_CAC_SEGMENT_COUNT		((off_t) 1 << XT_INDEX_CACHE_SEGMENT_SHIFTS)
#define IDX_CAC_SEGMENT_MASK		(IDX_CAC_SEGMENT_COUNT - 1)
#define IDX_CAC_MAX_FLUSH_COUNT		(XT_INDEX_CACHE_MAX_FLUSH_SEG_CNT * IDX_CAC_SEGMENT_COUNT)

#define IDX_CAC_BLOCK_FREE			0
#define IDX_CAC_BLOCK_READING		1
#define IDX_CAC_BLOCK_CLEAN			2
#define IDX_CAC_BLOCK_DIRTY			3
#define IDX_CAC_BLOCK_WRITING		4

typedef struct DcBlock {
	off_t				cb_address;						/* The block address. */
	u_int				cb_file_id;						/* The file id of the block. */
	xtWord4				cb_state;						/* Block status. */
	xtWord4				cb_ru_time;						/* If this is in the top 1/4 don't change position in MRU list. */
	struct DcBlock		*cb_next;						/* Pointer to next block on hash list, or next free block on free list. */
	struct DcBlock		*cb_dirty_next;					/* Double link for dirty blocks, next pointer. */
	struct DcBlock		*cb_dirty_prev;					/* Double link for dirty blocks, previous pointer. */
	struct DcBlock		*cb_mr_used;					/* More recently used blocks. */
	struct DcBlock		*cb_lr_used;					/* Less recently used blocks. */
	xtWord1				cb_data[XT_INDEX_PAGE_SIZE];
} DcBlockRec, *DcBlockPtr;

/* A disk cache segment. The cache is divided into a number of segments
 * to improve concurrency.
 */
typedef struct DcSegment {
	xt_mutex_type			cs_lock;						/* The cache segment lock. */
	xt_cond_type		cs_cond;
#ifndef MUTEX_FASTER_THAN_WLOCK
	xt_mutex_type		cs_cond_lock;					/* Required if "fast mutex" is a wrlock! */
#endif
	DcBlockPtr			*cs_hash_table;
	DcBlockPtr			cs_dirty_list[XT_INDEX_CACHE_FILE_SLOTS];
} DcSegmentRec, *DcSegmentPtr;

typedef struct DcGlobals {
	xt_mutex_type		cg_lock;						/* The public cache lock. */
	xt_cond_type		cg_cond;						/* The public cache wait condition. */
	DcSegmentRec		cg_segment[IDX_CAC_SEGMENT_COUNT];
	DcBlockPtr			cg_blocks;
	DcBlockPtr			cg_free_list;
	xtWord4				cg_free_count;
	DcBlockPtr			cg_lru_block;
	DcBlockPtr			cg_mru_block;
	xtWord4				cg_ru_now;					/* A counter as described by Jim Starkey (my thanks) */
	xtWord4				cg_hash_size;
	xtWord4				cg_block_count;
} DcGlobalsRec;

static DcGlobalsRec	ind_cac_globals;

#ifdef XT_USE_MYSYS
#ifdef xtPublic
#undef xtPublic
#endif
#include "my_global.h"
#include "my_sys.h"
#include "keycache.h"
KEY_CACHE my_cache;
#undef	pthread_rwlock_rdlock
#undef	pthread_rwlock_wrlock
#undef	pthread_rwlock_unlock
#undef	pthread_mutex_lock
#undef	pthread_mutex_unlock
#undef	pthread_cond_wait
#undef	pthread_cond_broadcast
#undef	xt_mutex_type
#define xtPublic
#endif

/*
 * Initialize the disk cache.
 */
xtPublic void xt_ind_init(XTThreadPtr self, size_t cache_size)
{
	DcBlockPtr	block;

#ifdef XT_USE_MYSYS
	init_key_cache(&my_cache, 1024, cache_size, 100, 300);
#endif
	/*
	 * This is required to ensure that the block
	 * works!
	 */
	(void) ASSERT(XT_INDEX_PAGE_SIZE == XT_INDEX_PAGE_SIZE);

	/* Determine the number of block that will fit into the given memory: */
	/*
	ind_cac_globals.cg_hash_size = (cache_size / (IDX_CAC_SEGMENT_COUNT * sizeof(DcBlockPtr) + sizeof(DcBlockRec))) / (IDX_CAC_SEGMENT_COUNT >> 1);
	ind_cac_globals.cg_block_count = (cache_size - (IDX_CAC_SEGMENT_COUNT * ind_cac_globals.cg_hash_size * sizeof(DcBlockPtr))) / sizeof(DcBlockRec);
	*/
	/* NEW: I no longer count the size of the directory: */
	ind_cac_globals.cg_block_count = cache_size / sizeof(DcBlockRec);
	ind_cac_globals.cg_hash_size = ind_cac_globals.cg_block_count / (IDX_CAC_SEGMENT_COUNT >> 1);

	try_(a) {
		for (u_int i=0; i<IDX_CAC_SEGMENT_COUNT; i++) {
			ind_cac_globals.cg_segment[i].cs_hash_table = (DcBlockPtr *) xt_calloc(self, ind_cac_globals.cg_hash_size * sizeof(DcBlockPtr));
			xt_init_mutex(self, &ind_cac_globals.cg_segment[i].cs_lock);
			xt_init_cond(self, &ind_cac_globals.cg_segment[i].cs_cond);
#ifndef MUTEX_FASTER_THAN_WLOCK
			xt_init_mutex(self, &ind_cac_globals.cg_segment[i].cs_cond_lock);
#endif
		}

		block = (DcBlockPtr) xt_malloc(self, ind_cac_globals.cg_block_count * sizeof(DcBlockRec));
		ind_cac_globals.cg_blocks = block; 
		xt_init_mutex(self, &ind_cac_globals.cg_lock);
		xt_init_cond(self, &ind_cac_globals.cg_cond);

		for (u_int i=0; i<ind_cac_globals.cg_block_count; i++) {
			block->cb_state = IDX_CAC_BLOCK_FREE;
			block->cb_next = ind_cac_globals.cg_free_list;
			ind_cac_globals.cg_free_list = block;
			block++;
		}
		ind_cac_globals.cg_free_count = ind_cac_globals.cg_block_count;
	}
	catch_(a) {
		xt_ind_exit(self);
		throw_();
	}
	cont_(a);
}

xtPublic void xt_ind_exit(XTThreadPtr self)
{
#ifdef XT_USE_MYSYS
	end_key_cache(&my_cache, 1);
#endif
	for (u_int i=0; i<IDX_CAC_SEGMENT_COUNT; i++) {
		if (ind_cac_globals.cg_segment[i].cs_hash_table) {
			xt_free(self, ind_cac_globals.cg_segment[i].cs_hash_table);
			ind_cac_globals.cg_segment[i].cs_hash_table = NULL;
			xt_free_mutex(&ind_cac_globals.cg_segment[i].cs_lock);
			xt_free_cond(&ind_cac_globals.cg_segment[i].cs_cond);
#ifndef MUTEX_FASTER_THAN_WLOCK
			xt_free_mutex(&ind_cac_globals.cg_segment[i].cs_cond_lock);
#endif
		}
	}

	if (ind_cac_globals.cg_blocks) {
		xt_free(self, ind_cac_globals.cg_blocks);
		ind_cac_globals.cg_blocks = NULL;
		xt_free_mutex(&ind_cac_globals.cg_lock);
		xt_free_cond(&ind_cac_globals.cg_cond);
	}
	memset(&ind_cac_globals, 0, sizeof(ind_cac_globals));
}

xtPublic xtInt8 xt_ind_get_usage()
{
	xtInt8 size = 0;

	size = (xtInt8) (ind_cac_globals.cg_block_count - ind_cac_globals.cg_free_count) * (xtInt8) sizeof(DcBlockRec);
	return size;
}

xtPublic xtInt8 xt_ind_get_size()
{
	xtInt8 size = 0;

	size = (xtInt8) ind_cac_globals.cg_block_count * (xtInt8) sizeof(DcBlockRec);
	return size;
}

xtPublic void xt_ind_check_cache(void)
{
	DcBlockPtr		block, pblock;
	u_int			used_count;
	u_int			free_count;
	DcSegmentPtr	seg;

	// Check the dirty list:
	for (u_int i=0; i<IDX_CAC_SEGMENT_COUNT; i++) {
		for (u_int j=0; j<XT_INDEX_CACHE_FILE_SLOTS; j++) {
			pblock = NULL;
			block = ind_cac_globals.cg_segment[i].cs_dirty_list[j];
			while (block) {
				ASSERT_NS(block->cb_state == IDX_CAC_BLOCK_DIRTY);
				ASSERT_NS(block->cb_dirty_prev == pblock);
				ASSERT_NS(block->cb_file_id % XT_INDEX_CACHE_FILE_SLOTS == j);
				pblock = block;
				block = block->cb_dirty_next;
			}
		}
	}

	// Check the LRU list:
	used_count = 0;
	pblock = NULL;
	block = ind_cac_globals.cg_lru_block;
	while (block) {
		used_count++;
		ASSERT_NS(block->cb_state != IDX_CAC_BLOCK_FREE);
		ASSERT_NS(block->cb_lr_used == pblock);
		if (block->cb_state == IDX_CAC_BLOCK_DIRTY) {
			if (!block->cb_dirty_prev) {
				//seg = &ind_cac_globals.cg_segment[(block->cb_address >> XT_INDEX_PAGE_SHIFTS) & IDX_CAC_SEGMENT_MASK];
				seg = &ind_cac_globals.cg_segment[((u_int) block->cb_address >> XT_INDEX_PAGE_SHIFTS) & IDX_CAC_SEGMENT_MASK];
				ASSERT_NS(seg->cs_dirty_list[block->cb_file_id % XT_INDEX_CACHE_FILE_SLOTS] == block);
			}
		}
		pblock = block;
		block = block->cb_mr_used;
	}
	ASSERT_NS(ind_cac_globals.cg_mru_block == pblock);
	ASSERT_NS(ind_cac_globals.cg_free_count + used_count == ind_cac_globals.cg_block_count);

	// Check the free list:
	free_count = 0;
	block = ind_cac_globals.cg_free_list;
	while (block) {
		free_count++;
		ASSERT_NS(block->cb_state == IDX_CAC_BLOCK_FREE);
		block = block->cb_next;
	}
	ASSERT_NS(ind_cac_globals.cg_free_count == free_count);
}

#ifdef FOR_DEBUG
static void ind_cac_check_on_dirty_list(DcSegmentPtr seg, DcBlockPtr block)
{
	DcBlockPtr	list_block, plist_block;
	xtBool		found = FALSE;
	
	plist_block = NULL;
	list_block = seg->cs_dirty_list[block->cb_file_id % XT_INDEX_CACHE_FILE_SLOTS];
	while (list_block) {
		ASSERT_NS(list_block->cb_state == IDX_CAC_BLOCK_DIRTY);
		ASSERT_NS(list_block->cb_dirty_prev == plist_block);
		if (list_block == block)
			found = TRUE;
		plist_block = list_block;
		list_block = list_block->cb_dirty_next;
	}
	ASSERT_NS(found);
}

static void ind_cac_check_dirty_list(DcSegmentPtr seg, DcBlockPtr block)
{
	DcBlockPtr list_block, plist_block;
	
	for (u_int j=0; j<XT_INDEX_CACHE_FILE_SLOTS; j++) {
		plist_block = NULL;
		list_block = seg->cs_dirty_list[j];
		while (list_block) {
			ASSERT_NS(list_block->cb_state == IDX_CAC_BLOCK_DIRTY);
			ASSERT_NS(block != list_block);
			ASSERT_NS(list_block->cb_dirty_prev == plist_block);
			plist_block = list_block;
			list_block = list_block->cb_dirty_next;
		}
	}
}

static void ind_cac_check_lru_list(DcBlockPtr block)
{
	DcBlockPtr list_block, plist_block;
	
	plist_block = NULL;
	list_block = ind_cac_globals.cg_lru_block;
	while (list_block) {
		ASSERT_NS(block != list_block);
		ASSERT_NS(list_block->cb_lr_used == plist_block);
		plist_block = list_block;
		list_block = list_block->cb_mr_used;
	}
	ASSERT_NS(ind_cac_globals.cg_mru_block == plist_block);
}
#endif

static void ind_cac_add_to_flush_buffer(u_int flush_count, DcBlockPtr *flush_buffer, DcBlockPtr block)
{
	register u_int		count = flush_count;
	register u_int		i;
	register u_int		guess;
	register xtInt8		r;

	i = 0;
	while (i < count) {
		guess = (i + count - 1) >> 1;
		r = (xtInt8) block->cb_address - (xtInt8) flush_buffer[guess]->cb_address;
		if (r == 0) {
			// Should not happen...
			ASSERT_NS(FALSE);
			return;
		}
		if (r < (xtInt8) 0)
			count = guess;
		else
			i = guess + 1;
	}

	/* Insert at position i */
	memmove(flush_buffer + i + 1, flush_buffer + i, (flush_count - i) * sizeof(DcBlockPtr));
	flush_buffer[i] = block;
}

xtPublic xtBool xt_ind_flush(XTOpenFilePtr file, xtBool *something_written)
{
	DcSegmentPtr	seg;
	DcBlockPtr		block, nblock;
	u_int			flush_count;
	DcBlockPtr		flush_buffer[IDX_CAC_MAX_FLUSH_COUNT];
	xtBool			more_to_flush = TRUE;
	u_int			file_idx = file->of_id % XT_INDEX_CACHE_FILE_SLOTS;

#ifdef XT_USE_MYSYS
	flush_key_blocks(&my_cache, file->of_filedes, FLUSH_KEEP);
	return OK;
#endif

#ifdef DEBUG_CHECK_CACHE
	//xt_ind_check_cache();
#endif
	while (more_to_flush) {
		flush_count = 0;
		more_to_flush = FALSE;

		for (u_int i=0; i<IDX_CAC_SEGMENT_COUNT; i++) {
			if (ind_cac_globals.cg_segment[i].cs_dirty_list[file_idx]) {
				seg = &ind_cac_globals.cg_segment[i];

				xt_lock_mutex_ns(&seg->cs_lock);
				
				block = seg->cs_dirty_list[file_idx];
				while (block) {
					nblock = block->cb_dirty_next;

					if (file->of_id == block->cb_file_id) {
						if (flush_count >= XT_INDEX_CACHE_MAX_FLUSH_SEG_CNT * (i+1)) {
							more_to_flush = TRUE;
							break;
						}

						ind_cac_add_to_flush_buffer(flush_count, flush_buffer, block);
						flush_count++;
						/* Remove from dirty list for writing: */
						if (seg->cs_dirty_list[file_idx] == block)
							seg->cs_dirty_list[file_idx] = block->cb_dirty_next;
						if (block->cb_dirty_next)
							block->cb_dirty_next->cb_dirty_prev = block->cb_dirty_prev;
						if (block->cb_dirty_prev)
							block->cb_dirty_prev->cb_dirty_next = block->cb_dirty_next;
						block->cb_dirty_next = NULL;
						block->cb_dirty_prev = NULL;
						block->cb_state = IDX_CAC_BLOCK_WRITING;
					}

					block = nblock;
				}

				xt_unlock_mutex_ns(&seg->cs_lock);
			}
		}

		if (!flush_count)
			break;

		for (u_int i=0; i<flush_count; i++) {
			*something_written = TRUE;
			if (!xt_pwrite_file(file, flush_buffer[i]->cb_address, XT_INDEX_PAGE_SIZE, flush_buffer[i]->cb_data)) {
				for (; i<flush_count; i++) {
					block = flush_buffer[i];
					//seg = &ind_cac_globals.cg_segment[(block->cb_address >> XT_INDEX_PAGE_SHIFTS) & IDX_CAC_SEGMENT_MASK];
					seg = &ind_cac_globals.cg_segment[((u_int) block->cb_address >> XT_INDEX_PAGE_SHIFTS) & IDX_CAC_SEGMENT_MASK];
					xt_lock_mutex_ns(&seg->cs_lock);

					/* Put it back on the list as dirty: */
					if (seg->cs_dirty_list[file_idx])
						seg->cs_dirty_list[file_idx]->cb_dirty_prev = block;
					block->cb_dirty_prev = NULL;
					block->cb_dirty_next = seg->cs_dirty_list[file_idx];
					seg->cs_dirty_list[file_idx] = block;
					block->cb_state = IDX_CAC_BLOCK_DIRTY;

					xt_cond_wakeall(&seg->cs_cond);
					xt_unlock_mutex_ns(&seg->cs_lock);
				}
#ifdef DEBUG_CHECK_CACHE
				//xt_ind_check_cache();
#endif
				return FAILED;
			}
			flush_buffer[i]->cb_state = IDX_CAC_BLOCK_CLEAN;
		}

		/* Wake all waiting for the flush: */
		for (u_int i=0; i<IDX_CAC_SEGMENT_COUNT; i++)
			xt_cond_wakeall(&ind_cac_globals.cg_segment[i].cs_cond);
	}
#ifdef DEBUG_CHECK_CACHE
	//xt_ind_check_cache();
#endif
	return OK;
}

/*
 * Free the least recently used block, return a pointer to the block
 */
static xtBool ind_cac_free_lru_blocks(XTOpenFilePtr of)
{
	DcBlockPtr		block, pblock;
	xtWord4			ru_time;
	u_int			file_id;
	off_t			address;
	//off_t			hash;
	DcSegmentPtr	seg;
	u_int			hash_idx;
	xtBool			have_global_lock = FALSE;

#ifdef DEBUG_CHECK_CACHE
	//xt_ind_check_cache();
#endif
	retry:
	if (!(block = ind_cac_globals.cg_lru_block))
		return OK;

	retry_2:
	ru_time = block->cb_ru_time;
	file_id = block->cb_file_id;
	address = block->cb_address;

	/*
	hash = (address >> XT_INDEX_PAGE_SHIFTS) ^ ((off_t) file_id << 28);
	seg = &ind_cac_globals.cg_segment[hash & IDX_CAC_SEGMENT_MASK];
	hash_idx = (hash >> XT_INDEX_CACHE_SEGMENT_SHIFTS) % ind_cac_globals.cg_hash_size;
	*/
	seg = &ind_cac_globals.cg_segment[((u_int) address >> XT_INDEX_PAGE_SHIFTS) & IDX_CAC_SEGMENT_MASK];
	hash_idx = (((u_int) (address >> (XT_INDEX_CACHE_SEGMENT_SHIFTS + XT_INDEX_PAGE_SHIFTS))) ^ (file_id << 16)) % ind_cac_globals.cg_hash_size;

	xt_lock_mutex_ns(&seg->cs_lock);

	free_more:
	pblock = NULL;
	block = seg->cs_hash_table[hash_idx];
	while (block) {
		if (block->cb_address == address && block->cb_file_id == file_id) {
			ASSERT_NS(block->cb_state != IDX_CAC_BLOCK_FREE);
			
			/* Try again if the block has been used in the meantime: */
			if (ru_time != block->cb_ru_time) {
				if (have_global_lock)
					/* Having this lock means we have already freed at least one block so
					 * don't bother to free more if we are having trouble.
					 */
					goto done_ok;

				xt_unlock_mutex_ns(&seg->cs_lock);
				goto retry;
			}

			/* Wait if the block is being read or written. */
			if (block->cb_state == IDX_CAC_BLOCK_READING || block->cb_state == IDX_CAC_BLOCK_WRITING) {
				if (have_global_lock)
					goto done_ok;

				if (!xt_timed_wait_cond_ns(&seg->cs_cond, &seg->cs_lock, &seg->cs_cond_lock, 100))
					goto failed;
				xt_unlock_mutex_ns(&seg->cs_lock);
				goto retry;
			}
			
			/* Skip this block if we cannot cannot clean it! */
			if (block->cb_state == IDX_CAC_BLOCK_DIRTY && of->of_id != file_id) {
				block = block->cb_mr_used;

				if (have_global_lock)
					goto done_ok;

				xt_unlock_mutex_ns(&seg->cs_lock);
				if (!block)
					return OK;
				goto retry_2;
			}
			goto free_the_block;
		}
		pblock = block;
		block = block->cb_next;
	}

	if (have_global_lock) {
		xt_unlock_mutex_ns(&ind_cac_globals.cg_lock);
		have_global_lock = FALSE;
	}
	xt_unlock_mutex_ns(&seg->cs_lock);
	goto retry;

	free_the_block:
	ASSERT_NS(block->cb_state == IDX_CAC_BLOCK_CLEAN || block->cb_state == IDX_CAC_BLOCK_DIRTY);
	if (block->cb_state == IDX_CAC_BLOCK_DIRTY) {
		u_long file_idx = file_id % XT_INDEX_CACHE_FILE_SLOTS;

		ASSERT_NS(of->of_id == file_id);

		if (have_global_lock) {
			xt_unlock_mutex_ns(&ind_cac_globals.cg_lock);
			have_global_lock = FALSE;
		}

		/* Remove from dirty list for writing: */
		if (seg->cs_dirty_list[file_idx] == block)
			seg->cs_dirty_list[file_idx] = block->cb_dirty_next;
		if (block->cb_dirty_next)
			block->cb_dirty_next->cb_dirty_prev = block->cb_dirty_prev;
		if (block->cb_dirty_prev)
			block->cb_dirty_prev->cb_dirty_next = block->cb_dirty_next;
		block->cb_dirty_next = NULL;
		block->cb_dirty_prev = NULL;
		block->cb_state = IDX_CAC_BLOCK_WRITING;

		xt_unlock_mutex_ns(&seg->cs_lock);
		if (!xt_pwrite_file(of, address, XT_INDEX_PAGE_SIZE, block->cb_data)) {
			xt_lock_mutex_ns(&seg->cs_lock);

			/* Put it back on the list as dirty: */
			if (seg->cs_dirty_list[file_idx])
				seg->cs_dirty_list[file_idx]->cb_dirty_prev = block;
			block->cb_dirty_prev = NULL;
			block->cb_dirty_next = seg->cs_dirty_list[file_idx];
			seg->cs_dirty_list[file_idx] = block;
			block->cb_state = IDX_CAC_BLOCK_DIRTY;

			xt_cond_wakeall(&seg->cs_cond);
			goto failed;
		}
		xt_lock_mutex_ns(&seg->cs_lock);
		block->cb_state = IDX_CAC_BLOCK_CLEAN;
		xt_cond_wakeall(&seg->cs_cond);
	}

	/* Block is clean, remove from the hash table: */
	if (pblock)
		pblock->cb_next = block->cb_next;
	else
		seg->cs_hash_table[hash_idx] = block->cb_next;

	/* Now free the block */
	if (!have_global_lock) {
		xt_lock_mutex_ns(&ind_cac_globals.cg_lock);
		have_global_lock = TRUE;
	}

	/* Remove from the MRU list: */
	if (ind_cac_globals.cg_lru_block == block)
		ind_cac_globals.cg_lru_block = block->cb_mr_used;
	if (ind_cac_globals.cg_mru_block == block)
		ind_cac_globals.cg_mru_block = block->cb_lr_used;
	if (block->cb_lr_used)
		block->cb_lr_used->cb_mr_used = block->cb_mr_used;
	if (block->cb_mr_used)
		block->cb_mr_used->cb_lr_used = block->cb_lr_used;

	/* Put the block on the free list: */
	block->cb_next = ind_cac_globals.cg_free_list;
	ind_cac_globals.cg_free_list = block;
	ind_cac_globals.cg_free_count++;
	block->cb_state = IDX_CAC_BLOCK_FREE;

	if (ind_cac_globals.cg_free_count < XT_INDEX_CACHE_MAX_FREE_COUNT) {
		/* Now that we have all the locks, try to free some more in this segment: */
		block = block->cb_mr_used;
		for (u_int i=0; block && i<IDX_CAC_SEGMENT_COUNT; i++) {
			ru_time = block->cb_ru_time;
			file_id = block->cb_file_id;
			address = block->cb_address;

			/*
			hash = (address >> XT_INDEX_PAGE_SHIFTS) ^ ((off_t) file_id << 28);
			if (seg == &ind_cac_globals.cg_segment[hash & IDX_CAC_SEGMENT_MASK]) {
				hash_idx = (hash >> XT_INDEX_CACHE_SEGMENT_SHIFTS) % ind_cac_globals.cg_hash_size;
				goto free_more;
			}
			*/
			if (seg == &ind_cac_globals.cg_segment[((u_int) address >> XT_INDEX_PAGE_SHIFTS) & IDX_CAC_SEGMENT_MASK]) {
				hash_idx = (((u_int) (address >> (XT_INDEX_CACHE_SEGMENT_SHIFTS + XT_INDEX_PAGE_SHIFTS))) ^ (file_id << 16)) % ind_cac_globals.cg_hash_size;
				goto free_more;
			}
			block = block->cb_mr_used;
		}
	}

	done_ok:
	xt_unlock_mutex_ns(&ind_cac_globals.cg_lock);
	xt_unlock_mutex_ns(&seg->cs_lock);
#ifdef DEBUG_CHECK_CACHE
	//xt_ind_check_cache();
#endif
	return OK;
	
	failed:
	xt_unlock_mutex_ns(&seg->cs_lock);
#ifdef DEBUG_CHECK_CACHE
	//xt_ind_check_cache();
#endif
	return FAILED;
}

#define IDX_CAC_NOTHING			0
#define IDX_CAC_READ				1
#define IDX_CAC_REMOVE_AND_READ	2

static DcBlockPtr ind_cac_find(XTOpenFilePtr file, off_t address, DcSegmentPtr *ret_seg, xtBool will_update)
{
	register DcSegmentPtr	seg;
	register DcBlockPtr		block;
	register u_int			hash_idx;
	register DcGlobalsRec	*dcg = &ind_cac_globals;

	seg = &dcg->cg_segment[((u_int) address >> XT_INDEX_PAGE_SHIFTS) & IDX_CAC_SEGMENT_MASK];
	hash_idx = (((u_int) (address >> (XT_INDEX_CACHE_SEGMENT_SHIFTS + XT_INDEX_PAGE_SHIFTS))) ^ (file->of_id << 16)) % dcg->cg_hash_size;

	xt_lock_mutex_ns(&seg->cs_lock);
	retry:
	block = seg->cs_hash_table[hash_idx];
	while (block) {
		if (block->cb_address == address && block->cb_file_id == file->of_id) {
			ASSERT_NS(block->cb_state != IDX_CAC_BLOCK_FREE);

			/* Wait if the block is being read or written.
			 * If we will just read the data, then we don't care
			 * if the buffer is being written.
			 */
			if (block->cb_state == IDX_CAC_BLOCK_READING || (will_update && block->cb_state == IDX_CAC_BLOCK_WRITING)) {
				if (!xt_timed_wait_cond_ns(&seg->cs_cond, &seg->cs_lock, &seg->cs_cond_lock, 100))
					break;
				goto retry;
			}

			*ret_seg = seg;
			return block;
		}
		block = block->cb_next;
	}
	
	/* Block not found: */
	xt_unlock_mutex_ns(&seg->cs_lock);
	return NULL;
}

static DcBlockPtr ind_cac_fetch(XTOpenFilePtr file, off_t address, DcSegmentPtr *ret_seg, xtBool will_update)
{
	register DcBlockPtr		block;
	register DcSegmentPtr	seg;
	register u_int			hash_idx;
	register DcGlobalsRec	*dcg = &ind_cac_globals;
	size_t					red_size;

	seg = &dcg->cg_segment[((u_int) address >> XT_INDEX_PAGE_SHIFTS) & IDX_CAC_SEGMENT_MASK];
	hash_idx = (((u_int) (address >> (XT_INDEX_CACHE_SEGMENT_SHIFTS + XT_INDEX_PAGE_SHIFTS))) ^ (file->of_id << 16)) % dcg->cg_hash_size;

	xt_lock_mutex_ns(&seg->cs_lock);
	retry:
	block = seg->cs_hash_table[hash_idx];
	while (block) {
		if (block->cb_address == address && block->cb_file_id == file->of_id) {
			ASSERT_NS(block->cb_state != IDX_CAC_BLOCK_FREE);

			/* Wait if the block is being read or written.
			 * If we will just read the data, then we don't care
			 * if the buffer is being written.
			 */
			if (block->cb_state == IDX_CAC_BLOCK_READING || (will_update && block->cb_state == IDX_CAC_BLOCK_WRITING)) {
				if (!xt_timed_wait_cond_ns(&seg->cs_cond, &seg->cs_lock, &seg->cs_cond_lock, 100)) {
					xt_unlock_mutex_ns(&seg->cs_lock);
					return NULL;
				}
				goto retry;
			}

			/* This block has been most recently used: */
			if (XT_TIME_DIFF(block->cb_ru_time, dcg->cg_ru_now) > (dcg->cg_block_count >> 1)) {
				/* Move to the front of the MRU list: */
				xt_lock_mutex_ns(&dcg->cg_lock);

				block->cb_ru_time = ++dcg->cg_ru_now;
				if (dcg->cg_mru_block != block) {
					/* Remove from the MRU list: */
					if (dcg->cg_lru_block == block)
						dcg->cg_lru_block = block->cb_mr_used;
					if (block->cb_lr_used)
						block->cb_lr_used->cb_mr_used = block->cb_mr_used;
					if (block->cb_mr_used)
						block->cb_mr_used->cb_lr_used = block->cb_lr_used;
	
					/* Make the block the most recently used: */
					if ((block->cb_lr_used = dcg->cg_mru_block))
						dcg->cg_mru_block->cb_mr_used = block;
					block->cb_mr_used = NULL;
					dcg->cg_mru_block = block;
					if (!dcg->cg_lru_block)
						dcg->cg_lru_block = block;
				}
				xt_unlock_mutex_ns(&dcg->cg_lock);
			}
			*ret_seg = seg;
#ifdef DEBUG_CHECK_CACHE
			//xt_ind_check_cache();
#endif
			return block;
		}
		block = block->cb_next;
	}
	
	/* Block not found: */
	free_some_blocks:
	if (!dcg->cg_free_list) {
		xt_unlock_mutex_ns(&seg->cs_lock);
		if (!ind_cac_free_lru_blocks(file))
			return NULL;
		xt_lock_mutex_ns(&seg->cs_lock);
		goto retry;
	}

	xt_lock_mutex_ns(&dcg->cg_lock);

	/* Get a free block: */
	if (!(block = dcg->cg_free_list)) {
		xt_unlock_mutex_ns(&dcg->cg_lock);
		goto free_some_blocks;
	}

	ASSERT_NS(block->cb_state == IDX_CAC_BLOCK_FREE);
	dcg->cg_free_list = block->cb_next;
	dcg->cg_free_count--;

	block->cb_address = address;
	block->cb_file_id = file->of_id;
	block->cb_state = IDX_CAC_BLOCK_READING;

	/* Make the block the most recently used: */
	block->cb_ru_time = ++dcg->cg_ru_now;
	if ((block->cb_lr_used = dcg->cg_mru_block))
		dcg->cg_mru_block->cb_mr_used = block;
	block->cb_mr_used = NULL;
	dcg->cg_mru_block = block;
	if (!dcg->cg_lru_block)
		dcg->cg_lru_block = block;

	xt_unlock_mutex_ns(&dcg->cg_lock);

	/* Add the block to the hash table: */
	block->cb_next = seg->cs_hash_table[hash_idx];
	seg->cs_hash_table[hash_idx] = block;
	block->cb_dirty_next = NULL;
	block->cb_dirty_prev = NULL;

	/* Read the block into memory: */
	xt_unlock_mutex_ns(&seg->cs_lock);

	if (!xt_pread_file(file, address, XT_INDEX_PAGE_SIZE, 0, block->cb_data, &red_size))
		return NULL;
	memset(block->cb_data + red_size, 0, XT_INDEX_PAGE_SIZE - red_size);

	xt_lock_mutex_ns(&seg->cs_lock);
	block->cb_state = IDX_CAC_BLOCK_CLEAN;
	xt_cond_wakeall(&seg->cs_cond);

	*ret_seg = seg;
#ifdef DEBUG_CHECK_CACHE
	//xt_ind_check_cache();
#endif
	return block;
}

xtPublic xtBool xt_ind_write_thru(XTOpenFilePtr file, off_t offset, size_t size, xtWord1 *data)
{
	off_t			address;
	DcBlockPtr		block;
	DcSegmentPtr	seg;
	size_t			boff;
	size_t			tfer;

#ifdef DEBUG_CHECK_CACHE
	//xt_ind_check_cache();
#endif
	/* Write the file: */
	if (!xt_pwrite_file(file, offset, size, data))
		return FAILED;

	/* Update the cache (if necessary): */
	address = offset & ~XT_INDEX_PAGE_MASK;
	boff = (size_t) (offset - address);
	tfer = XT_INDEX_PAGE_SIZE - boff;
	if (tfer > size)
		tfer = size;
	while (size > 0) {
		block = ind_cac_find(file, address, &seg, TRUE);
		if (block) {
			ASSERT_NS(block->cb_state == IDX_CAC_BLOCK_CLEAN || block->cb_state == IDX_CAC_BLOCK_DIRTY);
			memcpy(block->cb_data + boff, data, tfer);
			xt_unlock_mutex_ns(&seg->cs_lock);
		}
		size -= tfer;
		data += tfer;
		address += XT_INDEX_PAGE_SIZE;
		boff = 0;
		tfer = size;
		if (tfer > XT_INDEX_PAGE_SIZE)
			tfer = XT_INDEX_PAGE_SIZE;
	}
#ifdef DEBUG_CHECK_CACHE
	//xt_ind_check_cache();
#endif
	return OK;
}

xtPublic xtBool xt_ind_write(XTOpenFilePtr file, off_t offset, size_t size, xtWord1 *data)
{
	off_t			address;
	DcBlockPtr		block;
	DcSegmentPtr	seg;
	size_t			boff;
	size_t			tfer;

#ifdef DEBUG_CHECK_CACHE
	//xt_ind_check_cache();
#endif
	address = offset & ~XT_INDEX_PAGE_MASK;
	boff = (size_t) (offset - address);
	tfer = XT_INDEX_PAGE_SIZE - boff;
	if (tfer > size)
		tfer = size;
	while (size > 0) {
		block = ind_cac_fetch(file, address, &seg, TRUE);
		if (!block) {
#ifdef DEBUG_CHECK_CACHE
			//xt_ind_check_cache();
#endif
			return FAILED;
		}
		ASSERT_NS(block->cb_state == IDX_CAC_BLOCK_CLEAN || block->cb_state == IDX_CAC_BLOCK_DIRTY);
		memcpy(block->cb_data + boff, data, tfer);
		if (block->cb_state != IDX_CAC_BLOCK_DIRTY) {
			u_long file_idx = ((u_long) file->of_id) % XT_INDEX_CACHE_FILE_SLOTS;

//printf("PMC FILE %p IDX %lu\n", file, file_idx);
			if (seg->cs_dirty_list[file_idx])
				seg->cs_dirty_list[file_idx]->cb_dirty_prev = block;
			block->cb_dirty_prev = NULL;
			block->cb_dirty_next = seg->cs_dirty_list[file_idx];
			seg->cs_dirty_list[file_idx] = block;
			block->cb_state = IDX_CAC_BLOCK_DIRTY;
		}
		xt_unlock_mutex_ns(&seg->cs_lock);
		size -= tfer;
		data += tfer;
		address += XT_INDEX_PAGE_SIZE;
		boff = 0;
		tfer = size;
		if (tfer > XT_INDEX_PAGE_SIZE)
			tfer = XT_INDEX_PAGE_SIZE;
	}
#ifdef DEBUG_CHECK_CACHE
	//xt_ind_check_cache();
#endif
	return OK;
}

xtPublic xtBool xt_ind_read(XTOpenFilePtr file, off_t offset, size_t size, xtWord1 *data)
{
	off_t			address;
	DcBlockPtr		block;
	DcSegmentPtr	seg;
	size_t			boff;
	size_t			tfer;

#ifdef DEBUG_CHECK_CACHE
	//xt_ind_check_cache();
#endif
	address = offset & ~XT_INDEX_PAGE_MASK;
	boff = (size_t) (offset - address);
	tfer = XT_INDEX_PAGE_SIZE - boff;
	if (tfer > size)
		tfer = size;
	while (size > 0) {
		block = ind_cac_fetch(file, address, &seg, FALSE);
		if (!block) {
#ifdef DEBUG_CHECK_CACHE
			//xt_ind_check_cache();
#endif
			return FAILED;
		}
		memcpy(data, block->cb_data + boff, tfer);
		xt_unlock_mutex_ns(&seg->cs_lock);
		size -= tfer;
		data += tfer;
		address += XT_INDEX_PAGE_SIZE;
		boff = 0;
		tfer = size;
		if (tfer > XT_INDEX_PAGE_SIZE)
			tfer = XT_INDEX_PAGE_SIZE;
	}
#ifdef DEBUG_CHECK_CACHE
	//xt_ind_check_cache();
#endif
	return OK;
}

/*
 * An optimised read for 8 bytes (used by the row table).
 */
xtPublic xtBool xt_ind_read_8(XTOpenFilePtr file, off_t offset, xtWord8 *data)
{
	register DcBlockPtr		block;
	DcSegmentPtr			seg;
	off_t					address;

	address = offset & ~XT_INDEX_PAGE_MASK;
	if (!(block = ind_cac_fetch(file, address, &seg, FALSE)))
		return FAILED;
	ASSERT_NS(offset - address + 8 <= XT_INDEX_PAGE_SIZE);
	*data = *((xtWord8 *) (block->cb_data + (offset - address)));
	xt_unlock_mutex_ns(&seg->cs_lock);
	return OK;
}

xtPublic xtBool xt_ind_read_block(XTOpenFilePtr file, off_t address, xtWord1 *data)
{
	register DcBlockPtr		block;
	DcSegmentPtr			seg;

	ASSERT_NS(address == (address & ~XT_INDEX_PAGE_MASK));
	if (!(block = ind_cac_fetch(file, address, &seg, FALSE)))
		return FAILED;
	memcpy(data, block->cb_data, XT_INDEX_PAGE_SIZE);
	xt_unlock_mutex_ns(&seg->cs_lock);
	return OK;
}



