/* Copyright (c) 2005 PrimeBase Technologies GmbH
 *
 * PrimeBase XT
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * 2005-01-12	Paul McCullagh
 *
 * H&G2JCtL
 */

#include "xt_config.h"

#ifndef XT_WIN
#include <unistd.h>
#include <dirent.h>
#endif
#include <stdio.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/types.h>
#include <ctype.h>
#include <string.h>
#include <errno.h>

#include "strutil_xt.h"
#include "thread_xt.h"
#include "filesys_xt.h"
#include "memory_xt.h"
#include "cache_xt.h"
#include "sortedlist_xt.h"
#include "trace_xt.h"

#ifdef DEBUG
//#define DEBUG_PRINT_IO
//#define DEBUG_TRACE_IO
#endif

/* ----------------------------------------------------------------------
 * Globals
 */

typedef struct FsGlobals {
	xt_mutex_type		*fsg_lock;						/* The xtPublic cache lock. */
	u_int				fsg_current_id;
	XTSortedListPtr		fsg_open_files;
} FsGlobalsRec;

static FsGlobalsRec	fs_globals;

#ifdef XT_WIN
static int fs_get_win_error()
{
	return (int) GetLastError();
}

xtPublic void xt_get_win_message(char *buffer, size_t size, int err)
{
	FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM, NULL, err,
		MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
        buffer,
        size, NULL);
}
#endif

/* ----------------------------------------------------------------------
 * Open file list
 */

static XTFilePtr fs_new_file(XTThreadPtr self, char *file)
{
	XTFilePtr file_ptr;

	pushsr_(file_ptr, xt_free, (XTFilePtr) xt_calloc(self, sizeof(XTFileRec)));

	file_ptr->fil_path = xt_dup_string(self, file);
	file_ptr->fil_id = fs_globals.fsg_current_id++;
	if (!fs_globals.fsg_current_id)
		fs_globals.fsg_current_id++;
#ifndef XT_WIN
	file_ptr->fil_filedes = -1;
#endif

	popr_(); // Discard xt_free(file_ptr)
	return file_ptr;
}

static void fs_free_file(XTThreadPtr self, void *thunk, void *item)
{
#pragma unused(thunk)
	XTFilePtr	file_ptr = *((XTFilePtr *) item);

#ifndef XT_WIN
	if (file_ptr->fil_filedes != -1) {
		close(file_ptr->fil_filedes);
		//printf("close (FILE) %d %s\n", file_ptr->fil_filedes, file_ptr->fil_path);
		file_ptr->fil_filedes = -1;
	}
#endif

	/* Flush any cache before this file is invalid: */
	if (file_ptr->fil_path) {
		xt_free(self, file_ptr->fil_path);
		file_ptr->fil_path = NULL;
	}

	xt_free(self, file_ptr);
}

static int fs_comp_file(XTThreadPtr self, register const void *thunk, register const void *a, register const void *b)
{
#pragma unused(self, thunk)
	char		*file_name = (char *) a;
	XTFilePtr	file_ptr = *((XTFilePtr *) b);

	return strcmp(file_name, file_ptr->fil_path);
}

static int fs_comp_file_ci(XTThreadPtr self, register const void *thunk, register const void *a, register const void *b)
{
#pragma unused(self, thunk)
	char		*file_name = (char *) a;
	XTFilePtr	file_ptr = *((XTFilePtr *) b);

	return strcasecmp(file_name, file_ptr->fil_path);
}

/* ----------------------------------------------------------------------
 * init & exit
 */

xtPublic void xt_fs_init(XTThreadPtr self)
{
	fs_globals.fsg_open_files = xt_new_sortedlist(self,
		sizeof(XTFilePtr), 20, 20,
		pbxt_ignore_case ? fs_comp_file_ci : fs_comp_file,
		NULL, fs_free_file, TRUE, FALSE);
	fs_globals.fsg_lock = fs_globals.fsg_open_files->sl_lock;
	fs_globals.fsg_current_id = 1;
}

xtPublic void xt_fs_exit(XTThreadPtr self)
{
	if (fs_globals.fsg_open_files) {
		xt_free_sortedlist(self, fs_globals.fsg_open_files);
		fs_globals.fsg_open_files = NULL;
	}
	fs_globals.fsg_lock = NULL;
	fs_globals.fsg_current_id = 0;
}

/* ----------------------------------------------------------------------
 * File operations
 */

static void fs_set_stats(XTThreadPtr self, char *path)
{
	char		super_path[PATH_MAX];
	struct stat	stats;
	char		*ptr;

	ptr = xt_last_name_of_path(path);
	if (ptr == path) 
		strcpy(super_path, ".");
	else {
		xt_strcpy(PATH_MAX, super_path, path);

		if ((ptr = xt_last_name_of_path(super_path)))
			*ptr = 0;
	}
	if (stat(super_path, &stats) == -1)
		xt_throw_ferrno(XT_CONTEXT, errno, super_path);

	if (chmod(path, stats.st_mode) == -1)
		xt_throw_ferrno(XT_CONTEXT, errno, path);

	/*chown(path, stats.st_uid, stats.st_gid);*/
}

xtPublic char *xt_file_path(XTOpenFilePtr of)
{
	return of->of_file->fil_path;
}

xtBool xt_fs_exists(char *path)
{
	int err;

	err = access(path, F_OK);
	if (err == -1)
		return FALSE;
	return TRUE;
}

#ifdef XT_WIN
#define XT_INVALID_HANDLE		INVALID_HANDLE_VALUE
#else
#define XT_INVALID_HANDLE		(-1)
#endif

xtPublic xtBool xt_fs_delete(XTThreadPtr self, char *name)
{
#ifdef XT_WIN
	//printf("delete %s\n", name);
	if (!DeleteFile(name)) {
		int err = fs_get_win_error();

		if (!XT_FILE_NOT_FOUND(err)) {
			xt_throw_ferrno(XT_CONTEXT, err, name);
			return FAILED;
		}
	}
#else
	if (unlink(name) == -1) {
		int err = errno;

		if (err != ENOENT) {
			xt_throw_ferrno(XT_CONTEXT, err, name);
			return FAILED;
		}
	}
#endif
	return OK;
}

xtPublic xtBool xt_fs_file_not_found(int err)
{
#ifdef XT_WIN
	return XT_FILE_NOT_FOUND(err);
#else
	return err == ENOENT;
#endif
}

xtPublic void xt_fs_move(struct XTThread *self, char *from_path, char *to_path)
{
#ifdef XT_WIN
	if (!MoveFile(from_path, to_path))
		xt_throw_ferrno(XT_CONTEXT, fs_get_win_error(), from_path);
#else
	int err;

	if (link(from_path, to_path) == -1) {
		err = errno;
		xt_throw_ferrno(XT_CONTEXT, err, from_path);
	}

	if (unlink(from_path) == -1) {
		err = errno;
		unlink(to_path);
		xt_throw_ferrno(XT_CONTEXT, err, from_path);
	}
#endif
}

xtPublic xtBool xt_fs_rename(struct XTThread *self, char *from_path, char *to_path)
{
	int err;

	if (rename(from_path, to_path) == -1) {
		err = errno;
		xt_throw_ferrno(XT_CONTEXT, err, from_path);
		return FAILED;
	}
	return OK;
}

xtPublic xtBool xt_fs_stat(XTThreadPtr self, char *path, off_t *size, struct timespec *mod_time)
{
#ifdef XT_WIN
	HANDLE						fh;
	BY_HANDLE_FILE_INFORMATION	info;
	SECURITY_ATTRIBUTES			sa = { sizeof(SECURITY_ATTRIBUTES), 0, 0 };

	fh = CreateFile(
		path,
		GENERIC_READ,
		FILE_SHARE_READ,
		&sa,
		OPEN_EXISTING,
		FILE_ATTRIBUTE_NORMAL,
		NULL);
	if (fh == INVALID_HANDLE_VALUE) {
		xt_throw_ferrno(XT_CONTEXT, fs_get_win_error(), path);
		return FAILED;
	}

	if (!GetFileInformationByHandle(fh, &info)) {
		CloseHandle(fh);
		xt_throw_ferrno(XT_CONTEXT, fs_get_win_error(), path);
		return FAILED;
	}

	CloseHandle(fh);
	if (size)
		*size = (off_t) info.nFileSizeLow | (((off_t) info.nFileSizeHigh) << 32);
	if (mod_time)
		mod_time->tv.ft = info.ftLastWriteTime;
#else
	struct stat sb;

	if (stat(path, &sb) == -1) {
		xt_throw_ferrno(XT_CONTEXT, errno, path);
		return FAILED;
	}
	if (size)
		*size = sb.st_size;
	if (mod_time) {
		mod_time->tv_sec = sb.st_mtime;
#ifdef __USE_MISC
		/* This is the Linux version: */
		mod_time->tv_nsec = sb.st_mtim.tv_nsec;
#else
		/* This is the Mac OS X version: */
		mod_time->tv_nsec = sb.st_mtimespec.tv_nsec;
#endif
	}
#endif
	return OK;
}

void xt_fs_mkdir(XTThreadPtr self, char *name)
{
	char path[PATH_MAX];

	xt_strcpy(PATH_MAX, path, name);
	xt_remove_dir_char(path);

#ifdef XT_WIN
	{
		SECURITY_ATTRIBUTES	sa = { sizeof(SECURITY_ATTRIBUTES), 0, 0 };

		if (!CreateDirectory(path, &sa))
			xt_throw_ferrno(XT_CONTEXT, fs_get_win_error(), path);
	}
#else
	if (mkdir(path, S_IRWXU | S_IRWXG | S_IRWXO) == -1)
		xt_throw_ferrno(XT_CONTEXT, errno, path);

	try_(a) {
		fs_set_stats(self, path);
	}
	catch_(a) {
		xt_fs_rmdir(NULL, name);
		throw_();
	}
	cont_(a);
#endif
}

void xt_fs_mkpath(XTThreadPtr self, char *path)
{
	char *ptr;

	if (xt_fs_exists(path))
		return;

	if (!(ptr = (char *) xt_last_directory_of_path((c_char *) path)))
		return;
	if (ptr == path)
		return;
	ptr--;
	if (XT_IS_DIR_CHAR(*ptr)) {
		*ptr = 0;
		xt_fs_mkpath(self, path);
		*ptr = XT_DIR_CHAR;
		xt_fs_mkdir(self, path);
	}
}

xtBool xt_fs_rmdir(XTThreadPtr self, char *name)
{
	char path[PATH_MAX];

	xt_strcpy(PATH_MAX, path, name);
	xt_remove_dir_char(path);

#ifdef XT_WIN
	if (!RemoveDirectory(path)) {
		int err = fs_get_win_error();

		if (!XT_FILE_NOT_FOUND(err)) {
			xt_throw_ferrno(XT_CONTEXT, err, path);
			return FAILED;
		}
	}
#else
	if (rmdir(path) == -1) {
		int err = errno;

		if (err != ENOENT) {
			xt_throw_ferrno(XT_CONTEXT, err, path);
			return FAILED;
		}
	}
#endif
	return OK;
}

/* ----------------------------------------------------------------------
 * Open & Close operations
 */

xtPublic XTFilePtr xt_fs_get_file(XTThreadPtr self, char *file_name)
{
	XTFilePtr	file_ptr, *file_pptr;

	xt_sl_lock(self, fs_globals.fsg_open_files);
	pushr_(xt_sl_unlock, fs_globals.fsg_open_files);

	if ((file_pptr = (XTFilePtr *) xt_sl_find(self, fs_globals.fsg_open_files, file_name)))
		file_ptr = *file_pptr;
	else {
		file_ptr = fs_new_file(self, file_name);
		xt_sl_insert(self, fs_globals.fsg_open_files, file_name, &file_ptr);
	}
	file_ptr->fil_ref_count++;
	freer_(); // xt_sl_unlock(fs_globals.fsg_open_files)
	return file_ptr;
}

xtPublic void xt_fs_release_file(XTThreadPtr self, XTFilePtr file_ptr)
{
	xt_sl_lock(self, fs_globals.fsg_open_files);
	pushr_(xt_sl_unlock, fs_globals.fsg_open_files);

	file_ptr->fil_ref_count--;
	if (!file_ptr->fil_ref_count) {
		xt_sl_delete(self, fs_globals.fsg_open_files, file_ptr->fil_path);
	}

	freer_(); // xt_ht_unlock(fs_globals.fsg_lock)
}

static xtBool fs_open_file(XTThreadPtr self, XTOpenFilePtr of, int mode)
{
	int retried = FALSE;

	retry:
#ifdef XT_WIN
	SECURITY_ATTRIBUTES	sa = { sizeof(SECURITY_ATTRIBUTES), 0, 0 };
	DWORD				flags;

	if (mode & XT_FS_EXCLUSIVE)
		flags = CREATE_NEW;
	else if (mode & XT_FS_CREATE)
		flags = OPEN_ALWAYS;
	else
		flags = OPEN_EXISTING;

	of->of_filedes = CreateFile(
		of->of_file->fil_path,
		mode & XT_FS_READONLY ? GENERIC_READ : (GENERIC_READ | GENERIC_WRITE),
		FILE_SHARE_READ | FILE_SHARE_WRITE,
		&sa,
		flags,
		FILE_FLAG_RANDOM_ACCESS,
		NULL);
	if (of->of_filedes == INVALID_HANDLE_VALUE) {
		int err = fs_get_win_error();

		if (!(mode & XT_FS_MISSING_OK) || !XT_FILE_NOT_FOUND(err)) {
			if (!retried && (mode & XT_FS_MAKE_PATH) && XT_FILE_NOT_FOUND(err)) {
				char path[PATH_MAX];

				xt_strcpy(PATH_MAX, path, of->of_file->fil_path);
				xt_remove_last_name_of_path(path);
				xt_fs_mkpath(self, path);
				retried = TRUE;
				goto retry;
			}

			xt_throw_ferrno(XT_CONTEXT, err, of->of_file->fil_path);
		}

		/* File is missing, but don't throw an error. */
		return FAILED;
	}
	//printf("open %d %s\n", of->of_filedes, of->of_file->fil_path);
	return OK;
#else
	int flags = 0;

	if (mode & XT_FS_READONLY)
		flags = O_RDONLY;
	else
		flags = O_RDWR;
	if (mode & XT_FS_CREATE)
		flags |= O_CREAT;
	if (mode & XT_FS_EXCLUSIVE)
		flags |= O_EXCL;

	of->of_filedes = open(of->of_file->fil_path, flags, XT_MASK);
	if (of->of_filedes == -1) {
		int err = errno;

		if (!(mode & XT_FS_MISSING_OK) || err != ENOENT) {
			if (!retried && (mode & XT_FS_MAKE_PATH) && err == ENOENT) {
				char path[PATH_MAX];

				xt_strcpy(PATH_MAX, path, of->of_file->fil_path);
				xt_remove_last_name_of_path(path);
				xt_fs_mkpath(self, path);
				retried = TRUE;
				goto retry;
			}

			xt_throw_ferrno(XT_CONTEXT, err, of->of_file->fil_path);
		}

		/* File is missing, but don't throw an error. */
		return FAILED;
	}
	///printf("open %d %s\n", of->of_filedes, of->of_file->fil_path);
	return OK;
#endif
}

xtPublic XTOpenFilePtr xt_open_file(XTThreadPtr self, char *file, int mode)
{
	XTOpenFilePtr of;

	pushsr_(of, xt_close_file, (XTOpenFilePtr) xt_calloc(self, sizeof(XTOpenFileRec)));
	of->of_filedes = XT_INVALID_HANDLE;
	of->of_file = xt_fs_get_file(self, file);
	of->of_id = of->of_file->fil_id;

#ifdef XT_WIN
	if (!fs_open_file(self, of, mode)) {
		xt_close_file(self, of);
		of = NULL;
	}
#else
	if (of->of_file->fil_filedes == -1) {
		if (fs_open_file(self, of, mode)) {
			/* Comment this out to use file handle per thread: */
			xt_sl_lock(self, fs_globals.fsg_open_files);
			pushr_(xt_sl_unlock, fs_globals.fsg_open_files);
			if (of->of_file->fil_filedes == -1)
				of->of_file->fil_filedes = of->of_filedes;
			else {
				close(of->of_filedes);
				of->of_filedes = of->of_file->fil_filedes;
			}
			freer_(); // xt_ht_unlock(fs_globals.fsg_open_files)
		}
		else {
			xt_close_file(self, of);
			of = NULL;
		}
	}
	else
		of->of_filedes = of->of_file->fil_filedes;
#endif

	popr_(); // Discard xt_close_file(of)
	return of;
}

xtPublic XTOpenFilePtr xt_open_file_ns(char *file, int mode)
{
	XTThreadPtr		self = xt_get_self();
	XTOpenFilePtr	of;

	try_(a) {
		of = xt_open_file(self, file, mode);
	}
	catch_(a) {
		of = NULL;
	}
	cont_(a);
	return of;
}

xtPublic xtBool xt_open_file_ns(XTOpenFilePtr *fh, char *file, int mode)
{
	XTThreadPtr		self = xt_get_self();
	xtBool			ok = TRUE;

	try_(a) {
		*fh = xt_open_file(self, file, mode);
	}
	catch_(a) {
		ok = FALSE;
	}
	cont_(a);
	return ok;
}

xtPublic void xt_close_file(XTThreadPtr self, XTOpenFilePtr of)
{
	if (of->of_filedes != XT_INVALID_HANDLE) {
#ifdef XT_WIN
		CloseHandle(of->of_filedes);
		//printf("close %d %s\n", of->of_filedes, of->of_file->fil_path);
#else
		if (!of->of_file || of->of_filedes != of->of_file->fil_filedes) {
			close(of->of_filedes);
			//printf("close %d %s\n", of->of_filedes, of->of_file->fil_path);
		}
#endif

		of->of_filedes = XT_INVALID_HANDLE;
	}

	if (of->of_file) {
		xt_fs_release_file(self, of->of_file);
		of->of_file = NULL;
	}
	xt_free(self, of);
}

xtPublic xtBool xt_close_file_ns(XTOpenFilePtr of)
{
	XTThreadPtr self = xt_get_self();
	xtBool		failed = FALSE;

	try_(a) {
		xt_close_file(self, of);
	}
	catch_(a) {
		failed = TRUE;
	}
	cont_(a);
	return failed;
}

/* ----------------------------------------------------------------------
 * I/O operations
 */

xtPublic off_t xt_seek_eof_file(XTThreadPtr self, XTOpenFilePtr of)
{
#ifdef XT_WIN
	DWORD			result;
	LARGE_INTEGER	lpFileSize;

	result = SetFilePointer(of->of_filedes, 0, NULL, FILE_END);
	if (result == 0xFFFFFFFF) {
		xt_throw_ferrno(XT_CONTEXT, fs_get_win_error(), xt_file_path(of));
		return (off_t) -1;
	}

	if (!GetFileSizeEx(of->of_filedes, &lpFileSize)) {
		xt_throw_ferrno(XT_CONTEXT, fs_get_win_error(), xt_file_path(of));
		return (off_t) -1;
	}

	return lpFileSize.QuadPart;
#else
	off_t off;

	off = lseek(of->of_filedes, 0, SEEK_END);
	if (off == -1) {
		xt_throw_ferrno(XT_CONTEXT, errno, xt_file_path(of));
		return -1;
	}

     return off;
#endif
}

xtPublic xtBool xt_set_eof_file(XTThreadPtr self, XTOpenFilePtr of, off_t offset)
{
#ifdef XT_WIN
	LARGE_INTEGER liDistanceToMove;
	
	liDistanceToMove.QuadPart = offset;
	if (!SetFilePointerEx(of->of_filedes, liDistanceToMove, NULL, FILE_BEGIN)) {
		xt_throw_ferrno(XT_CONTEXT, fs_get_win_error(), xt_file_path(of));
		return FAILED;
	}

	if (!SetEndOfFile(of->of_filedes)) {
		xt_throw_ferrno(XT_CONTEXT, fs_get_win_error(), xt_file_path(of));
		return FAILED;
	}
#else
	if (ftruncate(of->of_filedes, offset) == -1) {
		xt_throw_ferrno(XT_CONTEXT, errno, xt_file_path(of));
		return FAILED;
	}
#endif
	return OK;
}

xtPublic xtBool xt_pwrite_file(XTOpenFilePtr of, off_t offset, size_t size, void *data)
{
#ifdef DEBUG_PRINT_IO
	printf("PBXT WRITE %s offs=%d size=%d\n", of->of_file->fil_path, (int) offset, (int) size);
#endif
#ifdef DEBUG_TRACE_IO
	xt_trace("/* %s */ pbxt_file_writ(\"%s\", %lu, %lu);\n", xt_trace_clock_diff(NULL), of->of_file->fil_path, (u_long) offset, (u_long) size);
#endif
#ifdef XT_WIN
	LARGE_INTEGER	liDistanceToMove;
	DWORD			result;
	
	liDistanceToMove.QuadPart = offset;
	if (!SetFilePointerEx(of->of_filedes, liDistanceToMove, NULL, FILE_BEGIN))
		return xt_register_ferrno(XT_REG_CONTEXT, fs_get_win_error(), xt_file_path(of));

	if (!WriteFile(of->of_filedes, data, size, &result, NULL))
		return xt_register_ferrno(XT_REG_CONTEXT, fs_get_win_error(), xt_file_path(of));

	if (result != size)
		return xt_register_ferrno(XT_REG_CONTEXT, ERROR_HANDLE_EOF, xt_file_path(of));

	return OK;
#else
	ssize_t write_size;

	write_size = pwrite(of->of_filedes, data, size, offset);
	if (write_size == -1)
		return xt_register_ferrno(XT_REG_CONTEXT, errno, xt_file_path(of));

	if ((size_t) write_size != size)
		return xt_register_ferrno(XT_REG_CONTEXT, ESPIPE, xt_file_path(of));

	return OK;
#endif
}

xtPublic xtBool xt_flush_file(XTOpenFilePtr of)
{
#ifdef DEBUG_PRINT_IO
	printf("PBXT FLUSH %s\n", of->of_file->fil_path);
#endif
#ifdef DEBUG_TRACE_IO
	xt_trace("/* %s */ pbxt_file_sync(\"%s\");\n", xt_trace_clock_diff(NULL), of->of_file->fil_path);
#endif
#ifdef XT_WIN
	if (!FlushFileBuffers(of->of_filedes))
		return xt_register_ferrno(XT_REG_CONTEXT, fs_get_win_error(), xt_file_path(of));
#else
	if (fsync(of->of_filedes) == -1)
		return xt_register_ferrno(XT_REG_CONTEXT, errno, xt_file_path(of));
#endif
	return OK;
}

xtBool xt_pread_file(XTOpenFilePtr of, off_t offset, size_t size, size_t min_size, void *data, size_t *red_size)
{
#ifdef DEBUG_PRINT_IO
	printf("PBXT READ %s offset=%d size=%d\n", of->of_file->fil_path, (int) offset, (int) size);
#endif
#ifdef DEBUG_TRACE_IO
	xt_trace("/* %s */ pbxt_file_read(\"%s\", %lu, %lu);\n", xt_trace_clock_diff(NULL), of->of_file->fil_path, (u_long) offset, (u_long) size);
#endif
#ifdef XT_WIN
	LARGE_INTEGER	liDistanceToMove;
	DWORD			result;

	liDistanceToMove.QuadPart = offset;
	if (!SetFilePointerEx(of->of_filedes, liDistanceToMove, NULL, FILE_BEGIN))
		return xt_register_ferrno(XT_REG_CONTEXT, fs_get_win_error(), xt_file_path(of));

	if (!ReadFile(of->of_filedes, data, size, &result, NULL))
		return xt_register_ferrno(XT_REG_CONTEXT, fs_get_win_error(), xt_file_path(of));

	if ((size_t) result < min_size)
		return xt_register_ferrno(XT_REG_CONTEXT, ERROR_HANDLE_EOF, xt_file_path(of));

	if (red_size)
		*red_size = (size_t) result;
	return OK;
#else
	ssize_t read_size;

	read_size = pread(of->of_filedes, data, size, offset);
	if (read_size == -1)
		return xt_register_ferrno(XT_REG_CONTEXT, errno, xt_file_path(of));

	/* Throw an error if read less than the minimum: */
	if ((size_t) read_size < min_size) {
//printf("PMC PBXT <-- offset:%llu, count:%lu \n", (u_llong) offset, (u_long) size);
		return xt_register_ferrno(XT_REG_CONTEXT, ESPIPE, xt_file_path(of));
	}

	if (red_size)
		*red_size = (size_t) read_size;
	return OK;
#endif
}

/* ----------------------------------------------------------------------
 * Directory operations
 */

/*
 * The filter may contain one '*' as wildcard.
 */
XTOpenDirPtr xt_dir_open(XTThreadPtr self, c_char *path, c_char *filter)
{
	XTOpenDirPtr	od;
#ifdef XT_WIN
	size_t			len;
#endif

	if (!(od = (XTOpenDirPtr) xt_calloc(self, sizeof(XTOpenDirRec))))
		return NULL;

#ifdef XT_WIN
	// path = path\(filter | *)
	len = strlen(path) + 1 + (filter ? strlen(filter) : 1) + 1;
	if (!(od->od_path = xt_malloc(self, len)))
		goto failed;

	strcpy(od->od_path, path);
	xt_add_dir_char(len, od->od_path);
	if (filter)
		strcat(od->od_path, filter);
	else
		strcat(od->od_path, "*");

	od->od_handle = INVALID_HANDLE_VALUE;
#else
	if (!(od->od_path = xt_dup_string(NULL, path)))
		goto failed;

	if (filter) {
		if (!(od->od_filter = xt_dup_string(NULL, filter)))
			goto failed;
	}

	od->od_dir = opendir(path);
	if (!od->od_dir) {
		xt_dir_close(self, od);
		xt_throw_ferrno(XT_CONTEXT, errno, path);
		return FAILED;
	}
#endif

	return od;

	failed:
	xt_dir_close(self, od);
	throw_();
	return NULL;
}

void xt_dir_close(XTThreadPtr self, XTOpenDirPtr od)
{
	if (od) {
#ifdef XT_WIN
		if (od->od_handle != INVALID_HANDLE_VALUE) {
			FindClose(od->od_handle);
			od->od_handle = INVALID_HANDLE_VALUE;
		}
#else
		if (od->od_dir) {
			closedir(od->od_dir);
			od->od_dir = NULL;
		}
		if (od->od_filter) {
			xt_free(self, od->od_filter);
			od->od_filter = NULL;
		}
#endif
		if (od->od_path) {
			xt_free(self, od->od_path);
			od->od_path = NULL;
		}
		xt_free(self, od);
	}
}

#ifdef XT_WIN
xtBool xt_dir_next(XTThreadPtr self, XTOpenDirPtr od)
{
	int err = 0;

	if (od->od_handle == INVALID_HANDLE_VALUE) {
		od->od_handle = FindFirstFile(od->od_path, &od->od_data);
		if (od->od_handle == INVALID_HANDLE_VALUE)
			err = fs_get_win_error();
	}
	else {
		if (!FindNextFile(od->od_handle, &od->od_data))
			err = fs_get_win_error();
	}

	if (err) {
		if (err != ERROR_NO_MORE_FILES) {
			if (err == ERROR_FILE_NOT_FOUND) {
				char path[PATH_MAX];

				xt_strcpy(PATH_MAX, path, od->od_path);
				xt_remove_last_name_of_path(path);
				if (!xt_fs_exists(path))
					xt_throw_ferrno(XT_CONTEXT, err, path);
			}
			else
				xt_throw_ferrno(XT_CONTEXT, err, od->od_path);
		}
		return FAILED;
	}

	return OK;
}
#else
static xtBool fs_match_filter(c_char *name, c_char *filter)
{
	while (*name && *filter) {
		if (*filter == '*') {
			if (filter[1] == *name)
				filter++;
			else
				name++;
		}
		else {
			if (*name != *filter)
				return FALSE;
			name++;
			filter++;
		}
	}
	if (!*name) {
		if (!*filter || (*filter == '*' && !filter[1]))
			return TRUE;
	}
	return FALSE;
}

xtBool xt_dir_next(XTThreadPtr self, XTOpenDirPtr od)
{
	int				err;
	struct dirent	*result;

	for (;;) {
		err = readdir_r(od->od_dir, &od->od_entry, &result);
		if (err) {
			xt_throw_ferrno(XT_CONTEXT, err, od->od_path);
			return FAILED;
		}
		if (!result)
			break;
		/* Filter out '.' and '..': */
		if (od->od_entry.d_name[0] == '.') {
			if (od->od_entry.d_name[1] == '.') {
				if (od->od_entry.d_name[2] == '\0')
					continue;
			}
			else {
				if (od->od_entry.d_name[1] == '\0')
					continue;
			}
		}
		if (!od->od_filter)
			break;
		if (fs_match_filter(od->od_entry.d_name, od->od_filter))
			break;
	}
	return result ? TRUE : FALSE;
}
#endif

char *xt_dir_name(XTThreadPtr self, XTOpenDirPtr od)
{
#ifdef XT_WIN
	return od->od_data.cFileName;
#else
	return od->od_entry.d_name;
#endif
}

xtBool xt_dir_is_file(XTThreadPtr self, XTOpenDirPtr od)
{
#ifdef XT_WIN
	if (od->od_data.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)
		return FALSE;
#else
	if (od->od_entry.d_type & DT_DIR)
		return FALSE;
#endif
	return TRUE;
}

off_t xt_dir_file_size(XTThreadPtr self, XTOpenDirPtr od)
{
#ifdef XT_WIN
	return (off_t) od->od_data.nFileSizeLow | (((off_t) od->od_data.nFileSizeHigh) << 32);
#else
	char	path[PATH_MAX];
	off_t	size;

	xt_strcpy(PATH_MAX, path, od->od_path);
	xt_add_dir_char(PATH_MAX, path);
	xt_strcat(PATH_MAX, path, od->od_entry.d_name);
	if (!xt_fs_stat(self, path, &size, NULL))
		return -1;
	return size;
#endif
}
