/* Copyright (c) 2005 PrimeBase Technologies GmbH
 *
 * PrimeBase XT
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * 2005-09-30	Paul McCullagh
 *
 * H&G2JCtL
 */

#include "xt_config.h"

#include <string.h>
#include <stdio.h>
#include <stddef.h>
#ifndef XT_WIN
#include <strings.h>
#endif

#include <m_ctype.h>
#include <my_base.h>

#include "memory_xt.h"
#include "index_xt.h"
#include "heap_xt.h"
#include "database_xt.h"
#include "strutil_xt.h"
#include "cache_xt.h"
#include "myxt_xt.h"
#include "trace_xt.h"

#ifdef DEBUG
#define MAX_SEARCH_DEPTH			32
//#define CHECK_AND_PRINT
#else
#define MAX_SEARCH_DEPTH			100
#endif

typedef struct IdxStackItem {
	XTIdxItemRec			i_pos;
	off_t					i_branch;
} IdxStackItemRec, *IdxStackItemPtr;

typedef struct IdxBranchStack {
	int						s_top;
	IdxStackItemRec			s_elements[MAX_SEARCH_DEPTH];
} IdxBranchStackRec, *IdxBranchStackPtr;

#ifdef DEBUG
static void idx_check_on_key(XTOpenTablePtr ot);
static void idx_print_index(XTOpenTablePtr ot, XTIndexPtr ind, xtBool with_lock);
static void idx_check_index(XTOpenTablePtr ot, XTIndexPtr ind);
#endif

static xtBool idx_insert_node(XTOpenTablePtr ot, XTIndexPtr ind, IdxBranchStackPtr stack, XTIdxKeyValuePtr key_value, off_t branch);

/*
 * -----------------------------------------------------------------------
 * Compare XT disk cache to MySQL disk cache:
 */

#ifdef XT_USE_MYSYS
#include "my_global.h"
#include "my_sys.h"
#include "keycache.h"

#undef	pthread_rwlock_rdlock
#undef	pthread_rwlock_wrlock
#undef	pthread_rwlock_unlock
#undef	pthread_mutex_lock
#undef	pthread_mutex_unlock
#undef	pthread_cond_wait
#undef	pthread_cond_broadcast
#undef	xt_mutex_type

#define xtPublic

KEY_CACHE my_cache;

static xtBool my_set_mem(xtWord1 *buf, size_t siz)
{
	memset(buf, 0, siz);
	return OK;
}

#define xt_ind_write(x, y, z, a)	(!key_cache_write(&my_cache, (x)->of_filedes, y, 3, (byte *) (a), XT_INDEX_PAGE_SIZE, 1024, 1))

#define xt_ind_read(x, y, z, a)	(key_cache_read(&my_cache, (x)->of_filedes, y, 3, (byte *) (a), z, 1024, 0) ? 1 : my_set_mem((xtWord1 *) (a), z))
#endif

/*
 * -----------------------------------------------------------------------
 * Stack functions
 */

static void idx_newstack(IdxBranchStackPtr stack)
{
	stack->s_top = 0;
}

static xtBool idx_push(IdxBranchStackPtr stack, off_t n, XTIdxItemPtr pos)
{
	if (stack->s_top == MAX_SEARCH_DEPTH) {
		xt_register_error(XT_REG_CONTEXT, XT_ERR_STACK_OVERFLOW, 0, "Core B-tree too deep");
		return FAILED;
	}
	stack->s_elements[stack->s_top].i_branch = n;
	if (pos)
		stack->s_elements[stack->s_top].i_pos = *pos;
	stack->s_top++;
	return OK;
}

static IdxStackItemPtr idx_pop(IdxBranchStackPtr stack)
{
	if (stack->s_top == 0)
		return NULL;
	stack->s_top--;
	return &stack->s_elements[stack->s_top];
}

static IdxStackItemPtr idx_top(IdxBranchStackPtr stack)
{
	if (stack->s_top == 0)
		return NULL;
	return &stack->s_elements[stack->s_top-1];
}

/*
 * -----------------------------------------------------------------------
 * Allocation of nodes
 */

typedef struct IdxFreeBlock {
	XTDiskValue1			if_status_1;
	XTDiskValue1			if_unused1_1;
	XTDiskValue2			if_unused2_2;
	XTDiskValue4			if_unused3_4;
	XTDiskValue8			if_next_block_8;
} IdxFreeBlockRec, *IdxFreeBlockPtr;

static xtBool idx_new_branch(XTOpenTablePtr ot, off_t *address)
{
	register XTTableHPtr	tab = ot->ot_table;
	off_t					wrote_pos;

	xt_lock_mutex_ns(&tab->tab_ind_lock);
	if ((wrote_pos = tab->tab_ind_free)) {
		/* Use the block on the free list: */
		IdxFreeBlockRec	free_block;

		if (!xt_ind_read(ot->ot_ind_file, wrote_pos, sizeof(IdxFreeBlockRec), (xtWord1 *) &free_block))
			goto failed;
		tab->tab_ind_free = XT_GET_DISK_8(free_block.if_next_block_8);
		tab->tab_ind_head_dirty = TRUE;
		xt_unlock_mutex_ns(&tab->tab_ind_lock);

		*address = wrote_pos;
		return OK;
	}

	/* PMC - Dont allow overflow! */
	if (tab->tab_ind_eof >= ( ((xtWord8) 1) << (32 + XT_INDEX_PAGE_SHIFTS) ) ) {
		xt_register_ixterr(XT_REG_CONTEXT, XT_ERR_INDEX_FILE_TO_LARGE, xt_file_path(ot->ot_ind_file));
		goto failed;
	}
	*address = tab->tab_ind_eof;
	tab->tab_ind_eof += XT_INDEX_PAGE_SIZE;
	tab->tab_ind_head_dirty = TRUE;

	xt_unlock_mutex_ns(&tab->tab_ind_lock);
	return OK;

	failed:
	xt_unlock_mutex_ns(&tab->tab_ind_lock);
	return FAILED;
}

static xtBool idx_free_branch(XTOpenTablePtr ot, off_t address)
{
	register XTTableHPtr	tab = ot->ot_table;
	IdxFreeBlockRec			free_block = { { 0 }, { 0 }, { 0 }, { 0 }, { 0 } };

	xt_lock_mutex_ns(&tab->tab_ind_lock);

	/* Use the block on the free list: */
	XT_SET_DISK_8(free_block.if_next_block_8, tab->tab_ind_free);
	if (!xt_ind_write(ot->ot_ind_file, address, sizeof(IdxFreeBlockRec), (xtWord1 *) &free_block))
		goto failed;
	tab->tab_ind_free = address;
	tab->tab_ind_head_dirty = TRUE;
	xt_unlock_mutex_ns(&tab->tab_ind_lock);
	return OK;
	
	failed:
	xt_unlock_mutex_ns(&tab->tab_ind_lock);
	return FAILED;
}

/*
 * -----------------------------------------------------------------------
 * Simple compare functions
 */

xtPublic int xt_compare_2_int4(XTIndexPtr ind, uint key_length, xtWord1 *key_value, xtWord1 *b_value)
{
	int r;

	ASSERT_NS(key_length == 4 || key_length == 8);
	r = (xtInt4) XT_GET_DISK_4(key_value) - (xtInt4) XT_GET_DISK_4(b_value);
	if (r == 0 && key_length > 4) {
		key_value += 4;
		b_value += 4;
		r = (xtInt4) XT_GET_DISK_4(key_value) - (xtInt4) XT_GET_DISK_4(b_value);
	}
	return r;
}

xtPublic int xt_compare_3_int4(XTIndexPtr ind, uint key_length, xtWord1 *key_value, xtWord1 *b_value)
{
	int r;

	ASSERT_NS(key_length == 4 || key_length == 8 || key_length == 12);
	r = (xtInt4) XT_GET_DISK_4(key_value) - (xtInt4) XT_GET_DISK_4(b_value);
	if (r == 0 && key_length > 4) {
		key_value += 4;
		b_value += 4;
		r = (xtInt4) XT_GET_DISK_4(key_value) - (xtInt4) XT_GET_DISK_4(b_value);
		if (r == 0 && key_length > 8) {
			key_value += 4;
			b_value += 4;
			r = (xtInt4) XT_GET_DISK_4(key_value) - (xtInt4) XT_GET_DISK_4(b_value);
		}
	}
	return r;
}

/*
 * -----------------------------------------------------------------------
 * Tree branch sanning (searching nodes and leaves)
 */

xtPublic void xt_scan_branch_single(XTIndexPtr ind, XTIdxBranchDPtr branch, register XTIdxKeyValuePtr value, register XTIdxResultRec *result)
{
	u_int				branch_size;
	u_int				node_ref_size;
	u_int				full_item_size;
	int					search_flags;
	register xtWord1	*base;
	register u_int		i;
	register xtWord1	*bitem;

	branch_size = XT_GET_DISK_2(branch->tb_size_2);
	node_ref_size = XT_IS_NODE(branch_size) ? XT_NODE_REF_SIZE : 0;

	result->sr_found = FALSE;
	result->sr_duplicate = FALSE;
	result->sr_item.i_total_size = XT_GET_BRANCH_DATA_SIZE(branch_size);
	ASSERT_NS((int) result->sr_item.i_total_size >= 0 && result->sr_item.i_total_size <= XT_INDEX_PAGE_SIZE-2);

	result->sr_item.i_item_size = ind->mi_key_size + XT_RECORD_REF_SIZE;
	full_item_size = result->sr_item.i_item_size + node_ref_size;
	result->sr_item.i_node_ref_size = node_ref_size;

	search_flags = value->sv_flags;
	base = branch->tb_data + node_ref_size;
	if (search_flags & XT_SEARCH_FIRST_FLAG)
		i = 0;
	else if (search_flags & XT_SEARCH_AFTER_LAST_FLAG)
		i = (result->sr_item.i_total_size - node_ref_size) / full_item_size;
	else {
		register u_int		guess;
		register u_int		count;
		register xtInt4		r;
		xtRecordID			key_record;

		key_record = value->sv_rec_id;
		count = (result->sr_item.i_total_size - node_ref_size) / full_item_size;

		ASSERT_NS(ind);
		i = 0;
		while (i < count) {
			guess = (i + count - 1) >> 1;

			bitem = base + guess * full_item_size;

			switch (ind->mi_single_type) {
				case HA_KEYTYPE_LONG_INT:
					r = XT_GET_DISK_4(value->sv_key) - XT_GET_DISK_4(bitem);
					break;
				case HA_KEYTYPE_ULONG_INT: {
					register xtWord4 a, b;
					
					a = XT_GET_DISK_4(value->sv_key);
					b = XT_GET_DISK_4(bitem);
					r = (a < b) ? -1 : (a == b ? 0 : 1);
					break;
				}
				default:
					/* Should not happen: */
					r = 1;
					break;
			}
			if (r == 0) {
				if (search_flags & XT_SEARCH_WHOLE_KEY) {
					xtRecordID	item_record;
					xtRowID		row_id;
					
					xt_get_record_ref(bitem + ind->mi_key_size, &item_record, &row_id);

					/* This should not happen because we should never
					 * try to insert the same record twice into the 
					 * index!
					 */
					result->sr_duplicate = TRUE;
					if (key_record == item_record) {
						result->sr_found = TRUE;
						result->sr_rec_id = item_record;
						result->sr_row_id = row_id;
						result->sr_branch = XT_GET_NODE_REF(bitem - node_ref_size);
						result->sr_item.i_item_offset = node_ref_size + guess * full_item_size;
						return;
					}
					if (key_record < item_record)
						r = -1;
					else
						r = 1;
				}
				else {
					result->sr_found = TRUE;
					/* -1 causes a search to the beginning of the duplicate list of keys.
					 * 1 causes a search to just after the key.
				 	*/
					if (search_flags & XT_SEARCH_AFTER_KEY)
						r = 1;
					else
						r = -1;
				}
			}

			if (r < 0)
				count = guess;
			else
				i = guess + 1;
		}
	}

	bitem = base + i * full_item_size;
	xt_get_res_record_ref(bitem + ind->mi_key_size, result);
	result->sr_branch = XT_GET_NODE_REF(bitem - node_ref_size);			/* Only valid if this is a node. */
	result->sr_item.i_item_offset = node_ref_size + i * full_item_size;
}

/*
 * We use a special binary search here. It basically assumes that the values
 * in the index are not unique.
 *
 * Even if they are unique, when we search for part of a key, then it is
 * effectively the case.
 *
 * So in the situation where we find duplicates in the index we usually
 * want to position ourselves at the beginning of the duplicate list.
 *
 * Alternatively a search can find the position just after a given key.
 *
 * To achieve this we make the following modifications:
 * - The result of the comparison is always returns 1 or -1. We only stop
 *   the search early in the case an exact match when inserting (but this
 *   should not happen anyway).
 * - The search never actually fails, but sets 'found' to TRUE if it
 *   sees the search key in the index.
 *
 * If the search value exists in the index we know that
 * this method will take us to the first occurrence of the key in the
 * index (in the case of -1) or to the first value after the
 * the search key in the case of 1.
 */
xtPublic void xt_scan_branch_fix(XTIndexPtr ind, XTIdxBranchDPtr branch, register XTIdxKeyValuePtr value, register XTIdxResultRec *result)
{
	u_int				branch_size;
	u_int				node_ref_size;
	u_int				full_item_size;
	int					search_flags;
	xtWord1				*base;
	register u_int		i;
	xtWord1				*bitem;

	branch_size = XT_GET_DISK_2(branch->tb_size_2);
	node_ref_size = XT_IS_NODE(branch_size) ? XT_NODE_REF_SIZE : 0;

	result->sr_found = FALSE;
	result->sr_duplicate = FALSE;
	result->sr_item.i_total_size = XT_GET_BRANCH_DATA_SIZE(branch_size);
	ASSERT_NS((int) result->sr_item.i_total_size >= 0 && result->sr_item.i_total_size <= XT_INDEX_PAGE_SIZE-2);

	result->sr_item.i_item_size = ind->mi_key_size + XT_RECORD_REF_SIZE;
	full_item_size = result->sr_item.i_item_size + node_ref_size;
	result->sr_item.i_node_ref_size = node_ref_size;

	search_flags = value->sv_flags;
	base = branch->tb_data + node_ref_size;
	if (search_flags & XT_SEARCH_FIRST_FLAG)
		i = 0;
	else if (search_flags & XT_SEARCH_AFTER_LAST_FLAG)
		i = (result->sr_item.i_total_size - node_ref_size) / full_item_size;
	else {
		register u_int		guess;
		register u_int		count;
		xtRecordID			key_record;
		int					r;

		key_record = value->sv_rec_id;
		count = (result->sr_item.i_total_size - node_ref_size) / full_item_size;

		ASSERT_NS(ind);
		i = 0;
		while (i < count) {
			guess = (i + count - 1) >> 1;

			bitem = base + guess * full_item_size;

			r = myxt_compare_key(ind, search_flags, value->sv_length, value->sv_key, bitem);

			if (r == 0) {
				if (search_flags & XT_SEARCH_WHOLE_KEY) {
					xtRecordID	item_record;
					xtRowID		row_id;

					xt_get_record_ref(bitem + ind->mi_key_size, &item_record, &row_id);

					/* This should not happen because we should never
					 * try to insert the same record twice into the 
					 * index!
					 */
					result->sr_duplicate = TRUE;
					if (key_record == item_record) {
						result->sr_found = TRUE;
						result->sr_rec_id = item_record;
						result->sr_row_id = row_id;
						result->sr_branch = XT_GET_NODE_REF(bitem - node_ref_size);
						result->sr_item.i_item_offset = node_ref_size + guess * full_item_size;
						return;
					}
					if (key_record < item_record)
						r = -1;
					else
						r = 1;
				}
				else {
					result->sr_found = TRUE;
					/* -1 causes a search to the beginning of the duplicate list of keys.
					 * 1 causes a search to just after the key.
				 	*/
					if (search_flags & XT_SEARCH_AFTER_KEY)
						r = 1;
					else
						r = -1;
				}
			}

			if (r < 0)
				count = guess;
			else
				i = guess + 1;
		}
	}

	bitem = base + i * full_item_size;
	xt_get_res_record_ref(bitem + ind->mi_key_size, result);
	result->sr_branch = XT_GET_NODE_REF(bitem - node_ref_size);			/* Only valid if this is a node. */
	result->sr_item.i_item_offset = node_ref_size + i * full_item_size;
}

xtPublic void xt_scan_branch_fix_simple(XTIndexPtr ind, XTIdxBranchDPtr branch, register XTIdxKeyValuePtr value, register XTIdxResultRec *result)
{
	u_int				branch_size;
	u_int				node_ref_size;
	u_int				full_item_size;
	int					search_flags;
	xtWord1				*base;
	register u_int		i;
	xtWord1				*bitem;

	branch_size = XT_GET_DISK_2(branch->tb_size_2);
	node_ref_size = XT_IS_NODE(branch_size) ? XT_NODE_REF_SIZE : 0;

	result->sr_found = FALSE;
	result->sr_duplicate = FALSE;
	result->sr_item.i_total_size = XT_GET_BRANCH_DATA_SIZE(branch_size);
	ASSERT_NS((int) result->sr_item.i_total_size >= 0 && result->sr_item.i_total_size <= XT_INDEX_PAGE_SIZE-2);

	result->sr_item.i_item_size = ind->mi_key_size + XT_RECORD_REF_SIZE;
	full_item_size = result->sr_item.i_item_size + node_ref_size;
	result->sr_item.i_node_ref_size = node_ref_size;

	search_flags = value->sv_flags;
	base = branch->tb_data + node_ref_size;
	if (search_flags & XT_SEARCH_FIRST_FLAG)
		i = 0;
	else if (search_flags & XT_SEARCH_AFTER_LAST_FLAG)
		i = (result->sr_item.i_total_size - node_ref_size) / full_item_size;
	else {
		register u_int		guess;
		register u_int		count;
		xtRecordID			key_record;
		int					r;

		key_record = value->sv_rec_id;
		count = (result->sr_item.i_total_size - node_ref_size) / full_item_size;

		ASSERT_NS(ind);
		i = 0;
		while (i < count) {
			guess = (i + count - 1) >> 1;

			bitem = base + guess * full_item_size;

			r = ind->mi_simple_comp_key(ind, value->sv_length, value->sv_key, bitem);

			if (r == 0) {
				if (search_flags & XT_SEARCH_WHOLE_KEY) {
					xtRecordID	item_record;
					xtRowID		row_id;

					xt_get_record_ref(bitem + ind->mi_key_size, &item_record, &row_id);

					/* This should not happen because we should never
					 * try to insert the same record twice into the 
					 * index!
					 */
					result->sr_duplicate = TRUE;
					if (key_record == item_record) {
						result->sr_found = TRUE;
						result->sr_rec_id = item_record;
						result->sr_row_id = row_id;
						result->sr_branch = XT_GET_NODE_REF(bitem - node_ref_size);
						result->sr_item.i_item_offset = node_ref_size + guess * full_item_size;
						return;
					}
					if (key_record < item_record)
						r = -1;
					else
						r = 1;
				}
				else {
					result->sr_found = TRUE;
					/* -1 causes a search to the beginning of the duplicate list of keys.
					 * 1 causes a search to just after the key.
				 	*/
					if (search_flags & XT_SEARCH_AFTER_KEY)
						r = 1;
					else
						r = -1;
				}
			}

			if (r < 0)
				count = guess;
			else
				i = guess + 1;
		}
	}

	bitem = base + i * full_item_size;
	xt_get_res_record_ref(bitem + ind->mi_key_size, result);
	result->sr_branch = XT_GET_NODE_REF(bitem - node_ref_size);			/* Only valid if this is a node. */
	result->sr_item.i_item_offset = node_ref_size + i * full_item_size;
}

/*
 * Variable length key values are stored as a sorted list. Since each list item has a variable length, we
 * must scan the list sequentially in order to find a key.
 */
xtPublic void xt_scan_branch_var(XTIndexPtr ind, XTIdxBranchDPtr branch, register XTIdxKeyValuePtr value, register XTIdxResultRec *result)
{
	u_int			branch_size;
	u_int			node_ref_size;
	int				search_flags;
	xtWord1			*base;
	xtWord1			*bitem;
	u_int			ilen;
	xtWord1			*bend;

	branch_size = XT_GET_DISK_2(branch->tb_size_2);
	node_ref_size = XT_IS_NODE(branch_size) ? XT_NODE_REF_SIZE : 0;

	result->sr_found = FALSE;
	result->sr_duplicate = FALSE;
	result->sr_item.i_total_size = XT_GET_BRANCH_DATA_SIZE(branch_size);
	ASSERT_NS((int) result->sr_item.i_total_size >= 0 && result->sr_item.i_total_size <= XT_INDEX_PAGE_SIZE-2);

	result->sr_item.i_node_ref_size = node_ref_size;

	search_flags = value->sv_flags;
	base = branch->tb_data + node_ref_size;
	bitem = base;
	bend = &branch->tb_data[result->sr_item.i_total_size];
	ilen = 0;
	if (bitem >= bend)
		goto done_ok;

	if (search_flags & XT_SEARCH_FIRST_FLAG)
		ilen = myxt_get_key_length(ind, bitem);
	else if (search_flags & XT_SEARCH_AFTER_LAST_FLAG) {
		bitem = bend;
		ilen = 0;
	}
	else {
		xtRecordID	key_record;
		int			r;

		key_record = value->sv_rec_id;

		ASSERT_NS(ind);
		while (bitem < bend) {
			ilen = myxt_get_key_length(ind, bitem);
			r = myxt_compare_key(ind, search_flags, value->sv_length, value->sv_key, bitem);
			if (r == 0) {
				if (search_flags & XT_SEARCH_WHOLE_KEY) {
					xtRecordID	item_record;
					xtRowID		row_id;

					xt_get_record_ref(bitem + ilen, &item_record, &row_id);

					/* This should not happen because we should never
					 * try to insert the same record twice into the 
					 * index!
					 */
					result->sr_duplicate = TRUE;
					if (key_record == item_record) {
						result->sr_found = TRUE;
						result->sr_item.i_item_size = ilen + XT_RECORD_REF_SIZE;;
						result->sr_rec_id = item_record;
						result->sr_row_id = row_id;
						result->sr_branch = XT_GET_NODE_REF(bitem - node_ref_size);
						result->sr_item.i_item_offset = bitem - branch->tb_data;
						return;
					}
					if (key_record < item_record)
						r = -1;
					else
						r = 1;
				}
				else {
					result->sr_found = TRUE;
					/* -1 causes a search to the beginning of the duplicate list of keys.
					 * 1 causes a search to just after the key.
				 	*/
					if (search_flags & XT_SEARCH_AFTER_KEY)
						r = 1;
					else
						r = -1;
				}
			}
			if (r <= 0)
				break;
			bitem += ilen + XT_RECORD_REF_SIZE + node_ref_size;
		}
	}

	done_ok:
	result->sr_item.i_item_size = ilen + XT_RECORD_REF_SIZE;
	xt_get_res_record_ref(bitem + ilen, result);
	result->sr_branch = XT_GET_NODE_REF(bitem - node_ref_size);			/* Only valid if this is a node. */
	result->sr_item.i_item_offset = bitem - branch->tb_data;
}

/* Go to the next item in the node. */
static void idx_next_branch_item(XTIndexPtr ind, XTIdxBranchDPtr branch, register XTIdxResultRec *result)
{
	xtWord1	*bitem;
	u_int	ilen;

	result->sr_item.i_item_offset += result->sr_item.i_item_size + result->sr_item.i_node_ref_size;
	bitem = branch->tb_data + result->sr_item.i_item_offset;
	if (ind->mi_fix_key)
		ilen = result->sr_item.i_item_size;
	else {
		ilen = myxt_get_key_length(ind, bitem) + XT_RECORD_REF_SIZE;
		result->sr_item.i_item_size = ilen;
	}
	xt_get_res_record_ref(bitem + ilen - XT_RECORD_REF_SIZE, result); /* (Only valid if i_item_offset < i_total_size) */
	result->sr_branch = XT_GET_NODE_REF(bitem - result->sr_item.i_node_ref_size);
}

xtPublic void xt_prev_branch_item_fix(XTIndexPtr ind, XTIdxBranchDPtr branch, register XTIdxResultRec *result)
{
	ASSERT_NS(result->sr_item.i_item_offset >= result->sr_item.i_item_size + result->sr_item.i_node_ref_size + result->sr_item.i_node_ref_size);
	result->sr_item.i_item_offset -= (result->sr_item.i_item_size + result->sr_item.i_node_ref_size);
	xt_get_res_record_ref(branch->tb_data + result->sr_item.i_item_offset + result->sr_item.i_item_size - XT_RECORD_REF_SIZE, result); /* (Only valid if i_item_offset < i_total_size) */
	result->sr_branch = XT_GET_NODE_REF(branch->tb_data + result->sr_item.i_item_offset - result->sr_item.i_node_ref_size);
}

xtPublic void xt_prev_branch_item_var(XTIndexPtr ind, XTIdxBranchDPtr branch, register XTIdxResultRec *result)
{
	xtWord1	*bitem;
	xtWord1	*bend;
	u_int	ilen;

	bitem = branch->tb_data + result->sr_item.i_node_ref_size;
	bend = &branch->tb_data[result->sr_item.i_item_offset];
	for (;;) {
		ilen = myxt_get_key_length(ind, bitem);
		if (bitem + ilen + XT_RECORD_REF_SIZE + result->sr_item.i_node_ref_size >= bend)
			break;
		bitem += ilen + XT_RECORD_REF_SIZE + result->sr_item.i_node_ref_size;
	}

	result->sr_item.i_item_size = ilen + XT_RECORD_REF_SIZE;
	xt_get_res_record_ref(bitem + ilen, result); /* (Only valid if i_item_offset < i_total_size) */
	result->sr_branch = XT_GET_NODE_REF(bitem - result->sr_item.i_node_ref_size);
	result->sr_item.i_item_offset = bitem - branch->tb_data;
}

static void idx_first_branch_item(XTIndexPtr ind, XTIdxBranchDPtr branch, register XTIdxResultPtr result)
{
	u_int branch_size;
	u_int node_ref_size;
	u_int key_data_size;

	branch_size = XT_GET_DISK_2(branch->tb_size_2);
	node_ref_size = XT_IS_NODE(branch_size) ? XT_NODE_REF_SIZE : 0;

	result->sr_found = FALSE;
	result->sr_duplicate = FALSE;
	result->sr_item.i_total_size = XT_GET_BRANCH_DATA_SIZE(branch_size);
	ASSERT_NS((int) result->sr_item.i_total_size >= 0 && result->sr_item.i_total_size <= XT_INDEX_PAGE_SIZE-2);

	if (ind->mi_fix_key)
		key_data_size = ind->mi_key_size;
	else {
		xtWord1 *bitem;

		bitem = branch->tb_data + node_ref_size;
		if (bitem < &branch->tb_data[result->sr_item.i_total_size])
			key_data_size = myxt_get_key_length(ind, bitem);
		else
			key_data_size = 0;
	}

	result->sr_item.i_item_size = key_data_size + XT_RECORD_REF_SIZE;
	result->sr_item.i_node_ref_size = node_ref_size;

	xt_get_res_record_ref(branch->tb_data + node_ref_size + key_data_size, result);
	result->sr_branch = XT_GET_NODE_REF(branch->tb_data);				/* Only valid if this is a node. */
	result->sr_item.i_item_offset = node_ref_size;
}

/*
 * Last means different things for leaf or node!
 */
xtPublic void xt_last_branch_item_fix(XTIndexPtr ind, XTIdxBranchDPtr branch, register XTIdxResultPtr result)
{
	u_int branch_size;
	u_int node_ref_size;

	branch_size = XT_GET_DISK_2(branch->tb_size_2);
	node_ref_size = XT_IS_NODE(branch_size) ? XT_NODE_REF_SIZE : 0;

	result->sr_found = FALSE;
	result->sr_duplicate = FALSE;
	result->sr_item.i_total_size = XT_GET_BRANCH_DATA_SIZE(branch_size);
	ASSERT_NS((int) result->sr_item.i_total_size >= 0 && result->sr_item.i_total_size <= XT_INDEX_PAGE_SIZE-2);

	result->sr_item.i_item_size = ind->mi_key_size + XT_RECORD_REF_SIZE;
	result->sr_item.i_node_ref_size = node_ref_size;

	if (node_ref_size) {
		result->sr_item.i_item_offset = result->sr_item.i_total_size;
		result->sr_branch = XT_GET_NODE_REF(&branch->tb_data[result->sr_item.i_item_offset - node_ref_size]);
	}
	else {
		if (result->sr_item.i_total_size) {
			result->sr_item.i_item_offset = result->sr_item.i_total_size - result->sr_item.i_item_size;
			xt_get_res_record_ref(branch->tb_data + result->sr_item.i_item_offset + ind->mi_key_size, result);
		}
		else
			/* Leaf is empty: */
			result->sr_item.i_item_offset = 0;
	}
}

xtPublic void xt_last_branch_item_var(XTIndexPtr ind, XTIdxBranchDPtr branch, register XTIdxResultPtr result)
{
	u_int	branch_size;
	u_int	node_ref_size;

	branch_size = XT_GET_DISK_2(branch->tb_size_2);
	node_ref_size = XT_IS_NODE(branch_size) ? XT_NODE_REF_SIZE : 0;

	result->sr_found = FALSE;
	result->sr_duplicate = FALSE;
	result->sr_item.i_total_size = XT_GET_BRANCH_DATA_SIZE(branch_size);
	ASSERT_NS((int) result->sr_item.i_total_size >= 0 && result->sr_item.i_total_size <= XT_INDEX_PAGE_SIZE-2);

	result->sr_item.i_node_ref_size = node_ref_size;

	if (node_ref_size) {
		result->sr_item.i_item_offset = result->sr_item.i_total_size;
		result->sr_branch = XT_GET_NODE_REF(&branch->tb_data[result->sr_item.i_item_offset - node_ref_size]);
		result->sr_item.i_item_size = 0;
	}
	else {
		if (result->sr_item.i_total_size) {
			xtWord1	*bitem;
			u_int	ilen;
			xtWord1	*bend;

			bitem = branch->tb_data + node_ref_size;;
			bend = &branch->tb_data[result->sr_item.i_total_size];
			ilen = 0;
			if (bitem < bend) {
				for (;;) {
					ilen = myxt_get_key_length(ind, bitem);
					if (bitem + ilen + XT_RECORD_REF_SIZE + node_ref_size >= bend)
						break;
					bitem += ilen + XT_RECORD_REF_SIZE + node_ref_size;
				}
			}

			result->sr_item.i_item_offset = bitem - branch->tb_data;
			xt_get_res_record_ref(bitem + ilen, result);
			result->sr_item.i_item_size = ilen + XT_RECORD_REF_SIZE;
		}
		else {
			/* Leaf is empty: */
			result->sr_item.i_item_offset = 0;
			result->sr_item.i_item_size = 0;
		}
	}
}

/*
 * Remove an item and save to disk.
 */
static xtBool idx_remove_branch_item_right(XTOpenTablePtr ot, XTIndexPtr ind, off_t address, XTIdxBranchDPtr branch, register XTIdxItemPtr item)
{
	u_int size = item->i_item_size + item->i_node_ref_size;

	/* Remove the node reference to the left of the item: */
	memmove(&branch->tb_data[item->i_item_offset],
		&branch->tb_data[item->i_item_offset + size],
		item->i_total_size - item->i_item_offset - size);
	item->i_total_size -= size;
	XT_SET_DISK_2(branch->tb_size_2, XT_MAKE_BRANCH_SIZE(item->i_total_size, item->i_node_ref_size));
	return xt_ind_write(ot->ot_ind_file, address, offsetof(XTIdxBranchDRec, tb_data) + item->i_total_size, (xtWord1 *) branch);
}

static xtBool idx_remove_branch_item_left(XTOpenTablePtr ot, XTIndexPtr ind, off_t address, XTIdxBranchDPtr branch, register XTIdxItemPtr item)
{
	u_int size = item->i_item_size + item->i_node_ref_size;

	/* Remove the node reference to the left of the item: */
	memmove(&branch->tb_data[item->i_item_offset - item->i_node_ref_size],
		&branch->tb_data[item->i_item_offset + item->i_item_size],
		item->i_total_size - item->i_item_offset - item->i_item_size);
	item->i_total_size -= size;
	XT_SET_DISK_2(branch->tb_size_2, XT_MAKE_BRANCH_SIZE(item->i_total_size, item->i_node_ref_size));
	return xt_ind_write(ot->ot_ind_file, address, offsetof(XTIdxBranchDRec, tb_data) + item->i_total_size, (xtWord1 *) branch);
}

static void idx_insert_leaf_item(XTIndexPtr ind, XTIdxBranchDPtr leaf, XTIdxKeyValuePtr value, XTIdxResultPtr result)
{
	xtWord1 *item;

	/* This will ensure we do not overwrite the end of the buffer: */
	ASSERT_NS(value->sv_length <= XT_INDEX_MAX_KEY_SIZE);
	memmove(&leaf->tb_data[result->sr_item.i_item_offset + value->sv_length + XT_RECORD_REF_SIZE],
		&leaf->tb_data[result->sr_item.i_item_offset],
		result->sr_item.i_total_size - result->sr_item.i_item_offset);
	item = &leaf->tb_data[result->sr_item.i_item_offset];
	memcpy(item, value->sv_key, value->sv_length);
	xt_set_val_record_ref(item + value->sv_length, value);
	result->sr_item.i_total_size += value->sv_length + XT_RECORD_REF_SIZE;
	XT_SET_DISK_2(leaf->tb_size_2, XT_MAKE_LEAF_SIZE(result->sr_item.i_total_size));
}

static void idx_insert_node_item(XTIndexPtr ind, XTIdxBranchDPtr leaf, XTIdxKeyValuePtr value, XTIdxResultPtr result, off_t branch)
{
	xtWord1 *item;

	/* This will ensure we do not overwrite the end of the buffer: */
	ASSERT_NS(value->sv_length <= XT_INDEX_MAX_KEY_SIZE);
	memmove(&leaf->tb_data[result->sr_item.i_item_offset + value->sv_length + XT_RECORD_REF_SIZE + result->sr_item.i_node_ref_size],
		&leaf->tb_data[result->sr_item.i_item_offset],
		result->sr_item.i_total_size - result->sr_item.i_item_offset);
	item = &leaf->tb_data[result->sr_item.i_item_offset];
	memcpy(item, value->sv_key, value->sv_length);
	xt_set_val_record_ref(item + value->sv_length, value);
	XT_SET_NODE_REF(item + value->sv_length + XT_RECORD_REF_SIZE, branch);
	result->sr_item.i_total_size += value->sv_length + XT_RECORD_REF_SIZE + result->sr_item.i_node_ref_size;
	XT_SET_DISK_2(leaf->tb_size_2, XT_MAKE_NODE_SIZE(result->sr_item.i_total_size));
}

static void idx_get_middle_branch_item(XTIndexPtr ind, XTIdxBranchDPtr branch, XTIdxKeyValuePtr value, XTIdxResultPtr result)
{
	xtWord1	*bitem;

	if (ind->mi_fix_key) {
		u_int full_item_size = result->sr_item.i_item_size + result->sr_item.i_node_ref_size;

		result->sr_item.i_item_offset = ((result->sr_item.i_total_size - result->sr_item.i_node_ref_size)
			/ full_item_size / 2 * full_item_size) + result->sr_item.i_node_ref_size;

		bitem = &branch->tb_data[result->sr_item.i_item_offset];
		value->sv_flags = XT_SEARCH_WHOLE_KEY;
		value->sv_length = result->sr_item.i_item_size - XT_RECORD_REF_SIZE;
		xt_get_record_ref(bitem + value->sv_length, &value->sv_rec_id, &value->sv_row_id);
		memcpy(value->sv_key, bitem, value->sv_length);
	}
	else {
		u_int	node_ref_size;
		u_int	ilen;
		xtWord1	*bend;

		node_ref_size = result->sr_item.i_node_ref_size;
		bitem = branch->tb_data + node_ref_size;;
		bend = &branch->tb_data[(result->sr_item.i_total_size - node_ref_size) / 2 + node_ref_size];
		ilen = 0;
		if (bitem < bend) {
			for (;;) {
				ilen = myxt_get_key_length(ind, bitem);
				if (bitem + ilen + XT_RECORD_REF_SIZE + node_ref_size >= bend)
					break;
				bitem += ilen + XT_RECORD_REF_SIZE + node_ref_size;
			}
		}

		result->sr_item.i_item_offset = bitem - branch->tb_data;
		result->sr_item.i_item_size = ilen + XT_RECORD_REF_SIZE;

		value->sv_flags = XT_SEARCH_WHOLE_KEY;
		value->sv_length = ilen;
		xt_get_record_ref(bitem + ilen, &value->sv_rec_id, &value->sv_row_id);
		memcpy(value->sv_key, bitem, value->sv_length);
	}
}

static size_t idx_write_branch_item(XTIndexPtr ind, xtWord1 *item, XTIdxKeyValuePtr value)
{
	memcpy(item, value->sv_key, value->sv_length);
	xt_set_val_record_ref(item + value->sv_length, value);
	return value->sv_length + XT_RECORD_REF_SIZE;
}

static xtBool idx_replace_node_key(XTOpenTablePtr ot, XTIndexPtr ind, IdxStackItemPtr item, IdxBranchStackPtr stack, u_int item_size, xtWord1 *item_buf)
{
	XTIdxBranchDPtr		branch = &ot->ot_ind_wbuf;
	off_t				new_branch;
	XTIdxResultRec		result;
	off_t				current = item->i_branch;
	u_int				new_size;
	XTIdxBranchDPtr		new_branch_ptr;
	XTIdxKeyValueRec	key_value;
	xtWord1				key_buf[XT_INDEX_MAX_KEY_SIZE];

	if (!xt_ind_read_block(ot->ot_ind_file, current, (xtWord1 *) branch))
		return FAILED;
	memmove(&branch->tb_data[item->i_pos.i_item_offset + item_size],
		&branch->tb_data[item->i_pos.i_item_offset + item->i_pos.i_item_size],
		item->i_pos.i_total_size - item->i_pos.i_item_offset - item->i_pos.i_item_size);
	memcpy(&branch->tb_data[item->i_pos.i_item_offset],
		item_buf, item_size);
	item->i_pos.i_total_size = item->i_pos.i_total_size + item_size - item->i_pos.i_item_size;
	XT_SET_DISK_2(branch->tb_size_2, XT_MAKE_NODE_SIZE(item->i_pos.i_total_size));

	if (item->i_pos.i_total_size <= XT_INDEX_PAGE_DATA_SIZE)
		return xt_ind_write(ot->ot_ind_file, current, offsetof(XTIdxBranchDRec, tb_data) + item->i_pos.i_total_size, (xtWord1 *) branch);

	/* The node has overflowed!! */
	result.sr_item = item->i_pos;

	/* Adjust the stack (we want the parents of the delete node): */
	for (;;) {
		if (idx_pop(stack) == item)
			break;
	}		

	/* We assume that value can be overwritten (which is the case) */
	key_value.sv_flags = XT_SEARCH_WHOLE_KEY;
	key_value.sv_key = key_buf;
	idx_get_middle_branch_item(ind, &ot->ot_ind_wbuf, &key_value, &result);

	if (!idx_new_branch(ot, &new_branch))
		return FAILED;

	/* Split the node: */
	new_size = result.sr_item.i_total_size - result.sr_item.i_item_offset - result.sr_item.i_item_size;
	new_branch_ptr = (XTIdxBranchDPtr) &ot->ot_ind_wbuf.tb_data[XT_INDEX_PAGE_DATA_SIZE];
	memmove(new_branch_ptr->tb_data, &ot->ot_ind_wbuf.tb_data[result.sr_item.i_item_offset + result.sr_item.i_item_size], new_size);

	XT_SET_DISK_2(new_branch_ptr->tb_size_2, XT_MAKE_NODE_SIZE(new_size));
	if (!xt_ind_write(ot->ot_ind_file, new_branch, offsetof(XTIdxBranchDRec, tb_data) + new_size, (xtWord1 *) new_branch_ptr))
		goto failed_2;

	/* Change the size of the old branch: */
	XT_SET_DISK_2(ot->ot_ind_wbuf.tb_size_2, XT_MAKE_NODE_SIZE(result.sr_item.i_item_offset));
	if (!xt_ind_write(ot->ot_ind_file, current, offsetof(XTIdxBranchDRec, tb_data) + result.sr_item.i_item_offset, (xtWord1 *) &ot->ot_ind_wbuf))
		goto failed_2;

	/* Insert the new branch into the parent node, using the new middle key value: */
	if (!idx_insert_node(ot, ind, stack, &key_value, new_branch))
		goto failed_3;

	return OK;

	failed_3:
	// Index may be inconsistant now...

	failed_2:
	idx_free_branch(ot, new_branch);

	return FAILED;
}

/*
 * -----------------------------------------------------------------------
 * Standard b-tree insert
 */

/*
 * Insert the given branch into the node on the top of the stack. If the stack
 * is empty we need to add a new root.
 */
static xtBool idx_insert_node(XTOpenTablePtr ot, XTIndexPtr ind, IdxBranchStackPtr stack, XTIdxKeyValuePtr key_value, off_t branch)
{
	IdxStackItemPtr	stack_item;
	off_t			new_branch;
	size_t			size;
	off_t			current;
	XTIdxResultRec	result;
	u_int			new_size;
	XTIdxBranchDPtr	new_branch_ptr;

	/* Insert a new branch (key, data)... */
	if (!(stack_item = idx_pop(stack))) {
		xtWord1 *ditem;

		/* New root */
		if (!idx_new_branch(ot, &new_branch))
			goto failed;

		ditem = ot->ot_ind_wbuf.tb_data;
		XT_SET_NODE_REF(ditem, ind->mi_root);
		ditem += XT_NODE_REF_SIZE;
		ditem += idx_write_branch_item(ind, ditem, key_value);
		XT_SET_NODE_REF(ditem, branch);
		ditem += XT_NODE_REF_SIZE;
		size = ditem - ot->ot_ind_wbuf.tb_data;
		XT_SET_DISK_2(ot->ot_ind_wbuf.tb_size_2, XT_MAKE_NODE_SIZE(size));
		if (!xt_ind_write(ot->ot_ind_file, new_branch, offsetof(XTIdxBranchDRec, tb_data) + size, (xtWord1 *) &ot->ot_ind_wbuf))
			goto failed_2;
		ind->mi_root = new_branch;
		ot->ot_table->tab_ind_head_dirty = TRUE;
		goto done_ok;
	}

	current = stack_item->i_branch;
	if (!xt_ind_read_block(ot->ot_ind_file, current, (xtWord1 *) &ot->ot_ind_wbuf))
		goto failed;
	ASSERT_NS(XT_IS_NODE(XT_GET_DISK_2(ot->ot_ind_wbuf.tb_size_2)));
	ind->mi_scan_branch(ind, &ot->ot_ind_wbuf, key_value, &result);

	idx_insert_node_item(ind, &ot->ot_ind_wbuf, key_value, &result, branch);
	if (result.sr_item.i_total_size <= XT_INDEX_PAGE_DATA_SIZE) {
		if (!xt_ind_write(ot->ot_ind_file, current, offsetof(XTIdxBranchDRec, tb_data) + result.sr_item.i_total_size, (xtWord1 *) &ot->ot_ind_wbuf))
			goto failed;
		goto done_ok;
	}

	/* We assume that value can be overwritten (which is the case) */
	idx_get_middle_branch_item(ind, &ot->ot_ind_wbuf, key_value, &result);

	if (!idx_new_branch(ot, &new_branch))
		goto failed;

	/* Split the node: */
	new_size = result.sr_item.i_total_size - result.sr_item.i_item_offset - result.sr_item.i_item_size;
	new_branch_ptr = (XTIdxBranchDPtr) &ot->ot_ind_wbuf.tb_data[XT_INDEX_PAGE_DATA_SIZE];
	memmove(new_branch_ptr->tb_data, &ot->ot_ind_wbuf.tb_data[result.sr_item.i_item_offset + result.sr_item.i_item_size], new_size);

	XT_SET_DISK_2(new_branch_ptr->tb_size_2, XT_MAKE_NODE_SIZE(new_size));
	if (!xt_ind_write(ot->ot_ind_file, new_branch, offsetof(XTIdxBranchDRec, tb_data) + new_size, (xtWord1 *) new_branch_ptr))
		goto failed_2;

	/* Change the size of the old branch: */
	XT_SET_DISK_2(ot->ot_ind_wbuf.tb_size_2, XT_MAKE_NODE_SIZE(result.sr_item.i_item_offset));
	if (!xt_ind_write(ot->ot_ind_file, current, offsetof(XTIdxBranchDRec, tb_data) + result.sr_item.i_item_offset, (xtWord1 *) &ot->ot_ind_wbuf))
		goto failed_2;

	/* Insert the new branch into the parent node, using the new middle key value: */
	if (!idx_insert_node(ot, ind, stack, key_value, new_branch))
		goto failed_3;

	done_ok:
	return OK;

	failed_3:
	// Index may be inconsistant now...

	failed_2:
	idx_free_branch(ot, new_branch);

	failed:
	return FAILED;
}

/*
 * Check all the duplicate variation in an index.
 * If one of them is visible, then we have a duplicate key
 * error.
 *
 * GOTCHA: This routine must use the write index buffer!
 */
static xtBool idx_check_duplicates(XTOpenTablePtr ot, XTIndexPtr ind, XTIdxKeyValuePtr key_value)
{
	IdxBranchStackRec	stack;
	off_t				current;
	XTIdxResultRec		result;
	xtBool				on_key = FALSE;
	xtXactID			xn_id;
	int					save_flags;				

	retry:
	idx_newstack(&stack);

	if (!(current = ind->mi_root))
		return OK;

	save_flags = key_value->sv_flags;
	key_value->sv_flags = 0;

	while (current) {
		if (!xt_ind_read_block(ot->ot_ind_file, current, (xtWord1 *) &ot->ot_ind_wbuf)) {
			key_value->sv_flags = save_flags;
			return FAILED;
		}
		ind->mi_scan_branch(ind, &ot->ot_ind_wbuf, key_value, &result);
		if (result.sr_found)
			/* If we have found the key in a node: */
			on_key = TRUE;
		if (!result.sr_item.i_node_ref_size)
			break;
		if (!idx_push(&stack, current, &result.sr_item)) {
			key_value->sv_flags = save_flags;
			return FAILED;
		}
		current = result.sr_branch;
	}

	key_value->sv_flags = save_flags;

	if (!on_key)
		return OK;

	for (;;) {
		if (result.sr_item.i_item_offset == result.sr_item.i_total_size) {
			IdxStackItemPtr node;

			/* We are at the end of a leaf node.
			 * Go up the stack to find the start position of the next key.
			 * If we find none, then we are the end of the index.
			 */
			while ((node = idx_pop(&stack))) {
				if (node->i_pos.i_item_offset < node->i_pos.i_total_size) {
					current = node->i_branch;
					if (!xt_ind_read_block(ot->ot_ind_file, current, (xtWord1 *) &ot->ot_ind_wbuf))
						return FAILED;
					xt_get_res_record_ref(&ot->ot_ind_wbuf.tb_data[node->i_pos.i_item_offset + node->i_pos.i_item_size - XT_RECORD_REF_SIZE], &result);
					result.sr_item = node->i_pos;
					goto check_value;
				}
			}
			break;
		}

		check_value:
		/* Quit the loop if the key is no longer matched! */
		if (myxt_compare_key(ind, 0, key_value->sv_length, key_value->sv_key, &ot->ot_ind_wbuf.tb_data[result.sr_item.i_item_offset]) != 0)
			break;
		/*
		if (key_value->sv_length != result.sr_item.i_item_size - XT_RECORD_REF_SIZE)
			break;
		if (memcmp(key_value->sv_key, &ot->ot_ind_wbuf.tb_data[result.sr_item.i_item_offset], key_value->sv_length) != 0)
			break;
		*/

		switch (xt_tab_maybe_committed(ot, result.sr_rec_id, &xn_id, NULL, NULL)) {
			case XT_MAYBE:
				/* Record is not committed, wait for the transaction. */
				xt_rwlock_unlock(&ind->mi_rwlock);
				if (!xt_xn_wait_for_xact(ot->ot_thread, xn_id, FALSE)) {
					xt_rwlock_wrlock(&ind->mi_rwlock);
					return FAILED;
				}
				xt_rwlock_wrlock(&ind->mi_rwlock);
				goto retry;			
			case XT_ERR:
				/* Error while reading... */
				return FAILED;
			case TRUE:
				/* Record is committed or belongs to me, duplicate key: */
				XT_DEBUG_TRACE(("DUPLICATE KEY tx=%d rec=%d\n", (int) ot->ot_thread->st_xact_data->xd_start_xn_id, (int) result.sr_rec_id));
				xt_register_xterr(XT_REG_CONTEXT, XT_ERR_DUPLICATE_KEY);
				return FAILED;
			case FALSE:
				/* Record is deleted or rolled-back: */
				break;
		}

		idx_next_branch_item(ind, &ot->ot_ind_wbuf, &result);

		if (result.sr_item.i_node_ref_size) {
			/* Go down to the bottom: */
			while (current) {
				if (!idx_push(&stack, current, &result.sr_item))
					return FAILED;
				current = result.sr_branch;
				if (!xt_ind_read_block(ot->ot_ind_file, current, (xtWord1 *) &ot->ot_ind_wbuf))
					return FAILED;
				idx_first_branch_item(ind, &ot->ot_ind_wbuf, &result);
				if (!result.sr_item.i_node_ref_size)
					break;
			}
		}
	}
	return OK;
}

/*
 * Insert a value into the given index. Return FALSE if an error occurs.
 */
xtPublic xtBool xt_idx_insert(XTOpenTablePtr ot, XTIndexPtr ind, xtRecordID rec_id, xtWord1 *rec_buf, xtWord1 *bef_buf, xtBool allow_dups)
{
	XTIdxKeyValueRec	key_value;
	xtWord1				key_buf[XT_INDEX_MAX_KEY_SIZE];
	IdxBranchStackRec	stack;
	off_t				current;
	off_t				new_branch;
	size_t				size;
	XTIdxResultRec		result;
	size_t				new_size;
	XTIdxBranchDPtr		new_branch_ptr;
	xtBool				check_for_dups = ind->mi_flags & (HA_UNIQUE_CHECK | HA_NOSAME) && !allow_dups;

#ifdef CHECK_AND_PRINT
	//idx_print_index(ot, ind, TRUE);
#endif

	key_value.sv_flags = XT_SEARCH_WHOLE_KEY;
	key_value.sv_rec_id = rec_id;
	key_value.sv_row_id = 0;		/* Always set the row ID to zero on insert (will be update by sweeper later). */
	key_value.sv_key = key_buf;
	key_value.sv_length = myxt_create_key_from_row(ind, key_buf, rec_buf, &check_for_dups);

//#ifdef xxxx
	if (bef_buf && check_for_dups) {
		/* If we have a before image, and we are required to check for duplicates.
		 * then compare the before image key with the after image key.
		 */
		xtWord1	bef_key_buf[XT_INDEX_MAX_KEY_SIZE];
		u_int	len;
		xtBool	has_no_null = TRUE;

		len = myxt_create_key_from_row(ind, bef_key_buf, bef_buf, &has_no_null);
		if (has_no_null) {
			/* If the before key has no null values, then compare with the after key value.
			 * We only have to check for duplicates if the key has changed!
			 */
			check_for_dups = myxt_compare_key(ind, 0, len, bef_key_buf, key_buf) != 0;
		}
	}
//#endif

	idx_newstack(&stack);

	xt_rwlock_wrlock(&ind->mi_rwlock);

	retry:
	if (!(current = ind->mi_root)) {
		if (!idx_new_branch(ot, &new_branch))
			goto failed;
		size = idx_write_branch_item(ind, ot->ot_ind_wbuf.tb_data, &key_value);
		XT_SET_DISK_2(ot->ot_ind_wbuf.tb_size_2, XT_MAKE_LEAF_SIZE(size));
		if (!xt_ind_write(ot->ot_ind_file, new_branch, offsetof(XTIdxBranchDRec, tb_data) + size, (xtWord1 *) &ot->ot_ind_wbuf))
			goto failed_2;
		ind->mi_root = new_branch;
		ot->ot_table->tab_ind_head_dirty = TRUE;
		goto done_ok;
	}

	while (current) {
		if (!xt_ind_read_block(ot->ot_ind_file, current, (xtWord1 *) &ot->ot_ind_wbuf))
			goto failed;
		ind->mi_scan_branch(ind, &ot->ot_ind_wbuf, &key_value, &result);
		if (result.sr_duplicate) {
			if (check_for_dups) {
				/* Duplicates are not allowed, at least one has been
				 * found.
				 */
				if (!idx_check_duplicates(ot, ind, &key_value))
					goto failed;
				/* We have checked all the "duplicate" variations. None of them are
				 * relevant. So this will cause a correct insert.
				 */
				check_for_dups = FALSE;
				idx_newstack(&stack);
				goto retry;
			}
		}
		if (result.sr_found) {
			/* Node found, can happen during recovery of indexes! */
			goto done_ok;
		}
		/* Stop when we get to a leaf: */
		if (!result.sr_item.i_node_ref_size)
			break;
		if (!idx_push(&stack, current, NULL))
			goto failed;
		current = result.sr_branch;
	}

	idx_insert_leaf_item(ind, &ot->ot_ind_wbuf, &key_value, &result);
	if (result.sr_item.i_total_size <= XT_INDEX_PAGE_DATA_SIZE) {
		if (!xt_ind_write(ot->ot_ind_file, current, offsetof(XTIdxBranchDRec, tb_data) + result.sr_item.i_total_size, (xtWord1 *) &ot->ot_ind_wbuf))
			goto failed;
		goto done_ok;
	}

#ifdef XT_INDEX_LOG_CHANGES
	u_int insert_offset = result.sr_item.i_item_offset;
	u_int insert_len = key_value.sv_length + XT_RECORD_REF_SIZE;
	u_int org_size = result.sr_item.i_total_size - insert_len;
#endif

	/* Key does not fit, must split... */
	idx_get_middle_branch_item(ind, &ot->ot_ind_wbuf, &key_value, &result);

	if (!idx_new_branch(ot, &new_branch))
		goto failed;

#ifdef XT_INDEX_LOG_CHANGES
	/* Note: the middle item is not copied into any of the
	 * 2 resulting nodes.
	 */
	if (result.sr_item.i_item_offset <= insert_offset) {
		/* This is were the node split occored before the insertion point. */
		u_int toff = result.sr_item.i_item_offset + result.sr_item.i_item_size;

		/* Copy the end part of the current node to the new node. */
		XT_COPY_TO_NEW(current, toff, org_size - toff, new_branch);
		if (insert_offset > toff) {
			/* Insert the the data into the new branch: */
			toff = insert_offset - toff;
			XT_INSERT_INTO_BRANCH(new_branch, toff, insert_len, &ot->ot_ind_wbuf.tb_data[insert_offset]);
		}
		/* Set the current node size. */
		XT_SET_NODE_SIZE(current, result.sr_item.i_item_offset);
	}
	else {
		/* This is the case where the split is after the insert point. */
		/* This is the original offset into the node: */
		u_int toff = result.sr_item.i_item_offset - insert_len;

		/* Copy to the new node, and set the size of the old: */
		XT_COPY_TO_NEW(current, toff, org_size - toff - result.sr_item.i_item_size, new_branch);
		XT_SET_NODE_SIZE(current, toff);

		/* Insert into the correct position in old node: */
		XT_INSERT_INTO_BRANCH(current, inserted_offset, insert_len, &ot->ot_ind_wbuf.tb_data[insert_offset]);
	}
#endif

	/* Copy and write the rest of the data to the new node: */
	new_size = result.sr_item.i_total_size - result.sr_item.i_item_offset - result.sr_item.i_item_size;
	new_branch_ptr = (XTIdxBranchDPtr) &ot->ot_ind_wbuf.tb_data[XT_INDEX_PAGE_DATA_SIZE];
	memmove(new_branch_ptr->tb_data, &ot->ot_ind_wbuf.tb_data[result.sr_item.i_item_offset + result.sr_item.i_item_size], new_size);

	XT_SET_DISK_2(new_branch_ptr->tb_size_2, XT_MAKE_LEAF_SIZE(new_size));
	if (!xt_ind_write(ot->ot_ind_file, new_branch, offsetof(XTIdxBranchDRec, tb_data) + new_size, (xtWord1 *) new_branch_ptr))
		goto failed_2;

	/* Modify the first node: */
	XT_SET_DISK_2(ot->ot_ind_wbuf.tb_size_2, XT_MAKE_LEAF_SIZE(result.sr_item.i_item_offset));
	if (!xt_ind_write(ot->ot_ind_file, current, offsetof(XTIdxBranchDRec, tb_data) + result.sr_item.i_item_offset, (xtWord1 *) &ot->ot_ind_wbuf))
		goto failed_2;

	/* Insert the new branch into the parent node, using the new middle key value: */
	if (!idx_insert_node(ot, ind, &stack, &key_value, new_branch))
		goto failed_3;


	done_ok:
	xt_rwlock_unlock(&ind->mi_rwlock);

#ifdef DEBUG
	//printf("INSERT OK\n");
	//idx_print_index(ot, ind, TRUE);
	//idx_check_index(ot, ind);
#endif
	return OK;

	failed_3:
	// Index may be inconsistant now...

	failed_2:
	idx_free_branch(ot, new_branch);

	failed:
	xt_rwlock_unlock(&ind->mi_rwlock);

#ifdef DEBUG
	//printf("INSERT FAILED\n");
	//idx_print_index(ot, ind, TRUE);
#endif
	return FAILED;
}

static xtBool idx_delete(XTOpenTablePtr ot, XTIndexPtr ind, XTIdxKeyValuePtr key_value)
{
	IdxBranchStackRec	stack;
	off_t				current;
	XTIdxResultRec		result;
	IdxStackItemPtr		delete_node = NULL;
	IdxStackItemPtr		current_top = NULL;

	idx_newstack(&stack);

	xt_rwlock_wrlock(&ind->mi_rwlock);

	if (!(current = ind->mi_root))
		goto done_ok;

	while (current) {
		if (!xt_ind_read_block(ot->ot_ind_file, current, (xtWord1 *) &ot->ot_ind_wbuf))
			goto failed;
		ind->mi_scan_branch(ind, &ot->ot_ind_wbuf, key_value, &result);
		if (!result.sr_item.i_node_ref_size) {
			if (result.sr_found) {
				if (!idx_remove_branch_item_right(ot, ind, current, &ot->ot_ind_wbuf, &result.sr_item))
					goto failed;
			}
			goto done_ok;
		}
		if (!idx_push(&stack, current, &result.sr_item))
			goto failed;
		if (result.sr_found)
			/* If we have found the key in a node: */
			break;
		current = result.sr_branch;
	}

	/* This is the item we will have to replace: */
	delete_node = idx_top(&stack);

	/* Follow the branch after this item: */
	idx_next_branch_item(ind, &ot->ot_ind_wbuf, &result);

	/* Go down the left-hand side until we reach a leaf: */
	while (current) {
		current = result.sr_branch;
		if (!xt_ind_read_block(ot->ot_ind_file, current, (xtWord1 *) &ot->ot_ind_wbuf))
			goto failed;
		idx_first_branch_item(ind, &ot->ot_ind_wbuf, &result);
		if (!result.sr_item.i_node_ref_size)
			break;
		if (!idx_push(&stack, current, &result.sr_item))
			goto failed;
	}

	/* Crawl back up the stack trace, looking for a key
	 * that can be used to replace the deleted key.
	 *
	 * Any empty nodes on the way up can be removed!
	 */
	if (result.sr_item.i_total_size > 0) {
		/* There is a key in the leaf, extract it, and put it in the node: */
		memcpy(key_value->sv_key, &ot->ot_ind_wbuf.tb_data[result.sr_item.i_item_offset], result.sr_item.i_item_size);
		if (!idx_remove_branch_item_right(ot, ind, current, &ot->ot_ind_wbuf, &result.sr_item))
			goto failed;
		if (!idx_replace_node_key(ot, ind, delete_node, &stack, result.sr_item.i_item_size, key_value->sv_key))
			goto failed;
		goto done_ok;
	}

	for (;;) {
		/* The current node/leaf is empty, remove it: */
		idx_free_branch(ot, current);

		current_top = idx_pop(&stack);
		current = current_top->i_branch;
		if (!xt_ind_read_block(ot->ot_ind_file, current, (xtWord1 *) &ot->ot_ind_wbuf))
			goto failed;
		
		if (current_top == delete_node) {
			/* All children have been removed. Delete the key and done: */
			if (!idx_remove_branch_item_right(ot, ind, current, &ot->ot_ind_wbuf, &current_top->i_pos))
				goto failed;
			goto done_ok;
		}

		if (current_top->i_pos.i_total_size > current_top->i_pos.i_node_ref_size) {
			/* Save the key: */
			memcpy(key_value->sv_key, &ot->ot_ind_wbuf.tb_data[current_top->i_pos.i_item_offset], current_top->i_pos.i_item_size);
			if (!idx_remove_branch_item_left(ot, ind, current, &ot->ot_ind_wbuf, &current_top->i_pos))
				goto failed;
			if (!idx_replace_node_key(ot, ind, delete_node, &stack, current_top->i_pos.i_item_size, key_value->sv_key))
				goto failed;
			goto done_ok;
		}
	}


	done_ok:
	xt_rwlock_unlock(&ind->mi_rwlock);

#ifdef DEBUG
	//printf("DELETE OK\n");
	//idx_print_index(ot, ind, TRUE);
	//idx_check_index(ot, ind);
#endif
	return OK;

	failed:
	xt_rwlock_unlock(&ind->mi_rwlock);
	return FAILED;
}

xtPublic void xt_idx_delete_current(XTOpenTablePtr ot, XTIndexPtr ind)
{
	XTIdxKeyValueRec	key_value;
	xtWord1				key_buf[XT_INDEX_MAX_KEY_SIZE + XT_MAX_RECORD_REF_SIZE];
	XTIdxItemRec		tmp_state;

	key_value.sv_flags = XT_SEARCH_WHOLE_KEY;
	xt_get_record_ref(&ot->ot_ind_rbuf.tb_data[ot->ot_ind_state.i_item_offset + ot->ot_ind_state.i_item_size - XT_RECORD_REF_SIZE], &key_value.sv_rec_id, &key_value.sv_row_id);
	key_value.sv_length = ot->ot_ind_state.i_item_size - XT_RECORD_REF_SIZE;
	key_value.sv_key = key_buf;
	memcpy(key_buf, &ot->ot_ind_rbuf.tb_data[ot->ot_ind_state.i_item_offset], key_value.sv_length);
	tmp_state = ot->ot_ind_state;

	if (!idx_delete(ot, ind, &key_value))
		xt_log_and_clear_exception_ns();

	/* Setup things so that next or prev will work!
	 * This basically means that we must force a search
	 * for the next key value.
	 *
	 * I assume that what I am trying to do is restore
	 * the search value, to that which we had before the
	 * delete.
	 *
	 * The problem is, key_buf can be changed by idx_delete().
	 *
	 * However, I am assuming that ot_ind_rbuf does not
	 * change so we can just restore the state?
	ot->ot_ind_state.i_node_ref_size = XT_NODE_REF_SIZE;
	ot->ot_ind_state.i_item_offset = 0;
	ot->ot_ind_state.i_item_size = key_value.sv_length + XT_RECORD_REF_SIZE;
	xt_set_val_record_ref(&ot->ot_ind_rbuf.tb_data[key_value.sv_length], &key_value);
	memcpy(ot->ot_ind_rbuf.tb_data, key_buf, key_value.sv_length);
	 */
	ot->ot_ind_state = tmp_state;
	/* This will force search, because it says that this is not a leaf
	 * node. Only leaf nodes can be stepped!
	 */
	ot->ot_ind_state.i_node_ref_size = XT_NODE_REF_SIZE;
}

xtPublic xtBool xt_idx_delete(XTOpenTablePtr ot, XTIndexPtr ind, xtRecordID rec_id, xtWord1 *rec_buf)
{
	XTIdxKeyValueRec	key_value;
	xtWord1				key_buf[XT_INDEX_MAX_KEY_SIZE + XT_MAX_RECORD_REF_SIZE];

	key_value.sv_flags = XT_SEARCH_WHOLE_KEY;
	key_value.sv_rec_id = rec_id;
	key_value.sv_row_id = 0;
	key_value.sv_key = key_buf;
	key_value.sv_length = myxt_create_key_from_row(ind, key_buf, rec_buf, NULL);

	return idx_delete(ot, ind, &key_value);
}

xtPublic xtBool xt_idx_update_row_id(XTOpenTablePtr ot, XTIndexPtr ind, xtRecordID rec_id, xtRowID row_id, xtWord1 *rec_buf)
{
	off_t				current;
	XTIdxResultRec		result;
	XTIdxKeyValueRec	key_value;
	xtWord1				key_buf[XT_INDEX_MAX_KEY_SIZE + XT_MAX_RECORD_REF_SIZE];

#ifdef CHECK_AND_PRINT
	idx_print_index(ot, ind, TRUE);
#endif
	key_value.sv_flags = XT_SEARCH_WHOLE_KEY;
	key_value.sv_rec_id = rec_id;
	key_value.sv_row_id = 0;
	key_value.sv_key = key_buf;
	key_value.sv_length = myxt_create_key_from_row(ind, key_buf, rec_buf, NULL);

	/* NOTE: Only a read lock is required for this!! */
	xt_rwlock_rdlock(&ind->mi_rwlock);

	if (!(current = ind->mi_root))
		goto done_ok;

	while (current) {
		if (!xt_ind_read_block(ot->ot_ind_file, current, (xtWord1 *) &ot->ot_ind_rbuf))
			goto failed;
		ind->mi_scan_branch(ind, &ot->ot_ind_rbuf, &key_value, &result);
		if (result.sr_found || !result.sr_item.i_node_ref_size)
			break;
		current = result.sr_branch;
	}

	if (result.sr_found) {
		xt_set_record_ref(&ot->ot_ind_rbuf.tb_data[result.sr_item.i_item_offset + result.sr_item.i_item_size - XT_RECORD_REF_SIZE], key_value.sv_rec_id, row_id);
		if (!xt_ind_write(ot->ot_ind_file, current, offsetof(XTIdxBranchDRec, tb_data) + result.sr_item.i_total_size, (xtWord1 *) &ot->ot_ind_rbuf))
			goto failed;
	}

	done_ok:
	xt_rwlock_unlock(&ind->mi_rwlock);

#ifdef DEBUG
	//idx_print_index(ot, ind, TRUE);
	//idx_check_index(ot, ind);
	//idx_check_on_key(ot);
#endif
	return OK;

	failed:
	xt_rwlock_unlock(&ind->mi_rwlock);
	return FAILED;
}

xtPublic void xt_idx_prep_key(XTIndexPtr ind, register XTIdxSearchKeyPtr search_key, int flags, xtWord1 *in_key_buf, size_t in_key_length)
{
	search_key->sk_key_value.sv_flags = flags;
	search_key->sk_key_value.sv_rec_id = 0;
	search_key->sk_key_value.sv_row_id = 0;
	search_key->sk_key_value.sv_key = search_key->sk_key_buf;
	search_key->sk_key_value.sv_length = myxt_create_key_from_key(ind, search_key->sk_key_buf, in_key_buf, in_key_length);
	search_key->sk_on_key = FALSE;
}

xtPublic xtBool xt_idx_research(XTOpenTablePtr ot, XTIndexPtr ind)
{
	XTIdxSearchKeyRec search_key;

	search_key.sk_key_value.sv_flags = XT_SEARCH_WHOLE_KEY;
	xt_get_record_ref(&ot->ot_ind_rbuf.tb_data[ot->ot_ind_state.i_item_offset + ot->ot_ind_state.i_item_size - XT_RECORD_REF_SIZE], &search_key.sk_key_value.sv_rec_id, &search_key.sk_key_value.sv_row_id);
	search_key.sk_key_value.sv_key = search_key.sk_key_buf;
	search_key.sk_key_value.sv_length = ot->ot_ind_state.i_item_size - XT_RECORD_REF_SIZE;
	search_key.sk_on_key = FALSE;
	memcpy(search_key.sk_key_buf, &ot->ot_ind_rbuf.tb_data[ot->ot_ind_state.i_item_offset], search_key.sk_key_value.sv_length);
	return xt_idx_search(ot, ind, &search_key);
}

/*
 * Search for a given key and position the current pointer on the first
 * key in the list of duplicates. If the key is not found the current
 * pointer is placed at the first position after the key.
 */
xtPublic xtBool xt_idx_search(XTOpenTablePtr ot, XTIndexPtr ind, register XTIdxSearchKeyPtr search_key)
{
	IdxBranchStackRec	stack;
	off_t				current;
	XTIdxResultRec		result;

#ifdef DEBUG
	//idx_print_index(ot, ind, TRUE);
#endif

	/* Calling from recovery, this is not the case.
	 * But the index read does not require a transaction!
	 * Only insert requires this to check for duplicates.
	if (!ot->ot_thread->st_xact_data) {
		xt_register_xterr(XT_REG_CONTEXT, XT_ERR_NO_TRANSACTION);
		return FAILED;
	}
	*/

	idx_newstack(&stack);

	ot->ot_curr_rec_id = 0;
	ot->ot_curr_row_id = 0;

	xt_rwlock_rdlock(&ind->mi_rwlock);

	if (!(current = ind->mi_root))
		goto done_ok;

	while (current) {
		if (!xt_ind_read_block(ot->ot_ind_file, current, (xtWord1 *) &ot->ot_ind_rbuf))
			goto failed;
		ind->mi_scan_branch(ind, &ot->ot_ind_rbuf, &search_key->sk_key_value, &result);
		if (result.sr_found)
			/* If we have found the key in a node: */
			search_key->sk_on_key = TRUE;
		if (!result.sr_item.i_node_ref_size)
			break;
		if (!idx_push(&stack, current, &result.sr_item))
			goto failed;
		current = result.sr_branch;
	}

	if (result.sr_item.i_item_offset == result.sr_item.i_total_size) {
		IdxStackItemPtr node;

		/* We are at the end of a leaf node.
		 * Go up the stack to find the start poition of the next key.
		 * If we find none, then we are the end of the index.
		 */
		while ((node = idx_pop(&stack))) {
			if (node->i_pos.i_item_offset < node->i_pos.i_total_size) {
				xtRecordID rec_id;

				if (!xt_ind_read_block(ot->ot_ind_file, node->i_branch, (xtWord1 *) &ot->ot_ind_rbuf))
					goto failed;
				xt_get_record_ref(&ot->ot_ind_rbuf.tb_data[node->i_pos.i_item_offset + node->i_pos.i_item_size - XT_RECORD_REF_SIZE], &rec_id, &ot->ot_curr_row_id);
				ot->ot_curr_rec_id = rec_id;
				ot->ot_ind_state = node->i_pos;
				break;
			}
		}
	}
	else {
		ot->ot_curr_rec_id = result.sr_rec_id;
		ot->ot_curr_row_id = result.sr_row_id;
		ot->ot_ind_state = result.sr_item;
	}

	done_ok:
	xt_rwlock_unlock(&ind->mi_rwlock);

#ifdef DEBUG
	//idx_print_index(ot, ind, TRUE);
	//idx_check_index(ot, ind);
	//idx_check_on_key(ot);
#endif
	return OK;

	failed:
	xt_rwlock_unlock(&ind->mi_rwlock);
	return FAILED;
}

xtPublic xtBool xt_idx_search_prev(XTOpenTablePtr ot, XTIndexPtr ind, register XTIdxSearchKeyPtr search_key)
{
	IdxBranchStackRec	stack;
	off_t				current;
	XTIdxResultRec		result;

#ifdef DEBUG
	//idx_print_index(ot, ind, TRUE);
#endif

	if (!ot->ot_thread->st_xact_data) {
		xt_register_xterr(XT_REG_CONTEXT, XT_ERR_NO_TRANSACTION);
		return FAILED;
	}

	idx_newstack(&stack);

	ot->ot_curr_rec_id = 0;
	ot->ot_curr_row_id = 0;

	xt_rwlock_rdlock(&ind->mi_rwlock);

	if (!(current = ind->mi_root))
		goto done_ok;

	while (current) {
		if (!xt_ind_read_block(ot->ot_ind_file, current, (xtWord1 *) &ot->ot_ind_rbuf))
			goto failed;
		ind->mi_scan_branch(ind, &ot->ot_ind_rbuf, &search_key->sk_key_value, &result);
		if (result.sr_found)
			/* If we have found the key in a node: */
			search_key->sk_on_key = TRUE;
		if (!result.sr_item.i_node_ref_size)
			break;
		if (!idx_push(&stack, current, &result.sr_item))
			goto failed;
		current = result.sr_branch;
	}

	if (result.sr_item.i_item_offset == 0) {
		IdxStackItemPtr node;

		/* We are at the end of a leaf node.
		 * Go up the stack to find the start poition of the next key.
		 * If we find none, then we are the end of the index.
		 */
		while ((node = idx_pop(&stack))) {
			if (node->i_pos.i_item_offset > node->i_pos.i_node_ref_size) {
				if (!xt_ind_read_block(ot->ot_ind_file, node->i_branch, (xtWord1 *) &ot->ot_ind_rbuf))
					goto failed;
				result.sr_item = node->i_pos;
				ind->mi_prev_item(ind, &ot->ot_ind_rbuf, &result);
				goto record_found;
			}
		}
		goto done_ok;
	}

	/* We must just step once to the left in this leaf node... */
	ind->mi_prev_item(ind, &ot->ot_ind_rbuf, &result);

	record_found:
	ot->ot_curr_rec_id = result.sr_rec_id;
	ot->ot_curr_row_id = result.sr_row_id;
	ot->ot_ind_state = result.sr_item;

	done_ok:
	xt_rwlock_unlock(&ind->mi_rwlock);

#ifdef DEBUG
	//idx_print_index(ot, ind, TRUE);
	//idx_check_index(ot, ind);
	//idx_check_on_key(ot);
#endif
	return OK;

	failed:
	xt_rwlock_unlock(&ind->mi_rwlock);
	return FAILED;
}

/*
 * Copy the current index value to the record.
 */
xtPublic xtBool xt_idx_read(XTOpenTablePtr ot, XTIndexPtr ind, xtWord1 *rec_buf)
{
	xtWord1	*bitem;

#ifdef DEBUG
	//idx_check_on_key(ot);
#endif
	bitem = ot->ot_ind_rbuf.tb_data + ot->ot_ind_state.i_item_offset;
	myxt_create_row_from_key(ot, ind, bitem, ot->ot_ind_state.i_item_size - XT_RECORD_REF_SIZE, rec_buf);
	return OK;
}

xtPublic xtBool xt_idx_next(register XTOpenTablePtr ot, register XTIndexPtr ind, register XTIdxSearchKeyPtr search_key)
{
	XTIdxKeyValueRec	key_value;
	xtWord1				key_buf[XT_INDEX_MAX_KEY_SIZE];
	XTIdxResultRec		result;
	IdxBranchStackRec	stack;
	off_t				current;

	if (!ot->ot_ind_state.i_node_ref_size && ot->ot_ind_state.i_item_offset < ot->ot_ind_state.i_total_size) {
		key_value.sv_key = &ot->ot_ind_rbuf.tb_data[ot->ot_ind_state.i_item_offset];
		key_value.sv_length = ot->ot_ind_state.i_item_size - XT_RECORD_REF_SIZE;

		result.sr_item = ot->ot_ind_state;
		idx_next_branch_item(ind, &ot->ot_ind_rbuf, &result);
		if (result.sr_item.i_item_offset < result.sr_item.i_total_size)
			goto check_on_key;
	}

	key_value.sv_flags = XT_SEARCH_WHOLE_KEY;
	xt_get_record_ref(&ot->ot_ind_rbuf.tb_data[ot->ot_ind_state.i_item_offset + ot->ot_ind_state.i_item_size - XT_RECORD_REF_SIZE], &key_value.sv_rec_id, &key_value.sv_row_id);
	key_value.sv_key = key_buf;
	key_value.sv_length = ot->ot_ind_state.i_item_size - XT_RECORD_REF_SIZE;
	memcpy(key_buf, &ot->ot_ind_rbuf.tb_data[ot->ot_ind_state.i_item_offset], key_value.sv_length);

	idx_newstack(&stack);

	xt_rwlock_rdlock(&ind->mi_rwlock);

	if (!(current = ind->mi_root)) {
		xt_rwlock_unlock(&ind->mi_rwlock);
		return OK;
	}

	while (current) {
		if (!xt_ind_read_block(ot->ot_ind_file, current, (xtWord1 *) &ot->ot_ind_rbuf))
			goto failed;
		ind->mi_scan_branch(ind, &ot->ot_ind_rbuf, &key_value, &result);
		if (result.sr_item.i_node_ref_size) {
			if (result.sr_found) {
				/* If we have found the key in a node: */
				idx_next_branch_item(ind, &ot->ot_ind_rbuf, &result);

				/* Go down to the bottom: */
				while (current) {
					if (!idx_push(&stack, current, &result.sr_item))
						goto failed;
					current = result.sr_branch;
					if (!xt_ind_read_block(ot->ot_ind_file, current, (xtWord1 *) &ot->ot_ind_rbuf))
						goto failed;
					idx_first_branch_item(ind, &ot->ot_ind_rbuf, &result);
					if (!result.sr_item.i_node_ref_size)
						break;
				}

				/* Is the leaf not empty, then we are done... */
				break;
			}
		}
		else {
			/* We have reached the leaf. */
			if (result.sr_found)
				/* If we have found the key in a leaf: */
				idx_next_branch_item(ind, &ot->ot_ind_rbuf, &result);
			/* If we did not find the key (although we should have). Our
			 * position is automatically the next one.
			 */
			break;
		}
		if (!idx_push(&stack, current, &result.sr_item))
			goto failed;
		current = result.sr_branch;
	}

	/* Check the current position in a leaf: */
	if (result.sr_item.i_item_offset == result.sr_item.i_total_size) {
		/* At the end: */
		IdxStackItemPtr node;

		/* We are at the end of a leaf node.
		 * Go up the stack to find the start poition of the next key.
		 * If we find none, then we are the end of the index.
		 */
		while ((node = idx_pop(&stack))) {
			if (node->i_pos.i_item_offset < node->i_pos.i_total_size) {
				if (!xt_ind_read_block(ot->ot_ind_file, node->i_branch, (xtWord1 *) &ot->ot_ind_rbuf))
					goto failed;
				result.sr_item = node->i_pos;
				xt_get_res_record_ref(&ot->ot_ind_rbuf.tb_data[result.sr_item.i_item_offset + result.sr_item.i_item_size - XT_RECORD_REF_SIZE], &result);
				xt_rwlock_unlock(&ind->mi_rwlock);
				goto check_on_key;
			}
		}

		/* No more keys: */
		if (search_key)
			search_key->sk_on_key = FALSE;
		ot->ot_curr_rec_id = 0;
		ot->ot_curr_row_id = 0;
		xt_rwlock_unlock(&ind->mi_rwlock);
		return OK;
	}

	xt_rwlock_unlock(&ind->mi_rwlock);

	check_on_key:
	/* Still on key? */
	if (search_key && search_key->sk_on_key) {
		/* GOTCHA: As a short-cut I was using a length compare
		 * and a memcmp() here to check whether we as still on
		 * the original search key.
		 * This does not work because it does not take into account
		 * trialing spaces (which are ignored in comparison).
		 * So lengths can be different, but values still equal.
		 * 
		 * NOTE: We have to use the original search flags for
		 * this compare.
		 */
		search_key->sk_on_key = myxt_compare_key(ind, search_key->sk_key_value.sv_flags, search_key->sk_key_value.sv_length,
			search_key->sk_key_value.sv_key, &ot->ot_ind_rbuf.tb_data[result.sr_item.i_item_offset]) == 0;
	}
	ot->ot_curr_rec_id = result.sr_rec_id;
	ot->ot_curr_row_id = result.sr_row_id;
	ot->ot_ind_state = result.sr_item;

	return OK;

	failed:
	xt_rwlock_unlock(&ind->mi_rwlock);
	return FAILED;
}

xtPublic xtBool xt_idx_prev(XTOpenTablePtr ot, XTIndexPtr ind, register XTIdxSearchKeyPtr search_key)
{
	XTIdxKeyValueRec	key_value;
	xtWord1				key_buf[XT_INDEX_MAX_KEY_SIZE];
	XTIdxResultRec		result;
	IdxBranchStackRec	stack;
	off_t				current;
	IdxStackItemPtr		node;

	if (!ot->ot_ind_state.i_node_ref_size && ot->ot_ind_state.i_item_offset > 0) {
		key_value.sv_key = &ot->ot_ind_rbuf.tb_data[ot->ot_ind_state.i_item_offset];
		key_value.sv_length = ot->ot_ind_state.i_item_size - XT_RECORD_REF_SIZE;

		result.sr_item = ot->ot_ind_state;
		ind->mi_prev_item(ind, &ot->ot_ind_rbuf, &result);
		goto check_on_key;
	}

	key_value.sv_flags = XT_SEARCH_WHOLE_KEY;
	key_value.sv_rec_id = ot->ot_curr_rec_id;
	key_value.sv_row_id = 0;
	key_value.sv_key = key_buf;
	key_value.sv_length = ot->ot_ind_state.i_item_size - XT_RECORD_REF_SIZE;
	memcpy(key_buf, &ot->ot_ind_rbuf.tb_data[ot->ot_ind_state.i_item_offset], key_value.sv_length);

	idx_newstack(&stack);

	xt_rwlock_rdlock(&ind->mi_rwlock);

	if (!(current = ind->mi_root)) {
		xt_rwlock_unlock(&ind->mi_rwlock);
		return OK;
	}

	while (current) {
		if (!xt_ind_read_block(ot->ot_ind_file, current, (xtWord1 *) &ot->ot_ind_rbuf))
			goto failed;
		ind->mi_scan_branch(ind, &ot->ot_ind_rbuf, &key_value, &result);
		if (result.sr_item.i_node_ref_size) {
			if (result.sr_found) {
				/* If we have found the key in a node: */

				/* Go down to the bottom: */
				while (current) {
					if (!idx_push(&stack, current, &result.sr_item))
						goto failed;
					current = result.sr_branch;
					if (!xt_ind_read_block(ot->ot_ind_file, current, (xtWord1 *) &ot->ot_ind_rbuf))
						goto failed;
					ind->mi_last_item(ind, &ot->ot_ind_rbuf, &result);
					if (!result.sr_item.i_node_ref_size)
						break;
				}

				/* Is the leaf not empty, then we are done... */
				if (result.sr_item.i_total_size == 0)
					break;
				goto unlock_check_on_key;
			}
		}
		else {
			/* We have reached the leaf.
			 * Whether we found the key or not, we have
			 * to move one to the left.
			 */
			if (result.sr_item.i_item_offset == 0)
				break;
			ind->mi_prev_item(ind, &ot->ot_ind_rbuf, &result);
			goto unlock_check_on_key;
		}
		if (!idx_push(&stack, current, &result.sr_item))
			goto failed;
		current = result.sr_branch;
	}

	/* We are at the start of a leaf node.
	 * Go up the stack to find the start poition of the next key.
	 * If we find none, then we are the end of the index.
	 */
	while ((node = idx_pop(&stack))) {
		if (node->i_pos.i_item_offset > node->i_pos.i_node_ref_size) {
			if (!xt_ind_read_block(ot->ot_ind_file, node->i_branch, (xtWord1 *) &ot->ot_ind_rbuf))
				goto failed;
			result.sr_item = node->i_pos;
			ind->mi_prev_item(ind, &ot->ot_ind_rbuf, &result);
			goto unlock_check_on_key;
		}
	}

	/* No more keys: */
	if (search_key)
		search_key->sk_on_key = FALSE;
	ot->ot_curr_rec_id = 0;
	ot->ot_curr_row_id = 0;

	xt_rwlock_unlock(&ind->mi_rwlock);
	return OK;

	unlock_check_on_key:
	xt_rwlock_unlock(&ind->mi_rwlock);

	check_on_key:
	/* Still on key? */
	if (search_key && search_key->sk_on_key) {
		search_key->sk_on_key = myxt_compare_key(ind, search_key->sk_key_value.sv_flags, search_key->sk_key_value.sv_length,
			search_key->sk_key_value.sv_key, &ot->ot_ind_rbuf.tb_data[result.sr_item.i_item_offset]) == 0;
	}
	ot->ot_curr_rec_id = result.sr_rec_id;
	ot->ot_curr_row_id = result.sr_row_id;
	ot->ot_ind_state = result.sr_item;
	return OK;

	failed:
	xt_rwlock_unlock(&ind->mi_rwlock);
	return FAILED;
}

/* Return TRUE if the record matches the current index search! */
xtPublic xtBool xt_idx_match_search(register XTOpenTablePtr ot, register XTIndexPtr ind, register XTIdxSearchKeyPtr search_key, xtWord1 *buf, int mode)
{
	int		r;
	xtWord1	key_buf[XT_INDEX_MAX_KEY_SIZE];

	myxt_create_key_from_row(ind, key_buf, (xtWord1 *) buf, NULL);
	r = myxt_compare_key(ind, search_key->sk_key_value.sv_flags, search_key->sk_key_value.sv_length, search_key->sk_key_value.sv_key, key_buf);
	switch (mode) {
		case XT_S_MODE_MATCH:
			return r == 0;
		case XT_S_MODE_NEXT:
			return r <= 0;
		case XT_S_MODE_PREV:
			return r >= 0;
	}
	return FALSE;
}

static void idx_set_index_selectivity(XTThreadPtr self, XTOpenTablePtr ot, XTIndexPtr ind)
{
	XTIdxSearchKeyRec	search_key;
	XTIndexSegPtr		key_seg;
	u_int				select_count = 0;
	xtWord1				key_buf[XT_INDEX_MAX_KEY_SIZE];
	u_int				key_len;
	xtWord1				*next_key_buf;
	u_int				next_key_len;
	u_int				curr_len;
	u_int				diff;
	u_int				i;

	ind->mi_select_total = 0;
	key_seg = ind->mi_seg;
	for (i=0; i < ind->mi_seg_count; key_seg++, i++) {
		key_seg->is_selectivity = 1;
		key_seg->is_recs_in_range = 1;
	}

	xt_idx_prep_key(ind, &search_key, XT_SEARCH_FIRST_FLAG, NULL, 0);
	if (!xt_idx_search(ot, ind, &search_key))
		return;

	/* Initialize the buffer with the first index valid index entry: */
	while (!select_count && ot->ot_curr_rec_id) {
		if (ot->ot_curr_row_id) {
			select_count++;

			key_len = ot->ot_ind_state.i_item_size - XT_RECORD_REF_SIZE;
			memcpy(key_buf, ot->ot_ind_rbuf.tb_data + ot->ot_ind_state.i_item_offset, key_len);
		}
		if (!xt_idx_next(ot, ind, &search_key))
			return;
	}

	while (select_count < 100 && ot->ot_curr_rec_id) {
		/* Check if the index entry is committed: */
		if (ot->ot_curr_row_id) {
			select_count++;

			next_key_len = ot->ot_ind_state.i_item_size - XT_RECORD_REF_SIZE;
			next_key_buf = ot->ot_ind_rbuf.tb_data + ot->ot_ind_state.i_item_offset;
		
			curr_len = 0;
			diff = FALSE;
			key_seg = ind->mi_seg;
			for (i=0; i < ind->mi_seg_count; key_seg++, i++) {
				curr_len += myxt_key_seg_length(key_seg, curr_len, key_buf);
				if (!diff && myxt_compare_key(ind, 0, curr_len, key_buf, next_key_buf) != 0)
					diff = i+1;
				if (diff)
					key_seg->is_selectivity++;
			}

			/* Store the key for the next comparison: */
			key_len = next_key_len;
			memcpy(key_buf, next_key_buf, key_len);
		}

		if (!xt_idx_next(ot, ind, &search_key))
			break;
	}

	if (select_count) {
		u_int recs;

		ind->mi_select_total = select_count;
		key_seg = ind->mi_seg;
		for (i=0; i < ind->mi_seg_count; key_seg++, i++) {
			recs = (u_int) ((double) select_count / (double) key_seg->is_selectivity + (double) 0.5);
			key_seg->is_recs_in_range = recs ? recs : 1;
		}
	}
}

xtPublic void xt_ind_set_index_selectivity(XTThreadPtr self, XTOpenTablePtr ot)
{
	XTTableHPtr		tab = ot->ot_table;
	XTIndexPtr		*ind;
	u_int			i;

	for (i=0, ind=tab->tab_dic.dic_keys; i<tab->tab_dic.dic_key_count; i++, ind++)
		idx_set_index_selectivity(self, ot, *ind);
}

/*
 * -----------------------------------------------------------------------
 * Print a b-tree
 */

static void idx_check_on_key(XTOpenTablePtr ot)
{
	u_int		offs = ot->ot_ind_state.i_item_offset + ot->ot_ind_state.i_item_size - XT_RECORD_REF_SIZE;
	xtRecordID	rec_id;
	xtRowID		row_id;
	
	if (ot->ot_curr_rec_id && ot->ot_ind_state.i_item_offset < ot->ot_ind_state.i_total_size) {
		xt_get_record_ref(&ot->ot_ind_rbuf.tb_data[offs], &rec_id, &row_id);
		
		ASSERT_NS(rec_id == ot->ot_curr_rec_id);
	}
}

static void idx_print_space(int depth)
{
	for (int i=0; i<depth; i++)
		printf(". ");
}

static void idx_print_node(XTOpenTablePtr ot, XTIndexPtr ind, int depth, off_t node)
{
	xtWord1			nbuf[XT_INDEX_PAGE_SIZE];
	XTIdxBranchDPtr	branch = (XTIdxBranchDPtr) nbuf;
	XTIdxResultRec	result;

	if (!xt_ind_read_block(ot->ot_ind_file, node, (xtWord1 *) branch))
		return;
	idx_first_branch_item(ind, branch, &result);
	if (result.sr_item.i_node_ref_size) {
		idx_print_space(depth);
		printf("%04llu -->\n", (u_llong) result.sr_branch);
		idx_print_node(ot, ind, depth+1, result.sr_branch);
	}

	while (result.sr_item.i_item_offset < result.sr_item.i_total_size) {
		idx_print_space(depth);
		printf("%llu# row=%lu ", (u_llong) result.sr_rec_id, (u_long) result.sr_row_id);
		myxt_print_key(ind, &branch->tb_data[result.sr_item.i_item_offset]);
		printf("\n");
		idx_next_branch_item(ind, branch, &result);
		if (result.sr_item.i_node_ref_size) {
			idx_print_space(depth);
			printf("%04llu -->\n", (u_llong) result.sr_branch);
			idx_print_node(ot, ind, depth+1, result.sr_branch);
		}
	}
}

static void idx_print_index(XTOpenTablePtr ot, XTIndexPtr ind, xtBool with_lock)
{
	register XTTableHPtr	tab = ot->ot_table;
	off_t					current;
	IdxFreeBlockRec			free_block;

	if (with_lock)
		xt_rwlock_wrlock(&ind->mi_rwlock);

	printf("INDEX %04llu -----------------------\n", (u_llong) ind->mi_root);
	if ((current = ind->mi_root))
		idx_print_node(ot, ind, 0, current);

	if (with_lock)
		xt_rwlock_unlock(&ind->mi_rwlock);

	xt_lock_mutex_ns(&tab->tab_ind_lock);
	printf("\nFREE: ");
	current = tab->tab_ind_free;
	while (current) {
		printf("%04llu ", (u_llong) current);
		if (!xt_ind_read(ot->ot_ind_file, current, sizeof(IdxFreeBlockRec), (xtWord1 *) &free_block)) {
			xt_log_and_clear_exception_ns();
			break;
		}
		current = XT_GET_DISK_8(free_block.if_next_block_8);
	}
	printf("\n-----------------------------\n");
	xt_unlock_mutex_ns(&tab->tab_ind_lock);
}

static void idx_check_node(XTOpenTablePtr ot, XTIndexPtr ind, int depth, off_t node)
{
	xtWord1			nbuf[XT_INDEX_PAGE_SIZE];
	XTIdxBranchDPtr	branch = (XTIdxBranchDPtr) nbuf;
	XTIdxResultRec	result;

	ASSERT_NS(node <= ot->ot_table->tab_ind_eof);
	if (!xt_ind_read_block(ot->ot_ind_file, node, (xtWord1 *) branch))
		return;
	idx_first_branch_item(ind, branch, &result);
	ASSERT_NS(result.sr_item.i_total_size + offsetof(XTIdxBranchDRec, tb_data) <= XT_INDEX_PAGE_SIZE);
	if (result.sr_item.i_node_ref_size) {
		idx_check_node(ot, ind, depth+1, result.sr_branch);
	}

	while (result.sr_item.i_item_offset < result.sr_item.i_total_size) {
		idx_next_branch_item(ind, branch, &result);
		if (result.sr_item.i_node_ref_size)
			idx_check_node(ot, ind, depth+1, result.sr_branch);
	}
}

static void idx_check_index(XTOpenTablePtr ot, XTIndexPtr ind)
{
	register XTTableHPtr	tab = ot->ot_table;
	off_t					current;
	IdxFreeBlockRec			free_block;

	xt_rwlock_wrlock(&ind->mi_rwlock);

	if ((current = ind->mi_root))
		idx_check_node(ot, ind, 0, current);

	xt_rwlock_unlock(&ind->mi_rwlock);

	xt_lock_mutex_ns(&tab->tab_ind_lock);
	current = tab->tab_ind_free;
	while (current) {
		if (!xt_ind_read(ot->ot_ind_file, current, sizeof(IdxFreeBlockRec), (xtWord1 *) &free_block)) {
			xt_log_and_clear_exception_ns();
			break;
		}
		current = XT_GET_DISK_8(free_block.if_next_block_8);
	}
	xt_unlock_mutex_ns(&tab->tab_ind_lock);
}

xtPublic void xt_print_index(XTOpenTablePtr ot, XTIndexPtr ind, xtBool with_lock)
{
	idx_print_index(ot, ind, with_lock);
}

xtPublic void xt_print_indices(XTOpenTablePtr ot)
{
	XTIndexPtr	*ind;

	ind = ot->ot_table->tab_dic.dic_keys;
	for (u_int i=0; i<ot->ot_table->tab_dic.dic_key_count; i++, ind++) {
		xt_print_index(ot, *ind, TRUE);
	}
}


