/* Copyright (c) 2007 PrimeBase Technologies GmbH
 *
 * BLOB Streaming for MySQL
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Paul McCullagh
 * H&G2JCtL
 *
 * 2007-06-01
 *
 * This file contains the BLOB streaming interface engines that
 * are streaming enabled.
 *
 */
#ifndef __streaming_unx_h__
#define __streaming_unx_h__

#include <stdio.h>
#include <sys/types.h>
#include <unistd.h>
#include <stdlib.h>
#include <fcntl.h>
#include <string.h>
#include <dirent.h>
#include <signal.h>
#include <ctype.h>

#ifdef USE_PRAGMA_INTERFACE
#pragma interface			/* gcc class implementation */
#endif

#define BS_SHARED_MEMORY_MAGIC			0x7E9A120C
#define BS_ENGINE_VERSION				1
#define BS_CALLBACK_VERSION				1
#define BS_SHARED_MEMORY_VERSION		1
#define BS_ENGINE_LIST_SIZE				80
#define BS_TEMP_FILE_PREFIX				"mybs_temp_"

#define BS_RESULT_MESSAGE_SIZE			300
#define BS_RESULT_STACK_SIZE			200

#define BS_BLOB_HANDLE_SIZE				300

#define SH_MASK							((S_IRUSR | S_IWUSR) | (S_IRGRP | S_IWGRP) | (S_IROTH))

#define BS_OK							0
#define BS_ERR_ENGINE					1							/* Internal engine error. */
#define BS_ERR_UNKNOWN_TABLE			2							/* Returned if the engine cannot open the given table. */
#define BS_ERR_NOT_FOUND				3							/* The BLOB cannot be found. */
#define BS_ERR_TABLE_LOCKED				4							/* Table is currently locked. */
#define BS_ERR_INCORRECT_URL			5
#define BS_ERR_AUTH_FAILED				6
#define BS_ERR_NOT_IMPLEMENTED			7
#define BS_ERR_UNKNOWN_DB				8
#define BS_ERR_REMOVING_REPO			9

#define BS_LOCK_NONE					0
#define BS_LOCK_READONLY				1
#define BS_LOCK_READ_WRITE				2

#define BS_XACT_NONE					0
#define BS_XACT_BEGIN					1
#define BS_XACT_COMMIT					2
#define BS_XACT_ROLLBACK				3

#define MYBS_ENGINE_REF_LEN				8
#define MYBS_BLOB_URL_SIZE				200

#define MYBS_FIELD_COL_SIZE				128
#define MYBS_FIELD_COND_SIZE			300

typedef struct MyBSResultRec {
	int						mr_code;								/* Engine specific error code. */ 
	char					mr_message[BS_RESULT_MESSAGE_SIZE];		/* Error message, required if non-zero return code. */
	char					mr_stack[BS_RESULT_STACK_SIZE];			/* Trace information about where the error occurred. */
} MyBSResultRec, *MyBSResultPtr;

typedef struct MyBSEngineRefRec {
	unsigned char			er_data[MYBS_ENGINE_REF_LEN];
} MyBSEngineRefRec, *MyBSEngineRefPtr;

typedef struct MyBSBlobURL {
	char					bu_data[MYBS_BLOB_URL_SIZE];
} MyBSBlobURLRec, *MyBSBlobURLPtr;

typedef struct MyBSFieldRef {
	char					fr_column[MYBS_FIELD_COL_SIZE];
	char					fr_cond[MYBS_FIELD_COND_SIZE];
} MyBSFieldRefRec, *MyBSFieldRefPtr;
/*
 * The engine must free its resources for the given thread.
 */
typedef void (*BSCloseConnFunc)(void *thd);

/* Before access BLOBs of a table, the streaming engine will open the table.
 * Open tables are managed as a pool by the streaming engine.
 * When a request is received, the streaming engine will ask all
 * registered engine to open the table. The engine must return a NULL
 * open_table pointer if it does not handle the table.
 * A callback allows an engine to request all open tables to be
 * closed by the streaming engine.
 */
typedef int (*BSOpenTableFunc)(void *thd, const char *table_url, void **open_table, MyBSResultPtr result);
typedef void (*BSCloseTableFunc)(void *thd, void *open_table);

/*
 * When the streaming engine wants to use an open table handle from the
 * pool, it calls the lock table function.
 */ 
typedef int (*BSLockTableFunc)(void *thd, int *xact, void *open_table, int lock_type, MyBSResultPtr result);
typedef int (*BSUnlockTableFunc)(void *thd, int xact, void *open_table, MyBSResultPtr result);

/* This function is used to locate and send a BLOB on the given stream.
 */
typedef int (*BSSendBLOBFunc)(void *thd, void *open_table, const char *blob_column, const char *blob_url, void *stream, MyBSResultPtr result);

/*
 * Lookup and engine reference, and return readable text.
 */
typedef int (*BSLookupRefFunc)(void *thd, void *open_table, unsigned short col_index, MyBSEngineRefPtr eng_ref, MyBSFieldRefPtr feild_ref, MyBSResultPtr result);

typedef struct MyBSEngineRec {
	int						bs_version;							/* BS_ENGINE_VERSION */
	int						bs_index;							/* The index into the engine list. */
	int						bs_removing;						/* TRUE (1) if the engine is being removed. */
	char					*bs_engine_name;
	void					*bs_engine_info;
	BSCloseConnFunc			bs_close_conn;
	BSOpenTableFunc			bs_open_table;
	BSCloseTableFunc		bs_close_table;
	BSLockTableFunc			bs_lock_table;
	BSUnlockTableFunc		bs_unlock_table;
	BSSendBLOBFunc			bs_send_blob;
	BSLookupRefFunc			bs_lookup_ref;
} MyBSEngineRec, *MyBSEnginePtr;

/*
 * This function should never be called directly, it is called
 * by deregisterEngine() below.
 */
typedef void (*ECDeregisterdFunc)(MyBSEnginePtr engine);

typedef void (*ECTableCloseAllFunc)(const char *table_url);

typedef int (*ECSetContentLenFunc)(void *stream, off_t len, MyBSResultPtr result);

typedef int (*ECWriteHeadFunc)(void *stream, MyBSResultPtr result);

typedef int (*ECWriteStreamFunc)(void *stream, void *buffer, size_t len, MyBSResultPtr result);

/*
 * The engine should call this function from
 * its own close connection function!
 */
typedef int (*ECCloseConnFunc)(void *thd, MyBSResultPtr result);

/*
 * Call this function before retaining or releasing BLOBs in a row.
 */
typedef int (*ECOpenTableFunc)(void **open_table, char *table_path, MyBSResultPtr result);

/*
 * Call this function when the operation is complete.
 */
typedef void (*ECCloseTableFunc)(void *open_table);

/*
 * Call this function for each BLOB to be retained. When a BLOB is used, the 
 * URL may be changed. The returned URL is valid as long as the the
 * table is open.
 *
 * The returned URL must be inserted into the row in place of the given
 * URL.
 */
typedef int (*ECUseBlobFunc)(void *open_table, char **ret_blob_url, char *blob_url, unsigned short col_index, MyBSResultPtr result);

/*
 * Reference Blobs that has been uploaded to the streaming engine.
 *
 * All BLOBs specified by the use blob function are retained by
 * this function.
 *
 * The engine reference is a (unaligned) 8 byte value which
 * identifies the row that the BLOBs are in.
 */
typedef int (*ECRetainBlobsFunc)(void *open_table, MyBSEngineRefPtr eng_ref, MyBSResultPtr result);

/*
 * If a row containing a BLOB is deleted, then the BLOBs in the
 * row must be released.
 *
 * Note: if a table is dropped, all the BLOBs referenced by the
 * table are automatically released.
 */
typedef int (*ECReleaseBlobFunc)(void *open_table, char *blob_url, unsigned short col_index, MyBSEngineRefPtr eng_ref, MyBSResultPtr result);

typedef int (*ECDropTable)(const char *table_path, MyBSResultPtr result);

typedef int (*ECRenameTable)(const char *from_table, const char *to_table, MyBSResultPtr result);

typedef struct MyBSCallbacksRec {
	int						cb_version;							/* BS_CALLBACK_VERSION */
	ECDeregisterdFunc		cb_deregister;
	ECTableCloseAllFunc		cb_table_close_all;
	ECSetContentLenFunc		cb_set_cont_len;
	ECWriteHeadFunc			cb_write_head;
	ECWriteStreamFunc		cb_write_stream;
	ECCloseConnFunc			cb_close_conn;
	ECOpenTableFunc			cb_open_table;
	ECCloseTableFunc		cb_close_table;
	ECUseBlobFunc			cb_use_blob;
	ECRetainBlobsFunc		cb_retain_blobs;
	ECReleaseBlobFunc		cb_release_blob;
	ECDropTable				cb_drop_table;
	ECRenameTable			cb_rename_table;
} MyBSCallbacksRec, *MyBSCallbacksPtr;

typedef struct MyBSSharedMemoryRec {
	int						sm_magic;							/* BS_SHARED_MEMORY_MAGIC */
	int						sm_version;							/* BS_SHARED_MEMORY_VERSION */
	volatile int			sm_shutdown_lock;					/* "Cheap" lock for shutdown! */
	MyBSCallbacksPtr		sm_callbacks;
	int						sm_reserved1[20];
	void					*sm_reserved2[20];
	int						sm_list_size;
	int						sm_list_len;
	MyBSEnginePtr			sm_engine_list[BS_ENGINE_LIST_SIZE];
} MyBSSharedMemoryRec, *MyBSSharedMemoryPtr;

#ifndef MYBS_API
Please define he value of MYBS_API
#endif

class MYBS_API
{
public:
	MYBS_API(): sharedMemory(NULL) { }

	~MYBS_API() { }

	/*
	 * Register the engine with the Streaming Engine.
	 */
	int registerEngine(MyBSEnginePtr engine, MyBSResultPtr result) {
		int err;

		deleteTempFiles();

		if ((err = getSharedMemory(true, result)))
			return err;

		for (int i=0; i<sharedMemory->sm_list_size; i++) {
			if (!sharedMemory->sm_engine_list[i]) {
				sharedMemory->sm_engine_list[i] = engine;
				engine->bs_index = i;
				if (i >= sharedMemory->sm_list_len)
					sharedMemory->sm_list_len = i+1;
				return BS_OK;
			}
		}
		
		result->mr_code = 15010;
		strcpy(BS_RESULT_MESSAGE_SIZE, result->mr_message, "Too many BLOB streaming engines already registered");
		*result->mr_stack = 0;
		return BS_ERR_ENGINE;
	}

	void lock() {
		while (sharedMemory->sm_shutdown_lock)
			usleep(10000);
		sharedMemory->sm_shutdown_lock++;
		while (sharedMemory->sm_shutdown_lock != 1) {
			usleep(random() % 10000);
			sharedMemory->sm_shutdown_lock--;
			usleep(10000);
			sharedMemory->sm_shutdown_lock++;
		}
	}

	void unlock() {
		sharedMemory->sm_shutdown_lock--;
	}

	void deregisterEngine(MyBSEnginePtr engine) {
		MyBSResultRec result;
		int err;

		if ((err = getSharedMemory(true, &result)))
			return;

		lock();

		bool empty = true;
		for (int i=0; i<sharedMemory->sm_list_len; i++) {
			if (sharedMemory->sm_engine_list[i]) {
				if (sharedMemory->sm_engine_list[i] == engine) {
					if (sharedMemory->sm_callbacks)
						sharedMemory->sm_callbacks->cb_deregister(engine);
					sharedMemory->sm_engine_list[i] = NULL;
				}
				else
					empty = false;
			}
		}

		unlock();

		if (empty) {
			char	temp_file[100];

			sharedMemory->sm_magic = 0;
			free(sharedMemory);
			sharedMemory = NULL;
			getTempFileName(temp_file, getpid());
			unlink(temp_file);
		}
	}

	void closeAllTables(const char *table_url)
	{
		MyBSResultRec	result;
		int					err;

		if ((err = getSharedMemory(true, &result)))
			return;

		if (sharedMemory->sm_callbacks)
			sharedMemory->sm_callbacks->cb_table_close_all(table_url);
	}

	int setContentLength(void *stream, off_t len, MyBSResultPtr result)
	{
		int err;

		if ((err = getSharedMemory(true, result)))
			return err;

		return sharedMemory->sm_callbacks->cb_set_cont_len(stream, len, result);
	}

	int writeHead(void *stream, MyBSResultPtr result)
	{
		int err;

		if ((err = getSharedMemory(true, result)))
			return err;

		return sharedMemory->sm_callbacks->cb_write_head(stream, result);
	}

	int writeStream(void *stream, void *buffer, size_t len, MyBSResultPtr result)
	{
		int err;

		if ((err = getSharedMemory(true, result)))
			return err;

		return sharedMemory->sm_callbacks->cb_write_stream(stream, buffer, len, result);
	}

	int closeConn(void *thd, MyBSResultPtr result)
	{
		int err;

		if ((err = getSharedMemory(true, result)))
			return err;

		if (!sharedMemory->sm_callbacks)
			return BS_OK;

		return sharedMemory->sm_callbacks->cb_close_conn(thd, result);
	}

	int openTable(void **open_table, char *table_path, MyBSResultPtr result)
	{
		int err;

		if ((err = getSharedMemory(true, result)))
			return err;

		if (!sharedMemory->sm_callbacks) {
			*open_table = NULL;
			return BS_OK;
		}

		return sharedMemory->sm_callbacks->cb_open_table(open_table, table_path, result);
	}

	int closeTable(void *open_table, MyBSResultPtr result)
	{
		int err;

		if ((err = getSharedMemory(true, result)))
			return err;

		if (sharedMemory->sm_callbacks && open_table)
			sharedMemory->sm_callbacks->cb_close_table(open_table);
		return BS_OK;
	}

	int couldBeURL(char *blob_url)
	/* ~*test/~1-150-2b5e0a7-0[.ext] */
	/* ~*test/_1-150-2b5e0a7-0[.ext] */
	{
		char	*ptr;
		size_t	len;

		if (blob_url) {
			if ((len = strlen(blob_url))) {
				/* Too short: */
				if (len <= 10)
					return 0;

				/* Required prefix: */
				/* NOTE: ~> is deprecated v0.5.4+, now use ~* */
				if (*blob_url != '~' || (*(blob_url + 1) != '>' && *(blob_url + 1) != '*'))
					return 0;

				ptr = blob_url + len - 1;

				/* Allow for an optional extension: */
				if (!isdigit(*ptr)) {
					while (ptr > blob_url && *ptr != '/' && *ptr != '.')
						ptr--;
					if (ptr == blob_url || *ptr != '.')
						return 0;
					if (ptr == blob_url || !isdigit(*ptr))
						return 0;
				}
	
				while (ptr > blob_url && isdigit(*ptr))
					ptr--;

				if (ptr == blob_url || *ptr != '-')
					return 0;
				ptr--;
				if (!isxdigit(*ptr))
					return 0;

				while (ptr > blob_url && isxdigit(*ptr))
					ptr--;

				if (ptr == blob_url || *ptr != '-')
					return 0;
				ptr--;
				if (!isdigit(*ptr))
					return 0;
					
				while (ptr > blob_url && isdigit(*ptr))
					ptr--;

				if (ptr == blob_url || *ptr != '-')
					return 0;
				ptr--;
				if (!isdigit(*ptr))
					return 0;

				while (ptr > blob_url && isdigit(*ptr))
					ptr--;

				/* NOTE: ^ and : are deprecated v0.5.4+, now use ! and ~ */
				if (ptr == blob_url || (*ptr != '^' && *ptr != ':' && *ptr != '_' && *ptr != '~'))
					return 0;
				ptr--;

				if (ptr == blob_url || *ptr != '/')
					return 0;
				ptr--;
				if (ptr == blob_url)
					return 0;
				return 1;
			}
		}
		return 0;
	}

	int useBlob(void *open_table, char **ret_blob_url, char *blob_url, unsigned short col_index, MyBSResultPtr result)
	{
		int err;

		if ((err = getSharedMemory(true, result)))
			return err;

		if (!couldBeURL(blob_url)) {
			*ret_blob_url = NULL;
			return BS_OK;
		}

		if (!sharedMemory->sm_callbacks) {
			result->mr_code = BS_ERR_INCORRECT_URL;
			strcpy(BS_RESULT_MESSAGE_SIZE, result->mr_message, "BLOB streaming engine (MyBS) not installed");
			*result->mr_stack = 0;
			return BS_ERR_INCORRECT_URL;
		}

		return sharedMemory->sm_callbacks->cb_use_blob(open_table, ret_blob_url, blob_url, col_index, result);
	}

	int retainBlobs(void *open_table, MyBSEngineRefPtr eng_ref, MyBSResultPtr result)
	{
		int err;

		if ((err = getSharedMemory(true, result)))
			return err;

		if (!sharedMemory->sm_callbacks)
			return BS_OK;
		return sharedMemory->sm_callbacks->cb_retain_blobs(open_table, eng_ref, result);
	}

	int releaseBlob(void *open_table, char *blob_url, unsigned short col_index, MyBSEngineRefPtr eng_ref, MyBSResultPtr result)
	{
		int err;

		if ((err = getSharedMemory(true, result)))
			return err;

		if (!sharedMemory->sm_callbacks)
			return BS_OK;

		if (!couldBeURL(blob_url))
			return BS_OK;

		return sharedMemory->sm_callbacks->cb_release_blob(open_table, blob_url, col_index, eng_ref, result);
	}

	int dropTable(const char *table_path, MyBSResultPtr result)
	{
		int err;

		if ((err = getSharedMemory(true, result)))
			return err;

		if (!sharedMemory->sm_callbacks)
			return BS_OK;
			
		return sharedMemory->sm_callbacks->cb_drop_table(table_path, result);
	}

	int renameTable(const char *from_table, const char *to_table, MyBSResultPtr result)
	{
		int err;

		if ((err = getSharedMemory(true, result)))
			return err;

		if (!sharedMemory->sm_callbacks)
			return BS_OK;
			
		return sharedMemory->sm_callbacks->cb_rename_table(from_table, to_table, result);
	}

	volatile MyBSSharedMemoryPtr sharedMemory;

private:
	int getSharedMemory(bool create, MyBSResultPtr result)
	{
		int		tmp_f;
		int		r;
		char	temp_file[100];

		if (sharedMemory)
			return BS_OK;

		getTempFileName(temp_file, getpid());
		tmp_f = open(temp_file, O_RDWR | (create ? O_CREAT : 0), SH_MASK);
		if (tmp_f == -1)
			return setOSResult(errno, "open", temp_file, result);

		r = lseek(tmp_f, 0, SEEK_SET);
		if (r == -1) {
			close(tmp_f);
			return setOSResult(errno, "lseek", temp_file, result);
		}
		ssize_t tfer;
		char buffer[100];
		
		tfer = read(tmp_f, buffer, 100);
		if (tfer == -1) {
			close(tmp_f);
			return setOSResult(errno, "read", temp_file, result);
		}

		buffer[tfer] = 0;
		sscanf(buffer, "%p", &sharedMemory);
		if (!sharedMemory || sharedMemory->sm_magic != BS_SHARED_MEMORY_MAGIC) {
			if (!create)
				return BS_OK;

			sharedMemory = (MyBSSharedMemoryPtr) calloc(1, sizeof(MyBSSharedMemoryRec));
			sharedMemory->sm_magic = BS_SHARED_MEMORY_MAGIC;
			sharedMemory->sm_version = BS_SHARED_MEMORY_VERSION;
			sharedMemory->sm_list_size = BS_ENGINE_LIST_SIZE;

			r = lseek(tmp_f, 0, SEEK_SET);
			if (r == -1) {
				close(tmp_f);
				return setOSResult(errno, "fseek", temp_file, result);
			}

			sprintf(buffer, "%p", sharedMemory);
			tfer = write(tmp_f, buffer, strlen(buffer));
			if (tfer != strlen(buffer)) {
				close(tmp_f);
				return setOSResult(errno, "write", temp_file, result);
			}
			r = fsync(tmp_f);
			if (r == -1) {
				close(tmp_f);
				return setOSResult(errno, "fsync", temp_file, result);
			}
		}
		else if (sharedMemory->sm_version != BS_SHARED_MEMORY_VERSION) {
			close(tmp_f);
			result->mr_code = -1000;
			*result->mr_stack = 0;
			strcpy(BS_RESULT_MESSAGE_SIZE, result->mr_message, "Shared memory version: ");		
			strcat(BS_RESULT_MESSAGE_SIZE, result->mr_message, sharedMemory->sm_version);		
			strcat(BS_RESULT_MESSAGE_SIZE, result->mr_message, ", does not match engine shared memory version: ");		
			strcat(BS_RESULT_MESSAGE_SIZE, result->mr_message, BS_SHARED_MEMORY_VERSION);		
			strcat(BS_RESULT_MESSAGE_SIZE, result->mr_message, ".");		
			return BS_ERR_ENGINE;
		}
		close(tmp_f);
		return BS_OK;
	}

	void strcpy(size_t size, char *to, const char *from)
	{
		if (size > 0) {
			size--;
			while (*from && size--)
				*to++ = *from++;
			*to = 0;
		}
	}

	void strcat(size_t size, char *to, const char *from)
	{
		while (*to && size--) to++;
		strcpy(size, to, from);
	}

	void strcat(size_t size, char *to, int val)
	{
		char buffer[100];

		sprintf(buffer, "%d", val);
		strcat(size, to, buffer);
	}

	int setOSResult(int err, char *func, char *file, MyBSResultPtr result) {
		char *msg;

		result->mr_code = err;
		*result->mr_stack = 0;
		strcpy(BS_RESULT_MESSAGE_SIZE, result->mr_message, "System call ");		
		strcat(BS_RESULT_MESSAGE_SIZE, result->mr_message, func);		
		strcat(BS_RESULT_MESSAGE_SIZE, result->mr_message, "() failed on ");		
		strcat(BS_RESULT_MESSAGE_SIZE, result->mr_message, file);		
		strcat(BS_RESULT_MESSAGE_SIZE, result->mr_message, ": ");		

#ifdef XT_WIN
		if (FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM, NULL, err, 0, iMessage + strlen(iMessage), BS_RESULT_MESSAGE_SIZE - strlen(iMessage), NULL)) {
			char *ptr;

			ptr = &iMessage[strlen(iMessage)];
			while (ptr-1 > err_msg) {
				if (*(ptr-1) != '\n' && *(ptr-1) != '\r' && *(ptr-1) != '.')
					break;
				ptr--;
			}
			*ptr = 0;

			strcat(BS_RESULT_MESSAGE_SIZE, result->mr_message, " (");
			strcat(BS_RESULT_MESSAGE_SIZE, result->mr_message, err);
			strcat(BS_RESULT_MESSAGE_SIZE, result->mr_message, ")");
			return BS_ERR_ENGINE;
		}
#endif

		msg = strerror(err);
		if (msg) {
			strcat(BS_RESULT_MESSAGE_SIZE, result->mr_message, msg);
			strcat(BS_RESULT_MESSAGE_SIZE, result->mr_message, " (");
			strcat(BS_RESULT_MESSAGE_SIZE, result->mr_message, err);
			strcat(BS_RESULT_MESSAGE_SIZE, result->mr_message, ")");
		}
		else {
			strcat(BS_RESULT_MESSAGE_SIZE, result->mr_message, "Unknown OS error code ");
			strcat(BS_RESULT_MESSAGE_SIZE, result->mr_message, err);
		}

		return BS_ERR_ENGINE;
	}

	void getTempFileName(char *temp_file, int pid)
	{
		sprintf(temp_file, "/tmp/" BS_TEMP_FILE_PREFIX "%d", pid);
	}

	bool startsWith(const char *cstr, const char *w_cstr)
	{
		while (*cstr && *w_cstr) {
			if (*cstr != *w_cstr)
				return false;
			cstr++;
			w_cstr++;
		}
		return *cstr || !*w_cstr;
	}

	void deleteTempFiles()
	{
		struct dirent	entry;
		struct dirent	*result;
		DIR				*odir;
		int				err;
		char			temp_file[100];

		if (!(odir = opendir("/tmp/")))
			return;
		err = readdir_r(odir, &entry, &result);
		while (!err && result) {
			if (startsWith(entry.d_name, BS_TEMP_FILE_PREFIX)) {
				int pid = atoi(entry.d_name + strlen(BS_TEMP_FILE_PREFIX));
				
				/* If the process does not exist: */
				if (kill(pid, 0) == -1 && errno == ESRCH) {
					getTempFileName(temp_file, pid);
					unlink(temp_file);
				}
			}
			
			err = readdir_r(odir, &entry, &result);
		}
		closedir(odir);
	}
};

#endif
