/* Copyright (c) 2005 PrimeBase Technologies GmbH
 *
 * PrimeBase XT
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * 2005-04-10	Paul McCullagh
 *
 * H&G2JCtL
 */

#include "xt_config.h"

#include <m_ctype.h>

#include <time.h>
#include <signal.h>

#include "xaction_xt.h"
#include "database_xt.h"
#include "strutil_xt.h"
#include "heap_xt.h"
#include "trace_xt.h"
#include "myxt_xt.h"
#include "tabcache_xt.h"

#ifdef DEBUG
//#define TRACE_VARIATIONS
#endif

static void xn_sw_wait_for_xact(XTThreadPtr self, XTDatabaseHPtr db, u_int hsecs);

/* ============================================================================================== */

typedef struct XNSWRecItem {
	xtTableID				ri_tab_id;
	xtRecordID				ri_rec_id;
} XNSWRecItemRec, *XNSWRecItemPtr;

typedef struct XNSWToFreeItem {
	xtTableID				ri_tab_id;			/* If non-zero, then this is the table of the data record to be freed.
												 * If zero, then this free the transaction below must be freed.
												 */
	union {
		xtRecordID			ri_rec_id;
		xtXactID			ri_xn_id;
	} x;
	xtXactID				ri_wait_xn_id;			/* Wait for this transaction to be cleaned (or being cleaned up)
												 * before freeing this resource. */
} XNSWToFreeItemRec, *XNSWToFreeItemPtr;

/* ----------------------------------------------------------------------
 * Cleanup Record list
 */

static int xn_compare_cu_item(XTThreadPtr self, register const void *thunk, register const void *a, register const void *b)
{
#pragma unused(self, thunk)
	XNSWRecItemPtr	x = (XNSWRecItemPtr) a;
	XNSWRecItemPtr	y = (XNSWRecItemPtr) b;

	if (x->ri_tab_id == y->ri_tab_id) {
		if (x->ri_rec_id == y->ri_rec_id)
			return 0;
		if (x->ri_rec_id < y->ri_rec_id)
			return -1;
		return 1;
	}
	if (x->ri_tab_id < y->ri_tab_id)
		return -1;
	return 1;
}

static void xn_free_cu_item(XTThreadPtr self, void *thunk, void *item)
{
#pragma unused(self, thunk, item)
}

/*
 * We create a list for the sweeper of records that need to be
 * cleaned up. This may happen if a record is written to the
 * database, and then the server crashes before the change
 * can be written and flushed to the log.
 *
 * If a record is not written to the log, it is not cleaned
 * up by the sweeper.
 *
 * Such records can be detected. The type of the record
 * indicates that the record has not been cleaned.
 * At the same time the transaction that wrote the record is
 * not in memory.
 *
 * When such a record is encountered, it is added to this list
 * and cleaned up by the sweeper.
 */
static void xn_add_cu_record(XTDatabaseHPtr db, xtTableID tab_id, xtRecordID rec_id)
{
	XNSWRecItemRec	cu_rec;

	cu_rec.ri_tab_id = tab_id;
	cu_rec.ri_rec_id = rec_id;

	xt_sl_lock(NULL, db->db_sw_cu_list);
	xt_sl_insert(NULL, db->db_sw_cu_list, &cu_rec, &cu_rec);
	xt_sl_unlock(NULL, db->db_sw_cu_list);
}

/*
 * Get the next cleanup record from the end of the list.
 */
static xtBool xn_get_cu_record(XTDatabaseHPtr db, xtTableID *tab_id, xtRecordID *rec_id)
{
	XNSWRecItemPtr cu_rec;

	xt_sl_lock(NULL, db->db_sw_cu_list);
	if ((cu_rec = (XNSWRecItemPtr) xt_sl_last_item(db->db_sw_cu_list))) {
		*tab_id = cu_rec->ri_tab_id;
		*rec_id = cu_rec->ri_rec_id;
		xt_sl_set_size(db->db_sw_cu_list, xt_sl_get_size(db->db_sw_cu_list) - 1);
	}
	xt_sl_unlock(NULL, db->db_sw_cu_list);
	return cu_rec != NULL;
}

/* ----------------------------------------------------------------------
 * Wait for transactions
 */

typedef struct XNWaitFor {
	xtXactID				wf_waiting_xn_id;		/* The transaction of the waiting thread. */
	xtXactID				wf_for_me_xn_id;		/* The transaction we are waiting for. */
} XNWaitForRec, *XNWaitForPtr;

static int xn_compare_wait_for(XTThreadPtr self, register const void *thunk, register const void *a, register const void *b)
{
#pragma unused(self, thunk)
	xtXactID		*x = (xtXactID *) a;
	XNWaitForPtr	y = (XNWaitForPtr) b;

	if (*x == y->wf_waiting_xn_id)
		return 0;
	if (xt_xn_is_before(*x, y->wf_waiting_xn_id))
		return -1;
	return 1;
}

static void xn_free_wait_for(XTThreadPtr self, void *thunk, void *item)
{
#pragma unused(self, thunk, item)
}

/*
 * A deadlock occurs when a transaction is waiting for itself!
 * For example A is waiting for B which is waiting for A.
 * By repeatedly scanning the wait_for list we can find out if a
 * transaction is waiting for itself.
 */
static xtBool xn_detect_deadlock(XTDatabaseHPtr db, xtXactID waiting, xtXactID for_me)
{
	XNWaitForPtr wf;

	for (;;) {
		if (waiting == for_me) {
			xt_register_xterr(XT_REG_CONTEXT, XT_ERR_DEADLOCK);
			return TRUE;
		}
		if (!(wf = (XNWaitForPtr) xt_sl_find(NULL, db->db_xn_wait_for, &for_me)))
			break;
		for_me = wf->wf_for_me_xn_id;
	}
	return FALSE;
}

/*
 * The given thread must wait for the specified transaction to terminate. This
 * function places the transaction of the thread on a list of waiting threads.
 *
 * Before waiting we make a check for deadlocks. A deadlock occurs
 * if waiting would introduce a cycle.
 */
xtPublic xtBool xt_xn_wait_for_xact(XTThreadPtr self, xtXactID xn_id, xtBool will_retry)
{
	XTDatabaseHPtr		db = self->st_database;
	XTXactDataPtr		xact;
	XNWaitForRec		wf;

	ASSERT(self->st_xact_data);

	wf.wf_waiting_xn_id = self->st_xact_data->xd_start_xn_id;
	wf.wf_for_me_xn_id = xn_id;

	xt_lock_mutex_ns(&db->db_xn_wait_lock);

	for (;;) {
		if (!(xact = xt_xn_get_xact(db, xn_id)))
			break;

		/* This is a dirty read, but it should work! */
		if ((xact->xd_flags & XT_XN_XAC_ENDED) || xact->xd_start_xn_id != xn_id)
			break;

		if (xn_detect_deadlock(db, wf.wf_waiting_xn_id, wf.wf_for_me_xn_id))
			goto failed;

		if (!xt_sl_insert(NULL, db->db_xn_wait_for, &wf.wf_waiting_xn_id, &wf))
			goto failed;

		if ((xact->xd_flags & XT_XN_XAC_ENDED) || xact->xd_start_xn_id != xn_id)
			break;

		/* Timed wait because it is possible that transaction quits before
		 * we go to sleep.
		 */
		if (!xt_timed_wait_cond(NULL, &db->db_xn_wait_cond, &db->db_xn_wait_lock, 2 * 1000)) {
			xt_sl_delete(self, db->db_xn_wait_for, &wf.wf_waiting_xn_id);
			goto failed;
		}

		xt_sl_delete(self, db->db_xn_wait_for, &wf.wf_waiting_xn_id);
		
		if (will_retry)
			break;
	}

	xt_unlock_mutex_ns(&db->db_xn_wait_lock);
	return OK;

	failed:
	xt_unlock_mutex_ns(&db->db_xn_wait_lock);
	return FAILED;
}

xtPublic void xt_xn_wakeup_transactions(XTDatabaseHPtr db, xtBool wait_for_sweeper)
{
	/* This flag makes the gap for the race condition
	 * very small.
	 *
	 * However, this posibility still remains because
	 * we do not lock the mutex db_xn_wait_lock here.
	 *
	 * The reason is that it is too expensive.
	 *
	 * In the event that the wakeup is missed the sleeper
	 * wait will timeout eventually.
	 */
	db->db_sw_wakeup = TRUE;
	if (!xt_broadcast_cond(NULL, &db->db_xn_wait_cond))
		xt_log_and_clear_exception_ns();
}

/* ----------------------------------------------------------------------
 * Utilities
 */

//#define HIGH_X
#ifdef HIGH_X
u_long tot_alloced;
u_long high_alloced;
u_long not_clean_max;
u_long in_ram_max;
#endif

static void xn_free_xact(XTDatabaseHPtr db, XTXactSegPtr seg, XTXactDataPtr xact)
{
#ifdef HIGH_X
	tot_alloced--;
#endif
	/* This indicates the structure is free: */
	xact->xd_start_xn_id = 0;
	if ((xtWord1 *) xact >= db->db_xn_data && (xtWord1 *) xact < db->db_xn_data_end) {
		/* Put it in the free list: */
		xact->xd_next_xact = seg->xs_free_list;
		seg->xs_free_list = xact;
		return;
	}
	xt_free_ns(xact);
}

/*
 * GOTCHA: The value db->db_xn_curr_id may be a bit larger
 * than the actual transaction created because there is
 * a gap between the issude of the transaction ID
 * and the creation of a memory structure.
 * (indicated here: [*])
 *
 * This function returns the actuall current transaction ID.
 * This is the number of the last transaction actually
 * created in memory.
 *
 * This means that if you call xt_xn_get_xact() with any
 * number less than or equal to this value, not finding
 * the transaction means it has already ended!
 */
xtPublic xtXactID xt_xn_get_curr_id(XTDatabaseHPtr db)
{
	int						i;
	xtXactID				curr_xn_id;
	register XTXactSegPtr 	seg = db->db_xn_idx;

	/* Find the highest transaction ID actually created... */
	curr_xn_id = seg->xt_last_xn_id;
	seg++;
	for (i=1; i<XT_XN_NO_OF_SEGMENTS; i++, seg++) {
		if (xt_xn_is_before(curr_xn_id, seg->xt_last_xn_id))
			curr_xn_id = seg->xt_last_xn_id;
	}
	return curr_xn_id;
}

xtPublic XTXactDataPtr xt_xn_add_old_xact(XTDatabaseHPtr db, xtXactID xn_id)
{
	register XTXactDataPtr	xact;
	register XTXactSegPtr 	seg;
	register XTXactDataPtr	*hash;

	seg = &db->db_xn_idx[xn_id & XT_XN_SEGMENT_MASK];
	xt_rwlock_wrlock(&seg->xs_tab_lock);
	hash = &seg->xs_table[(xn_id >> XT_XN_SEGMENT_SHIFTS) % XT_XN_HASH_TABLE_SIZE];
	xact = *hash;
	while (xact) {
		if (xact->xd_start_xn_id == xn_id)
			goto done_ok;
		xact = xact->xd_next_xact;
	}

	if ((xact = seg->xs_free_list))
		seg->xs_free_list = xact->xd_next_xact;
	else {
		/* We have used up all the free transaction slots,
		 * the sweeper should work faster to free them
		 * up...
		 */
		db->db_sw_faster = TRUE;
		if (!(xact = (XTXactDataPtr) xt_malloc_ns(sizeof(XTXactDataRec)))) {
			xt_rwlock_unlock(&seg->xs_tab_lock);
			return NULL;
		}
	}

	xact->xd_next_xact = *hash;
	*hash = xact;

	xact->xd_start_xn_id = xn_id;
	xact->xd_end_xn_id = 0;
	xact->xd_end_time = 0;
	xact->xd_begin_log = 0;
	xact->xd_flags = 0;

	/* Get the largest transaction id. */
	if (xt_xn_is_before(seg->xt_last_xn_id, xn_id))
		seg->xt_last_xn_id = xn_id;

	done_ok:
	xt_rwlock_unlock(&seg->xs_tab_lock);
#ifdef HIGH_X
	tot_alloced++;
	if (tot_alloced > high_alloced)
		high_alloced = tot_alloced;
#endif
	return xact;
}

static XTXactDataPtr xn_add_new_xact(XTDatabaseHPtr db, xtXactID xn_id)
{
	register XTXactDataPtr	xact;
	register XTXactSegPtr 	seg;
	register XTXactDataPtr	*hash;

	seg = &db->db_xn_idx[xn_id & XT_XN_SEGMENT_MASK];
	xt_rwlock_wrlock(&seg->xs_tab_lock);
	hash = &seg->xs_table[(xn_id >> XT_XN_SEGMENT_SHIFTS) % XT_XN_HASH_TABLE_SIZE];

	if ((xact = seg->xs_free_list))
		seg->xs_free_list = xact->xd_next_xact;
	else {
		/* We have used up all the free transaction slots,
		 * the sweeper should work faster to free them
		 * up...
		 */
		db->db_sw_faster = TRUE;
		if (!(xact = (XTXactDataPtr) xt_malloc_ns(sizeof(XTXactDataRec)))) {
			xt_rwlock_unlock(&seg->xs_tab_lock);
			return NULL;
		}
	}

	xact->xd_next_xact = *hash;
	*hash = xact;

	xact->xd_start_xn_id = xn_id;
	xact->xd_end_xn_id = 0;
	xact->xd_end_time = 0;
	xact->xd_begin_log = 0;
	xact->xd_flags = 0;

	seg->xt_last_xn_id = xn_id;
	xt_rwlock_unlock(&seg->xs_tab_lock);
#ifdef HIGH_X
	tot_alloced++;
	if (tot_alloced > high_alloced)
		high_alloced = tot_alloced;
#endif
	return xact;
}

xtPublic XTXactDataPtr xt_xn_get_xact(XTDatabaseHPtr db, xtXactID xn_id)
{
	register XTXactSegPtr 	seg;
	register XTXactDataPtr	xact;

	seg = &db->db_xn_idx[xn_id & XT_XN_SEGMENT_MASK];
	xt_rwlock_rdlock(&seg->xs_tab_lock);
	xact = seg->xs_table[(xn_id >> XT_XN_SEGMENT_SHIFTS) % XT_XN_HASH_TABLE_SIZE];
	while (xact) {
		if (xact->xd_start_xn_id == xn_id)
			break;
		xact = xact->xd_next_xact;
	}
	xt_rwlock_unlock(&seg->xs_tab_lock);
	return xact;
}

/*
 * Delete a transaction, return TRUE if the transaction
 * was found.
 */
xtPublic xtBool xt_xn_delete_xact(XTDatabaseHPtr db, xtXactID xn_id)
{
	XTXactDataPtr	xact, pxact = NULL;
	XTXactSegPtr 	seg;

	seg = &db->db_xn_idx[xn_id & XT_XN_SEGMENT_MASK];
	xt_rwlock_wrlock(&seg->xs_tab_lock);
	xact = seg->xs_table[(xn_id >> XT_XN_SEGMENT_SHIFTS) % XT_XN_HASH_TABLE_SIZE];
	while (xact) {
		if (xact->xd_start_xn_id == xn_id) {
			if (pxact)
				pxact->xd_next_xact = xact->xd_next_xact;
			else
				 seg->xs_table[(xn_id >> XT_XN_SEGMENT_SHIFTS) % XT_XN_HASH_TABLE_SIZE] = xact->xd_next_xact;
			xn_free_xact(db, seg, xact);
			xt_rwlock_unlock(&seg->xs_tab_lock);
			return TRUE;
		}
		pxact = xact;
		xact = xact->xd_next_xact;
	}
	xt_rwlock_unlock(&seg->xs_tab_lock);
	return FALSE;
}

//#define DEBUG_RAM_LIST
#ifdef DEBUG_RAM_LIST

#define DEBUG_RAM_LIST_SIZE			80

int					check_ram_init_count = 0;
xt_rwlock_type		check_ram_lock;
xtXactID			check_ram_trns[DEBUG_RAM_LIST_SIZE];
int					check_ram_dummy;

static void check_ram_init(void)
{
	if (check_ram_init_count == 0)
		xt_init_rwlock(NULL, &check_ram_lock);
	check_ram_init_count++;
}

static void check_ram_free(void)
{
	check_ram_init_count--;
	if (check_ram_init_count == 0)
		xt_free_rwlock(&check_ram_lock);
}

static void check_ram_min_id(XTDatabaseHPtr db)
{
	int i;

	xt_rwlock_rdlock(&check_ram_lock);
	for (i=0; i<DEBUG_RAM_LIST_SIZE; i++) {
		if (check_ram_trns[i] && xt_xn_is_before(check_ram_trns[i], db->db_xn_min_ram_id)) {
			/* This should never happen! */
			XTXactDataPtr x_ptr;

			check_ram_dummy = 0;
			for (i=0; i<DEBUG_RAM_LIST_SIZE; i++) {
				if (check_ram_trns[i]) {
					x_ptr = xt_xn_get_xact(db, check_ram_trns[i]);
					check_ram_dummy = 1;
				}
			}
			break;
		}
	}
	xt_rwlock_unlock(&check_ram_lock);
}

static void check_ram_add(xtXactID xn_id)
{
	int i;
	
	xt_rwlock_wrlock(&check_ram_lock);
	for (i=0; i<DEBUG_RAM_LIST_SIZE; i++) {
		if (!check_ram_trns[i]) {
			check_ram_trns[i] = xn_id;
			xt_rwlock_unlock(&check_ram_lock);
			return;
		}
	}
	xt_rwlock_unlock(&check_ram_lock);
	printf("DEBUG --- List too small\n");
}

static void check_ram_del(xtXactID xn_id)
{
	int i;
	
	xt_rwlock_wrlock(&check_ram_lock);
	for (i=0; i<DEBUG_RAM_LIST_SIZE; i++) {
		if (check_ram_trns[i] == xn_id) {
			check_ram_trns[i] = 0;
			xt_rwlock_unlock(&check_ram_lock);
			return;
		}
	}
	xt_rwlock_unlock(&check_ram_lock);
}
#endif

/* ----------------------------------------------------------------------
 * Init and Exit
 */

xtPublic void xt_xn_init_db(XTThreadPtr self, XTDatabaseHPtr db)
{
	XTXactDataPtr	xact;
	XTXactSegPtr	seg;

#ifdef DEBUG_RAM_LIST
	check_ram_init();
#endif
	xt_init_mutex(self, &db->db_xn_id_lock);
	xt_init_mutex(self, &db->db_xn_wait_lock);
	xt_init_cond(self, &db->db_xn_wait_cond);
	xt_init_mutex(self, &db->db_wr_lock);
	xt_init_cond(self, &db->db_wr_cond);

	/* Pre-allocate transaction data structures: */
	db->db_xn_data = (xtWord1 *) xt_malloc(self, sizeof(XTXactDataRec) * XT_XN_DATA_ALLOC_COUNT * XT_XN_NO_OF_SEGMENTS);
	db->db_xn_data_end = db->db_xn_data + sizeof(XTXactDataRec) * XT_XN_DATA_ALLOC_COUNT * XT_XN_NO_OF_SEGMENTS;
	xact = (XTXactDataPtr) db->db_xn_data;
	for (u_int i=0; i<XT_XN_NO_OF_SEGMENTS; i++) {
		seg = &db->db_xn_idx[i];
		xt_init_rwlock(self, &seg->xs_tab_lock);
		for (u_int j=0;  j<XT_XN_DATA_ALLOC_COUNT; j++) {
			xact->xd_next_xact = seg->xs_free_list;
			seg->xs_free_list = xact;
			xact++;
		}
	}

	/* Initialize the data logs: */
	db->db_datalogs.dlc_init(self, db); 

	/* Setup the transaction log: */
	db->db_xlog.xlog_setup(self, db, xt_db_log_file_threshold, xt_db_log_buffer_size, xt_db_transaction_buffer_size, xt_db_log_file_count);

	db->db_xn_end_time = 1;

	/* Initializing the restart file, also does
	 * recovery. This returns the log position after recovery.
	 *
	 * This is the log position where the writer thread will
	 * begin. The writer thread writes changes to the database that
	 * have been flushed to the log.
	 */
	xt_xres_init(self, db);

	/* Initialize the "last transaction in memory", by default
	 * this is the current transaction ID, which is the ID
	 * of the last transaction.
	 */
	for (u_int i=0; i<XT_XN_NO_OF_SEGMENTS; i++) {
		seg = &db->db_xn_idx[i];
		xt_init_rwlock(self, &seg->xs_tab_lock);
		seg->xt_last_xn_id = db->db_xn_curr_id;
	}

	/*
	 * The next transaction to clean is the lowest transaction
	 * in memory:
	 */
	db->db_xn_to_clean_id = db->db_xn_min_ram_id;

	/*
	 * No transactions are running, so the minimum transaction
	 * ID is the next one to run:
	 */
	db->db_xn_min_run_id = db->db_xn_curr_id + 1;

	db->db_xn_wait_for = xt_new_sortedlist(self, sizeof(XNWaitForRec), 100, 50, xn_compare_wait_for, db, xn_free_wait_for, FALSE, FALSE);
#ifdef DEBUG
	db->db_sw_cu_list = xt_new_sortedlist(self, sizeof(XNSWRecItemRec), 1, 2, xn_compare_cu_item, db, xn_free_cu_item, TRUE, FALSE);
#else
	db->db_sw_cu_list = xt_new_sortedlist(self, sizeof(XNSWRecItemRec), 100, 50, xn_compare_cu_item, db, xn_free_cu_item, TRUE, FALSE);
#endif
}

xtPublic void xt_xn_exit_db(XTThreadPtr self, XTDatabaseHPtr db)
{
#ifdef HIGH_X
	printf("=========> MOST TXs CURR ALLOC: %lu\n", tot_alloced);
	printf("=========> MOST TXs HIGH ALLOC: %lu\n", high_alloced);
	printf("=========> MAX TXs NOT CLEAN: %lu\n", not_clean_max);
	printf("=========> MAX TXs IN RAM: %lu\n", in_ram_max);
#endif

	xt_stop_sweeper(self, db);	// Should be done already!
	xt_stop_writer(self, db);	// Should be done already!

	xt_xres_exit(self, db);
	db->db_xlog.xlog_exit(self);

	db->db_datalogs.dlc_exit(self); 

	for (u_int i=0; i<XT_XN_NO_OF_SEGMENTS; i++) {
		XTXactSegPtr 	seg;

		seg = &db->db_xn_idx[i];
		for (u_int j=0; j<XT_XN_HASH_TABLE_SIZE; j++) {
			XTXactDataPtr	xact, nxact;
			
			xact = seg->xs_table[j];
			while (xact) {
				nxact = xact->xd_next_xact;
				xn_free_xact(db, seg, xact);
				xact = nxact;
			}
		}
		xt_free_rwlock(&seg->xs_tab_lock);
	}
	if (db->db_xn_wait_for) {
		xt_free_sortedlist(self, db->db_xn_wait_for);
		db->db_xn_wait_for = NULL;
	}
	if (db->db_sw_cu_list) {
		xt_free_sortedlist(self, db->db_sw_cu_list);
		db->db_sw_cu_list = NULL;
	}
	if (db->db_xn_data) {
		xt_free(self, db->db_xn_data);
		db->db_xn_data = NULL;
		db->db_xn_data_end = NULL;
	}

	xt_free_cond(&db->db_wr_cond);
	xt_free_mutex(&db->db_wr_lock);
	xt_free_cond(&db->db_xn_wait_cond);
	xt_free_mutex(&db->db_xn_wait_lock);
	xt_free_mutex(&db->db_xn_id_lock);
#ifdef DEBUG_RAM_LIST
	check_ram_free();
#endif
}

xtPublic void xt_xn_init_thread(XTThreadPtr self, int what_for)
{
	ASSERT(self->st_database);

	if (!xt_init_row_lock_list(&self->st_lock_list))
		xt_throw(self);
	switch (what_for) {
		case XT_FOR_COMPACTOR:
			self->st_xact_buf.xbuf_init(self, xt_db_transaction_buffer_size, FALSE);
			self->st_dlog_buf.dlb_init(self->st_database, xt_db_log_buffer_size);
			break;
		case XT_FOR_WRITER:
			/* The writer does not need a transaction buffer. */
			self->st_xact_buf.xbuf_init(self, 0, FALSE);
			self->st_dlog_buf.dlb_init(self->st_database, 0);
			break;
		case XT_FOR_SWEEPER:
			self->st_xact_buf.xbuf_init(self, xt_db_transaction_buffer_size, FALSE);
			self->st_dlog_buf.dlb_init(self->st_database, 0);
			break;
		case XT_FOR_USER:
			self->st_xact_buf.xbuf_init(self, xt_db_transaction_buffer_size, TRUE);
			self->st_dlog_buf.dlb_init(self->st_database, xt_db_log_buffer_size);
			break;
	}
}

xtPublic void xt_xn_exit_thread(XTThreadPtr self)
{
	if (self->st_xact_data)
		xt_xn_rollback(self);
	self->st_xact_buf.xbuf_exit(self);
	self->st_dlog_buf.dlb_exit(self);
	xt_exit_row_lock_list(&self->st_lock_list);
}

/* ----------------------------------------------------------------------
 * Begin and End Transactions
 */

xtPublic xtBool xt_xn_begin(XTThreadPtr self)
{
	XTDatabaseHPtr	db = self->st_database;
	xtXactID		xn_id;

	ASSERT(!self->st_xact_data);

	xt_mutex_spinlock(&db->db_xn_id_lock);
	xn_id = ++db->db_xn_curr_id;
	xt_unlock_mutex_ns(&db->db_xn_id_lock);

#ifdef HIGH_X
	if (xt_xn_is_before(not_clean_max, xn_id - db->db_xn_to_clean_id))
		not_clean_max = xn_id - db->db_xn_to_clean_id;
	if (xt_xn_is_before(in_ram_max, xn_id - db->db_xn_min_ram_id))
		in_ram_max = xn_id - db->db_xn_min_ram_id;
#endif
	/* [*] This is the gap between incrementing the ID,
	 * and creating the transaction in memory.
	 * See xt_xn_get_curr_id().
	 */

	if (!(self->st_xact_data = xn_add_new_xact(db, xn_id)))
		return FAILED;
	self->st_xact_buf.xbuf_data_written = FALSE;
	
	/* All transactions that committed before or at this time
	 * are this one are visible: */
	self->st_visible_time = db->db_xn_end_time;

	XT_DEBUG_TRACE(("BEGIN %p tx=%d\n", self, (int) self->st_xact_data->xd_start_xn_id));
#ifdef TRACE_VARIATIONS
	xt_trace("%s begin: T%d\n", self->t_name, (int) self->st_xact_data->xd_start_xn_id);
#endif
#ifdef PBXT_TRACE_STAT
	xt_tracef_query(self, "BEGIN T%lu\n", (u_long) self->st_xact_data->xd_start_xn_id);
#endif
	return OK;
}

static xtBool xn_end_xact(XTThreadPtr self, u_int status)
{
	XTXactDataPtr xact;

	ASSERT(self->st_xact_data);
	if ((xact = self->st_xact_data)) {
		XTDatabaseHPtr	db = self->st_database;
		xtXactID		end_id = db->db_xn_curr_id;
		xtXactID		xn_id = xact->xd_start_xn_id;
		xtWord4			end_seq = ++db->db_xn_end_time;
		
		if (self->st_xact_buf.xbuf_data_written) {
			/* The transaction wrote something: */
			XTXactEndEntryDRec	entry;
			xtWord4				sum;

			sum = XT_CHECKSUM4_XACT(xn_id) ^ XT_CHECKSUM4_XACT(end_id);
			entry.xe_status_1 = status;
			entry.xe_checksum_1 = XT_CHECKSUM_1(sum);
			XT_SET_DISK_4(entry.xe_xact_id_4, xn_id);
			XT_SET_DISK_4(entry.xe_xend_id_4, end_id);

			/* Flush the data log: */
			if (!self->st_dlog_buf.dlb_flush_log(TRUE))
				return FAILED;

			/* Write and flush the transaction log: */
			if (!self->st_xact_buf.xbuf_log_data(self, sizeof(XTXactEndEntryDRec), (XTXactLogBufferDPtr) &entry, TRUE))
				return FAILED;

			/* Drop locks is you have any: */
			self->st_lock_list.xt_remove_all_locks(db);

			xact->xd_end_time = end_seq;
			xact->xd_end_xn_id = end_id;

			/* Setting this flag completes the transaction (do it last)! */
			if (status == XT_LOG_ENT_COMMIT)
				xact->xd_flags |= (XT_XN_XAC_COMMITTED | XT_XN_XAC_ENDED);
			else
				xact->xd_flags |= XT_XN_XAC_ENDED;
		}
		else {
			/* Read-only transaction can be removed, immediately */
			xact->xd_end_time = end_seq;
			xact->xd_end_xn_id = end_id;

			/* Drop locks is you have any: */
			self->st_lock_list.xt_remove_all_locks(db);

			if (xt_xn_delete_xact(db, xn_id)) {
				if (db->db_xn_min_ram_id == xn_id)
					db->db_xn_min_ram_id = xn_id+1;
			}
		}
		if (db->db_xn_min_run_id == xn_id)
			db->db_xn_min_run_id = xn_id+1;

		self->st_xact_data = NULL;

		/* Wake up sweeper, and other waiting threads: */
		xt_xn_wakeup_transactions(db, FALSE);
	}
	return OK;
}

xtPublic xtBool xt_xn_commit(XTThreadPtr self)
{
#ifdef PBXT_TRACE_STAT
	xt_tracef_query(self, "COMMIT T%lu\n", (u_long) self->st_xact_data->xd_start_xn_id);
#endif
#ifdef TRACE_VARIATIONS
	xt_trace("%s commit: T%d\n", self->t_name, (int) self->st_xact_data->xd_start_xn_id);
#endif
	XT_DEBUG_TRACE(("COMMIT %p tx=%d\n", self, (int) self->st_xact_data->xd_start_xn_id));
	return xn_end_xact(self, XT_LOG_ENT_COMMIT);
}

xtPublic xtBool xt_xn_rollback(XTThreadPtr self)
{
#ifdef PBXT_TRACE_STAT
	xt_tracef_query(self, "ABORT T%lu\n", (u_long) self->st_xact_data->xd_start_xn_id);
#endif
#ifdef TRACE_VARIATIONS
	xt_trace("%s abort: T%d\n", self->t_name, (int) self->st_xact_data->xd_start_xn_id);
#endif
	XT_DEBUG_TRACE(("ABORT %p tx=%d\n", self, (int) self->st_xact_data->xd_start_xn_id));
	return xn_end_xact(self, XT_LOG_ENT_ABORT);
}

xtPublic xtBool xt_xn_log_tab_id(XTThreadPtr self, xtTableID tab_id)
{
	XTXactNewTabEntryDRec	entry;

	entry.xt_status_1 = XT_LOG_ENT_NEW_TAB;
	entry.xt_checksum_1 = XT_CHECKSUM_1(tab_id);
	XT_SET_DISK_4(entry.xt_tab_id_4, tab_id);
	return self->st_xact_buf.xbuf_log_data(self, sizeof(XTXactNewTabEntryDRec), (XTXactLogBufferDPtr) &entry, TRUE);
}

/*
 * XT_XN_ABORTED - Transaction was aborted.
 * XT_XN_MY_UPDATE - The record was update by me.
 * XT_XN_OTHER_UPDATE - The record was updated by someone else.
 * XT_XN_COMMITTED - The transaction was committed.
 */
xtPublic int xt_xn_status(XTOpenTablePtr ot, xtXactID xn_id, xtRecordID rec_id)
{
	register XTThreadPtr	self = ot->ot_thread;
	register XTXactDataPtr	xact;

	if (xt_xn_is_before(xn_id, self->st_database->db_xn_min_ram_id)) {
		/* Not in RAM, rollback done: */
//*DBG*/xt_dump_trace();
//*DBG*/xt_dump_xlogs(self->st_database);
//*DBG*/xt_check_table(self, ot);
		xn_add_cu_record(ot->ot_table->tab_db, ot->ot_table->tab_id, rec_id);
		return XT_XN_ABORTED;
	}
	if (xn_id == self->st_xact_data->xd_start_xn_id)
		return XT_XN_MY_UPDATE;
	if (!(xact = xt_xn_get_xact(self->st_database, xn_id))) {
		/* Not in RAM, rollback done: */
		xn_add_cu_record(ot->ot_table->tab_db, ot->ot_table->tab_id, rec_id);
		return XT_XN_ABORTED;
	}
	if (!(xact->xd_flags & XT_XN_XAC_ENDED))
		/* Transaction not ended, may be visible. */
		return XT_XN_OTHER_UPDATE;
	/* Visible if the transaction was committed: */
	if (xact->xd_flags & XT_XN_XAC_COMMITTED) {
		if (!xt_xn_is_before(self->st_visible_time, xact->xd_end_time))  // was self->st_visible_time >= xact->xd_end_time
			return XT_XN_VISIBLE;
		return XT_XN_NOT_VISIBLE;
	}
	return XT_XN_ABORTED;
}

/* ----------------------------------------------------------------------
 * S W E E P E R    P R O C E S S
 */

typedef struct XNSweeperState {
	XTDatabaseHPtr			ss_db;
	XTXactSeqReadRec		ss_seqread;
	XTDataBufferRec			ss_databuf;
	u_int					ss_call_cnt;
	XTBasicQueueRec			ss_to_free;
	xtBool					ss_flush_pending;
	XTOpenTablePtr			ss_ot;
} XNSweeperStateRec, *XNSweeperStatePtr;

static XTOpenTablePtr xn_sw_get_open_table(XTThreadPtr self, XNSweeperStatePtr ss, xtTableID tab_id, int *r)
{
	if (ss->ss_ot) {
		if (ss->ss_ot->ot_table->tab_id == tab_id)
			return ss->ss_ot;

		/* Flush the table indexes: */
		/* This is now done by the checkpointer thread...
		 * see xres_cp_flush_indices()
		if (!xt_flush_table_index(ss->ss_ot))
			xt_throw(self);
		*/

		xt_db_return_table_to_pool(self, ss->ss_ot);
		ss->ss_ot = NULL;
	}

	if (!ss->ss_ot) {
		if (!(ss->ss_ot = xt_db_open_pool_table(self, ss->ss_db, tab_id, r, TRUE)))
			return NULL;
	}

	return ss->ss_ot;
}

static void xn_sw_close_open_table(XTThreadPtr self, XNSweeperStatePtr ss)
{
	if (ss->ss_ot) {
		/* Flush the table indexes: */
		/* This is now done by the checkpointer thread
		 * see xres_cp_flush_indices()
		if (!xt_flush_table_index(ss->ss_ot))
			xt_throw(self);
		*/

		xt_db_return_table_to_pool(self, ss->ss_ot);
		ss->ss_ot = NULL;
	}
}

/*
 * A thread can set the db_sw_faster = TRUE to make
 * the sweeper go faster.
 */
static void xn_sw_could_go_faster(XTThreadPtr self, XTDatabaseHPtr db)
{
	if (db->db_sw_faster) {
		if (!db->db_sw_fast) {
			xt_set_normal_priority(self);
			db->db_sw_fast = TRUE;
		}
		db->db_sw_faster = FALSE;
	}
}

static void xn_sw_could_go_slower(XTThreadPtr self, XTDatabaseHPtr db)
{
	if (db->db_sw_fast && !db->db_sw_faster) {
		xt_set_low_priority(self);
		db->db_sw_fast = FALSE;
	}
}

/*
 * Service the "to free" queue. Return TRUE if something was
 * freed.
 */
static xtBool xn_sw_service_to_free(XTThreadPtr self, XNSweeperStatePtr ss, xtBool in_cleanup)
{
	XTDatabaseHPtr		db = ss->ss_db;
	XNSWToFreeItemPtr	free_item;
	xtBool				something_freed = FALSE;
	xtTableID			tab_id;
	XTOpenTablePtr		ot = NULL;

	if ((free_item = (XNSWToFreeItemPtr) xt_bq_get(&ss->ss_to_free))) {
		do {
			xn_sw_could_go_faster(self, db);

			if (xt_xn_is_before(db->db_xn_min_run_id, db->db_xn_to_clean_id)) {
				if (in_cleanup)
					/* If we are currently in cleanup, then
					 * the transaction current being cleaned
					 * has quit!
					 */
					db->db_xn_min_run_id = db->db_xn_to_clean_id+1;
				else
					db->db_xn_min_run_id = db->db_xn_to_clean_id;
			}
			
			/* Before we can free this resource, we must wait
			 * until the transaction is clean!
			 */
			if (!xt_xn_is_before(free_item->ri_wait_xn_id, db->db_xn_min_run_id)) { // was >=
				/* Now we have to check to see if we cannot move
				 * the minimum run ID forward.
				 */
				XTXactDataPtr	xact;
				xtXactID		tmp_xn_id; // Protect against concurrent update!
				xtXactID		xn_curr_xn_id = xt_xn_get_curr_id(db);

				for (;;) {
					tmp_xn_id = db->db_xn_min_run_id;
					if (xt_xn_is_before(xn_curr_xn_id, tmp_xn_id))
						break;
					if ((xact = xt_xn_get_xact(db, tmp_xn_id)) && !(xact->xd_flags & XT_XN_XAC_ENDED))
						/* The transaction is still running... */
						break;
					db->db_xn_min_run_id = tmp_xn_id+1;
				}
				if (!xt_xn_is_before(free_item->ri_wait_xn_id, db->db_xn_min_run_id))
					break;
			}

			/* The transaction is clean, I can free the resource... */
			if ((tab_id = free_item->ri_tab_id)) {
				/* Free the data record: */
				if ((ot = xn_sw_get_open_table(self, ss, tab_id, NULL))) {
					ss->ss_flush_pending = TRUE;
					if (!xt_tab_free_record(ot, XT_LOG_ENT_REC_FREED, free_item->x.ri_rec_id, FALSE))
						throw_();
				}
			}
			else {
				/* Free the transaction record in memory: */
				xt_xn_delete_xact(db, free_item->x.ri_xn_id);

				/* Recalculate the minimum memory transaction: */
				ASSERT(!xt_xn_is_before(free_item->x.ri_xn_id, db->db_xn_min_ram_id));
				
				if (db->db_xn_min_ram_id == free_item->x.ri_xn_id) {
					db->db_xn_min_ram_id = free_item->x.ri_xn_id+1;
				}
				else {
					xtXactID xn_id;
					xtXactID xn_curr_xn_id = xt_xn_get_curr_id(db);

					while (!xt_xn_is_before(xn_curr_xn_id, db->db_xn_min_ram_id)) { // was db->db_xn_min_ram_id <= xn_curr_xn_id
						/* db_xn_min_ram_id may be changed, by some other process! */
						xn_id = db->db_xn_min_ram_id;
						if (xt_xn_get_xact(db, xn_id))
							break;
						db->db_xn_min_ram_id = xn_id+1;
					}
				}
			}

			xt_bq_next(&ss->ss_to_free);
		} while ((free_item = (XNSWToFreeItemPtr) xt_bq_get(&ss->ss_to_free)));
	}

	return something_freed;
}

/* Add a record to the "to free" queue. We note the current
 * transaction at the time this is done. The record will
 * only be freed once this transaction terminated, together
 * with all transactions that started before it! 
 *
 * The reason for this is that a sequential scan or some
 * other operation may read a committed record which is no longer
 * valid because it is no longer the latest variation (the first
 * variation reachable from the row pointer).
 *
 * In this case, the sweeper will free the variation.
 * If the variation is re-used and committed before
 * the sequential scan or read completes, and by some
 * fluke is used by the same record as previously,
 * the system will think the record is valid
 * again.
 *
 * Without re-reading the record the sequential
 * scan or other read will find it on the variation list, and
 * return the record data as if valid!
 *
 * ------------ 2008-01-03
 *
 * An example of this is:
 *
 * Assume we have 3 records.
 * The 3rd record is deleted, and committed.
 * Before cleanup can be performed
 * a sequential scan takes a copy of the records.
 *
 * Now assume a new insert is done before
 * the sequential scan gets to the 3rd record.
 *
 * The insert allocates the 3rd row and 3rd record
 * again.
 *
 * Now, when the sequential scan gets to the old copy of the 3rd record,
 * this is valid because the row points to this record again.
 *
 * HOWEVER! I have now changed the sequential scan so that it accesses
 * the records from the cache, without making a copy.
 *
 * This means that this problem cannot occur because the sequential scan
 * always reads the current data from the cache.
 *
 * There is also no race condition (although no lock is taken), because
 * the record is writen before the row (see here [(5)]).
 *
 * This means that the row does not point to the record before the
 * record has been modified.
 *
 * Once the record has been modified then the sequential scan will see
 * that the record belongs to a new transaction.
 *
 * If the row pointer was set before the record updated then a race
 * condition would exist when the sequential scan reads the record
 * after the insert has updated the row pointer but before it has
 * changed the record.
 *
 * AS A RESULT:
 *
 * I believe I can remove the delayed free record!
 *
 * This means I can combine the REMOVE and FREE operations.
 *
 * This is good because this takes care of the problem
 * that records are lost when:
 *
 * The server crashes when the delayed free list still has items on it.
 * AND
 * The transaction that freed the records has been cleaned, and this
 * fact has been committed to the log.
 *
 * So I have removed the delay here: [(6)]
 */
static void xn_sw_add_record_to_free(XTThreadPtr self, XNSweeperStatePtr ss, XTOpenTablePtr ot, xtRecordID rec_id)
{
	XNSWToFreeItemRec free_item;

	if ((ss->ss_to_free.bq_front - ss->ss_to_free.bq_back) >= XT_TN_MAX_TO_FREE) {
		/* If the queue is full, try to free some items:
		 * We use the call count to avoid doing this every time,
		 * when the queue overflows!
		 */
		if ((ss->ss_call_cnt % XT_TN_MAX_TO_FREE_CHECK) == 0)
			xn_sw_service_to_free(self, ss, FALSE);
		ss->ss_call_cnt++;
	}

	free_item.ri_wait_xn_id = ss->ss_db->db_xn_curr_id;
	free_item.ri_tab_id = ot->ot_table->tab_id;
	free_item.x.ri_rec_id = rec_id;

#ifdef DEBUG
	XNSWToFreeItemPtr free_list = (XNSWToFreeItemPtr) ss->ss_to_free.bq_data;
	for (u_int i=ss->ss_to_free.bq_back; i<ss->ss_to_free.bq_front; i++) {
		if (free_list[i].ri_tab_id == ot->ot_table->tab_id && free_list[i].x.ri_rec_id == rec_id)
			ASSERT(FALSE);
	}
#endif
	xt_bq_add(self, &ss->ss_to_free, &free_item);
}

/*
 * As above, but instead a transaction is added to the "to free" queue.
 *
 * It is important that transactions remain in memory until all
 * currently running transactions have ended. This is because
 * sequential and index scans have copies of old data.
 *
 * In the old data a record may not be indicated as cleaned. Such
 * a record is considered invalid if the transaction is not in RAM.
 *
 * GOTCHA:
 *
 * And this problem is demonstrated by the following example
 * which was derived from flush_table.test.
 *
 * Each handler command below is a separate transaction.
 * However the buffer is loaded by 'read first'.
 * Depending on when cleanup occurs, records can disappear
 * in some of the next commands.
 *
 * 2 solutions for the test. Use begin ... commit around
 * handler open ... close. Or use analyze table t1 before
 * open. analyze table waits for the sweeper to complete!
 *
 * create table dummy(table_id char(20) primary key);
 * let $1=100;
 * while ($1)
 * {
 *   drop table if exists t1;
 *   create table t1(table_id char(20) primary key);
 *   insert into t1 values ('Record-01');
 *   insert into t1 values ('Record-02');
 *   insert into t1 values ('Record-03');
 *   insert into t1 values ('Record-04');
 *   insert into t1 values ('Record-05');
 *   handler t1 open;
 *   handler t1 read first limit 1;
 *   handler t1 read next limit 1;
 *   handler t1 read next limit 1;
 *   handler t1 read next limit 1;
 *   handler t1 close;
 *   commit;
 *   dec $1;
 * }
 * 
 */
static void xn_sw_add_xact_to_free(XTThreadPtr self, XNSweeperStatePtr ss, xtXactID xn_id)
{
	XNSWToFreeItemRec free_item;

	if ((ss->ss_to_free.bq_front - ss->ss_to_free.bq_back) >= XT_TN_MAX_TO_FREE) {
		/* If the queue is full, try to free some items:
		 * We use the call count to avoid doing this every time,
		 * when the queue overflows!
		 */
		if ((ss->ss_call_cnt % XT_TN_MAX_TO_FREE_CHECK) == 0)
			/* GOTCHA: This call was not locking the sweeper,
			 * this could cause failure, of course:
			 */
			xn_sw_service_to_free(self, ss, TRUE);
		ss->ss_call_cnt++;
	}

	free_item.ri_wait_xn_id = ss->ss_db->db_xn_curr_id;
	free_item.ri_tab_id = 0;
	free_item.x.ri_xn_id = xn_id;

	xt_bq_add(self, &ss->ss_to_free, &free_item);
}

static void xt_sw_delete_variations(XTThreadPtr self, XNSweeperStatePtr ss, XTOpenTablePtr ot, xtRecordID rec_id, xtRowID row_id, xtXactID xn_id)
{
	xtRecordID prev_var_rec_id;

	while (rec_id) {
		switch (xt_tab_remove_record(ot, rec_id, ss->ss_databuf.db_data, &prev_var_rec_id, FALSE, row_id, xn_id)) {
			case XT_ERR:
				throw_();
				return;
			case TRUE:
				break;
		}
		rec_id = prev_var_rec_id;
	}
}

static void xt_sw_delete_variation(XTThreadPtr self, XNSweeperStatePtr ss, XTOpenTablePtr ot, xtRecordID rec_id, xtBool clean_delete, xtRowID row_id, xtXactID xn_id)
{
	xtRecordID prev_var_rec_id;

	switch (xt_tab_remove_record(ot, rec_id, ss->ss_databuf.db_data, &prev_var_rec_id, clean_delete, row_id, xn_id)) {
		case XT_ERR:
			throw_();
			return;
		case TRUE:
			break;
		case FALSE:
			break;
	}
}

/* Set rec_type to this value in order to force cleanup, without
 * a check.
 */
#define XN_FORCE_CLEANUP		XT_TAB_STATUS_FREED

/*
 * Read the record to be cleaned. Return TRUE if the cleanup has already been done.
 */
static xtBool xn_sw_cleanup_done(XTThreadPtr self, XTOpenTablePtr ot, xtRecordID rec_id, xtXactID xn_id, u_int rec_type, u_int stat_id, xtRowID row_id, XTTabRecHeadDPtr rec_head)
{
	if (!xt_tab_get_rec_data(ot, rec_id, sizeof(XTTabRecHeadDRec), (xtWord1 *) rec_head))
		throw_();

	if (rec_type == XN_FORCE_CLEANUP) {
		if (XT_REC_IS_FREE(rec_head->tr_rec_type_1))
			return TRUE;
	}
	else {
		/* Transaction must match: */
		if (XT_GET_DISK_4(rec_head->tr_xact_id_4) != xn_id)
			return TRUE;

		/* Record header must match expected value from
		 * log or clean has been done, or is not required.
		 *
		 * For example, it is not required if a record
		 * has been overwritten in a transaction.
		 */
		if (rec_head->tr_rec_type_1 != rec_type ||
			rec_head->tr_stat_id_1 != stat_id)
			return TRUE;

		/* Row must match: */
		if (XT_GET_DISK_4(rec_head->tr_row_id_4) != row_id)
			return TRUE;
	}

	return FALSE;
}

static void xn_sw_clean_indices(XTThreadPtr self, XTOpenTablePtr ot, xtRecordID rec_id, xtRowID row_id, xtWord1 *rec_data, xtWord1 *rec_buffer)
{
	XTTableHPtr	tab = ot->ot_table;
	u_int		cols_req;
	XTIndexPtr	*ind;

	if (!tab->tab_dic.dic_key_count)
		return;

	cols_req = tab->tab_dic.dic_ind_cols_req;
	if (XT_REC_IS_FIXED(rec_data[0]))
		rec_buffer = rec_data + XT_REC_FIX_HEADER_SIZE;
	else {
		if (XT_REC_IS_VARIABLE(rec_data[0])) {
			if (!myxt_load_row(ot, rec_data + XT_REC_FIX_HEADER_SIZE, rec_buffer, cols_req))
				goto failed;
		}
		else if (XT_REC_IS_EXT_DLOG(rec_data[0])) {
			ASSERT(cols_req);
			if (cols_req && cols_req <= tab->tab_dic.dic_fix_col_count) {
				if (!myxt_load_row(ot, rec_data + XT_REC_EXT_HEADER_SIZE, rec_buffer, cols_req))
					goto failed;
			}
			else {
				if (rec_data != ot->ot_row_rbuffer)
					memcpy(ot->ot_row_rbuffer, rec_data, tab->tab_dic.dic_rec_size);
				if (!xt_tab_load_ext_data(ot, rec_id, rec_buffer, cols_req))
					goto failed;
			}
		}
		else
			/* This is possible, the record has already been cleaned up. */
			return;
	}

	ind = tab->tab_dic.dic_keys;
	for (u_int i=0; i<tab->tab_dic.dic_key_count; i++, ind++) {
		if (!xt_idx_update_row_id(ot, *ind, rec_id, row_id, rec_buffer))
			xt_log_and_clear_exception_ns();
	}
	return;
	
	failed:
	xt_log_and_clear_exception_ns();
}

/*
 * Return TRUE if the cleanup was done. FAILED if cleanup could not be done
 * because dictionary information is not available.
 */
static xtBool xn_sw_cleanup_variation(XTThreadPtr self, XNSweeperStatePtr ss, XTXactDataPtr xact, xtTableID tab_id, xtRecordID rec_id, u_int status, u_int rec_type, u_int stat_id, xtRowID row_id, xtWord1 *rec_buf)
{
	XTOpenTablePtr		ot;
	XTTableHPtr			tab;
	XTTabRecHeadDRec	rec_head;
	xtRecordID			after_rec_id;
	xtXactID			xn_id;
	int					r;

	if (!(ot = xn_sw_get_open_table(self, ss, tab_id, &r))) {
		/* The table no longer exists, consider cleanup done: */
		switch (r) {
			case XT_TAB_NOT_FOUND:
				break;
			case XT_TAB_NO_DICTIONARY:
			case XT_TAB_POOL_CLOSED:
				return FALSE;
		}
		return TRUE;
	}

	tab = ot->ot_table;

	/* Make sure the buffer is large enough! */
	xt_db_set_size(self, &ss->ss_databuf, (size_t) tab->tab_dic.dic_buf_size);

	xn_id = xact->xd_start_xn_id;
	if (xact->xd_flags & XT_XN_XAC_COMMITTED) {
		/* The transaction has been committed. Clean the record and
		 * remove variations no longer in use.
		 */
		switch (status) {
			case XT_LOG_ENT_REC_MODIFIED:
			case XT_LOG_ENT_UPDATE:
			case XT_LOG_ENT_UPDATE_FL:
			case XT_LOG_ENT_UPDATE_BG:
			case XT_LOG_ENT_UPDATE_FL_BG:
				if (xn_sw_cleanup_done(self, ot, rec_id, xn_id, rec_type, stat_id, row_id, &rec_head))
					goto done_ok;
				after_rec_id = XT_GET_DISK_4(rec_head.tr_prev_rec_id_4);
				xt_sw_delete_variations(self, ss, ot, after_rec_id, row_id, xn_id);
				rec_head.tr_rec_type_1 |= XT_TAB_STATUS_CLEANED_BIT;
				XT_SET_NULL_DISK_4(rec_head.tr_prev_rec_id_4);
				if (!xt_tab_put_log_op_rec_data(ot, XT_LOG_ENT_REC_CLEANED, 0, rec_id, offsetof(XTTabRecHeadDRec, tr_prev_rec_id_4) + XT_RECORD_ID_SIZE, (xtWord1 *) &rec_head))
					throw_();
				xn_sw_clean_indices(self, ot, rec_id, row_id, rec_buf, ss->ss_databuf.db_data);
				break;
			case XT_LOG_ENT_INSERT:
			case XT_LOG_ENT_INSERT_FL:
			case XT_LOG_ENT_INSERT_BG:
			case XT_LOG_ENT_INSERT_FL_BG:
				/* In the case of insert, we avoid reading the record into cache! */
				rec_head.tr_rec_type_1 = rec_type | XT_TAB_STATUS_CLEANED_BIT;
				if (!xt_tab_put_log_op_rec_data(ot, XT_LOG_ENT_REC_CLEANED_1, 0, rec_id, 1, (xtWord1 *) &rec_head.tr_rec_type_1))
					throw_();
				xn_sw_clean_indices(self, ot, rec_id, row_id, rec_buf, ss->ss_databuf.db_data);
				break;
			case XT_LOG_ENT_DELETE:
			case XT_LOG_ENT_DELETE_FL:
			case XT_LOG_ENT_DELETE_BG:
			case XT_LOG_ENT_DELETE_FL_BG:
				if (xn_sw_cleanup_done(self, ot, rec_id, xn_id, rec_type, stat_id, row_id, &rec_head))
					goto done_ok;
				after_rec_id = XT_GET_DISK_4(rec_head.tr_prev_rec_id_4);
				xt_sw_delete_variations(self, ss, ot, after_rec_id, row_id, xn_id);
				xt_sw_delete_variation(self, ss, ot, rec_id, TRUE, row_id, xn_id);
				if (row_id) {
					if (!xt_tab_free_row(ot, tab, row_id))
						throw_();
				}
				break;
		}
	}
	else {
		/* The transaction has been aborted. Remove the variation from the
		 * variation list. If this means the list is empty, then remove
		 * the record as well.
		 */
		xtRecordID			first_rec_id, next_rec_id, prev_rec_id;
		XTTabRecHeadDRec	prev_rec_head;

		if (xn_sw_cleanup_done(self, ot, rec_id, xn_id, rec_type, stat_id, row_id, &rec_head))
			goto done_ok;

		if (!row_id)
			row_id = XT_GET_DISK_4(rec_head.tr_row_id_4);
		after_rec_id = XT_GET_DISK_4(rec_head.tr_prev_rec_id_4);

		/* Now remove the record from the variation list,
		 * (if it is still on the list).
		 */
		xt_rwlock_wrlock(&tab->tab_row_rwlock[row_id % XT_ROW_RWLOCKS]);

		/* Find the variation before the variation we wish to remove: */
		if (!(xt_tab_get_row(ot, row_id, &first_rec_id)))
			goto failed;
		prev_rec_id = 0;
		next_rec_id = first_rec_id;
		while (next_rec_id != rec_id) {
			if (!next_rec_id) {
				/* The record was not found in the list (we are done) */
				xt_rwlock_unlock(&tab->tab_row_rwlock[row_id % XT_ROW_RWLOCKS]);
				goto unlink_done;
			}
			if (!xt_tab_get_rec_data(ot, next_rec_id, sizeof(XTTabRecHeadDRec), (xtWord1 *) &prev_rec_head)) {
				xt_log_and_clear_exception(self);
				break;
			}
			prev_rec_id = next_rec_id;
			next_rec_id = XT_GET_DISK_4(prev_rec_head.tr_prev_rec_id_4);
		}

		if (next_rec_id == rec_id) {
			/* The record was found on the list: */
			if (prev_rec_id) {
				/* Unlink the deleted variation:
				 * I have found the following sequence:
				 *
				 * 17933 in use  1906112
				 * 1906112 delete      xact=2901   row=17933 prev=2419240
				 * 2419240 delete      xact=2899   row=17933 prev=2153360
				 * 2153360 record-X C  xact=2599   row=17933 prev=0 Xlog=151 Xoff=16824 Xsiz=100
				 *
				 * Despite the following facts which should prevent chains from
				 * forming:
				 *
				 * --- Only one transaction can modify a row
				 * at any one time. So it is not possible for a new change
				 * to be linked onto an uncommitted change.
				 * 
				 * --- Transactions that modify the same row
				 * twice do not allocate a new record for each change.
				 *
				 * -- A change that has been
				 * rolled back will not be linked onto. Instead
				 * the new transaction will link to the last.
				 * Comitted record.
				 *
				 * So if the sweeper is slow in doing its job
				 * we can have the situation that a number of records
				 * can refer to the last committed record of the
				 * row.
				 *
				 * Only one will be reference by the row pointer.
				 *
				 * The other, will all have been rolled back.
				 * This occurs over here: [(4)]
				 */
				XT_SET_DISK_4(prev_rec_head.tr_prev_rec_id_4, after_rec_id);
				if (!xt_tab_put_log_op_rec_data(ot, XT_LOG_ENT_REC_UNLINKED, 0, prev_rec_id, offsetof(XTTabRecHeadDRec, tr_prev_rec_id_4) + XT_RECORD_ID_SIZE, (xtWord1 *) &prev_rec_head))
					goto failed;
			}
			else {
				/* Variation to be removed at the front of the list. */
				ASSERT(rec_id == first_rec_id);
				if (after_rec_id) {
					/* Unlink the deleted variation, from the front of the list: */
					if (!xt_tab_set_row(ot, XT_LOG_ENT_ROW_SET, row_id, after_rec_id))
						goto failed;
				}
				else {
					/* No more variations, remove the row: */
					if (!xt_tab_free_row(ot, tab, row_id))
						goto failed;
				}
			}
		}

		xt_rwlock_unlock(&tab->tab_row_rwlock[row_id % XT_ROW_RWLOCKS]);

		/* Note: even when not found on the row list, the record must still
		 * be freed.
		 *
		 * There might be an exception to this, but there are very definite
		 * cases where this is required, for example when an unreferenced
		 * record is fouund and added to the clean up list xn_add_cu_record().
		 */

		unlink_done:
		/* Delete the extended record and index entries:
		 *
		 * NOTE! This must be done after we have release the row lock. Because
		 * a thread that does a duplicate check locks the index, and then
		 * check whether a row is valid, and can deadlock with
		 * code that locks a row, then an index!
		 *
		 * However, this should all be OK, because the variation has been removed from the
		 * row variation list at this stage, and now just need to be deleted.
		 */
		xt_sw_delete_variation(self, ss, ot, rec_id, FALSE, row_id, xn_id);
	}

	done_ok:
	return OK;

	failed:
	xt_rwlock_unlock(&tab->tab_row_rwlock[row_id % XT_ROW_RWLOCKS]);
	throw_();
	return FAILED;
}

/* Go through all updated records of a transaction and cleanup.
 * This means, of the transaction was aborted, then all the variations written
 * by the transaction must be removed.
 * If the transaction was committed then we remove older variations.
 * If a delete was committed this can lead to the row being removed.
 *
 * After a transaction has been cleaned it can be removed from RAM.
 * If this was the last transaction in a log, and the log has reached
 * threshold, and the log is no longer in exclusive use, then the log
 * can be deleted.
 *
 * This function returns OK if the transaction was cleaned up.
 */
static xtBool xn_sw_cleanup_xact(XTThreadPtr self, XNSweeperStatePtr ss, XTXactDataPtr xact)
{
	XTDatabaseHPtr		db = ss->ss_db;
	XTXactLogBufferDPtr	record;
	xtTableID			tab_id;
	xtRecordID			rec_id;
	xtBool				cleanup_complete = FALSE;
	xtXactID			xn_id;
	xtRowID				row_id;

	if (!db->db_xlog.xlog_seq_start(&ss->ss_seqread, xact->xd_begin_log, xact->xd_begin_offset, FALSE))
		xt_throw(self);

	for (;;) {
		if (self->t_quit)
			goto stop_cleanup;

		xn_sw_could_go_faster(self, db);

		if (!db->db_xlog.xlog_seq_next(&ss->ss_seqread, &record, FALSE))
			xt_throw(self);
		if (!record)
			goto stop_cleanup;
		switch (record->xh.xh_status_1) {
			case XT_LOG_ENT_NEW_LOG:
				if (!db->db_xlog.xlog_seq_start(&ss->ss_seqread, XT_GET_DISK_4(record->xl.xl_log_id_4), 0, FALSE))
					xt_throw(self);
				break;
			case XT_LOG_ENT_COMMIT:
			case XT_LOG_ENT_ABORT:
				xn_id = XT_GET_DISK_4(record->xe.xe_xact_id_4);
				if (xn_id == xact->xd_start_xn_id)
					goto break_loop;
				break;
			case XT_LOG_ENT_REC_MODIFIED:
			case XT_LOG_ENT_UPDATE:
			case XT_LOG_ENT_INSERT:
			case XT_LOG_ENT_DELETE:
			case XT_LOG_ENT_UPDATE_BG:
			case XT_LOG_ENT_INSERT_BG:
			case XT_LOG_ENT_DELETE_BG:
				xn_id = XT_GET_DISK_4(record->xu.xu_xact_id_4);
				if (xn_id != xact->xd_start_xn_id)
					break;
				tab_id = XT_GET_DISK_4(record->xu.xu_tab_id_4);
				rec_id = XT_GET_DISK_4(record->xu.xu_rec_id_4);
				row_id = XT_GET_DISK_4(record->xu.xu_row_id_4);
				if (!xn_sw_cleanup_variation(self, ss, xact, tab_id, rec_id, record->xu.xu_status_1, record->xu.xu_rec_type_1, record->xu.xu_stat_id_1, row_id, &record->xu.xu_rec_type_1))
					goto stop_cleanup;
				break;
			case XT_LOG_ENT_UPDATE_FL:
			case XT_LOG_ENT_INSERT_FL:
			case XT_LOG_ENT_DELETE_FL:
			case XT_LOG_ENT_UPDATE_FL_BG:
			case XT_LOG_ENT_INSERT_FL_BG:
			case XT_LOG_ENT_DELETE_FL_BG:
				xn_id = XT_GET_DISK_4(record->xf.xf_xact_id_4);
				if (xn_id != xact->xd_start_xn_id)
					break;
				tab_id = XT_GET_DISK_4(record->xf.xf_tab_id_4);
				rec_id = XT_GET_DISK_4(record->xf.xf_rec_id_4);
				row_id = XT_GET_DISK_4(record->xf.xf_row_id_4);
				if (!xn_sw_cleanup_variation(self, ss, xact, tab_id, rec_id, record->xf.xf_status_1, record->xf.xf_rec_type_1, record->xf.xf_stat_id_1, row_id, &record->xf.xf_rec_type_1))
					goto stop_cleanup;
				break;
			default:
				break;
		}
	}
	break_loop:
	cleanup_complete = TRUE;

	stop_cleanup:
	if (!cleanup_complete)
		return FAILED;

	/* Write the log to indicate the transaction has been cleaned: */
	XTXactCleanupEntryDRec cu;

	cu.xc_status_1 = XT_LOG_ENT_CLEANUP;
	cu.xc_checksum_1 = XT_CHECKSUM_1(XT_CHECKSUM4_XACT(xact->xd_start_xn_id));
	XT_SET_DISK_4(cu.xc_xact_id_4, xact->xd_start_xn_id);

	if (!self->st_xact_buf.xbuf_log_data(self, sizeof(XTXactCleanupEntryDRec), (XTXactLogBufferDPtr) &cu, FALSE))
		return FAILED;

	ss->ss_flush_pending = TRUE;

	xn_sw_add_xact_to_free(self, ss, xact->xd_start_xn_id);
	return OK;
}

/* Cleanup records in the list. Return TRUE if any
 * record was cleaned up.
 */
static xtBool xn_sw_cleanup_records(XTThreadPtr self, XNSweeperStatePtr ss)
{
	XTDatabaseHPtr	db = ss->ss_db;
	xtTableID		tab_id = 0;
	xtRecordID		rec_id = 0;
	XTXactDataRec	xact;

	if (!xt_sl_get_size(db->db_sw_cu_list))
		return FALSE;

	xact.xd_start_xn_id = 0;
	xact.xd_end_xn_id = 0;
	xact.xd_end_time = 0;
	xact.xd_begin_log = 0;
	xact.xd_begin_offset = 0;
	xact.xd_flags = 0;
	while (!self->t_quit && xn_get_cu_record(db, &tab_id, &rec_id)) {
		xn_sw_could_go_faster(self, db);

		ss->ss_flush_pending = TRUE;
		if (!xn_sw_cleanup_variation(self, ss, &xact, tab_id, rec_id, 0, XN_FORCE_CLEANUP, 0, 0, NULL)) {
			xn_add_cu_record(db, tab_id, rec_id);
			break;
		}
	}

	xn_sw_service_to_free(self, ss, FALSE);
	
	return TRUE;
}

static void xn_free_sw_state(XTThreadPtr self, XNSweeperStatePtr ss)
{
	xn_sw_close_open_table(self, ss);
	if (ss->ss_db)
		ss->ss_db->db_xlog.xlog_seq_exit(&ss->ss_seqread);
	xt_db_set_size(self, &ss->ss_databuf, 0);
	xt_bq_set_size(self, &ss->ss_to_free, 0);
}

static void xn_sw_main(XTThreadPtr self)
{
	XTDatabaseHPtr		db = self->st_database;
	XNSweeperStatePtr	ss;
	XTXactDataPtr		xact, xact2;

	xt_set_low_priority(self);

	alloczr_(ss, xn_free_sw_state, sizeof(XNSweeperStateRec), XNSweeperStatePtr);
	ss->ss_db = db;

	if (!db->db_xlog.xlog_seq_init(&ss->ss_seqread, xt_db_transaction_buffer_size))
		xt_throw(self);

	ss->ss_to_free.bq_item_size = sizeof(XNSWToFreeItemRec);
	ss->ss_to_free.bq_max_waste = XT_TN_MAX_TO_FREE_WASTE;
	ss->ss_to_free.bq_item_inc = XT_TN_MAX_TO_FREE_INC;
	ss->ss_call_cnt = 0;
	ss->ss_flush_pending = FALSE;

	while (!self->t_quit) {
		while (!self->t_quit) {
			/* We are just about to check the condition for sleeping,
			 * so if the condition for sleeping holds, then we wil
			 * exit the loop and sleep.
			 *
			 * We will then sleep if nobody sets the flag before we
			 * actually do sleep!
			 */
			db->db_sw_wakeup = FALSE;
			if (xt_xn_is_before(xt_xn_get_curr_id(db), db->db_xn_to_clean_id))
				break;
			xn_sw_could_go_faster(self, db);

			if ((xact = xt_xn_get_xact(db, db->db_xn_to_clean_id))) {
				xtXactID xn_id;

				if (!(xact->xd_flags & XT_XN_XAC_ENDED))
					/* Transaction has not yet committed. */
					goto sleep;

				/* Check if we can cleanup the transaction.
				 * We do this by checking to see if there is any running
				 * transaction which start before the end of this transaction.
				 */
				xn_id = xact->xd_start_xn_id;
				while (xt_xn_is_before(xn_id, xact->xd_end_xn_id)) {
					xn_id++;
					if ((xact2 = xt_xn_get_xact(db, xn_id))) {
						if (!(xact2->xd_flags & XT_XN_XAC_ENDED))
							/* A transaction was started before the end of
							 * the transaction we wish to sweep, and this
							 * transaction has not committed, the we have to
							 * wait.
							 */
							goto sleep;
					}
				}
				
				/* Can cleanup the transaction, and move to the next. */
				if (xact->xd_flags & XT_XN_XAC_LOGGED) {
					if (!xn_sw_cleanup_xact(self, ss, xact)) {
						/* We failed to clean (try again later)... */
						goto sleep;
					}
				}
				else {
					/* This was a read-only transaction, it is safe to
					 * just remove the transaction structure from memory.
					 * (should not be necessary because RO transactions
					 * do this themselves):
					 */
					if (xt_xn_delete_xact(db, db->db_xn_to_clean_id)) {
						if (db->db_xn_min_ram_id == db->db_xn_to_clean_id)
							db->db_xn_min_ram_id = db->db_xn_to_clean_id+1;
					}
				}
			}
			
			/* Move on to clean the next: */
			db->db_xn_to_clean_id++;
		}

		sleep:
		if (xn_sw_cleanup_records(self, ss))
			continue;
			
		if (xn_sw_service_to_free(self, ss, TRUE))
			continue;

		xn_sw_close_open_table(self, ss);

		xn_sw_could_go_slower(self, db);

		/* Shrink the free list, if it is empty, and larger then
		 * the default:
		 */
		if (ss->ss_to_free.bq_size > XT_TN_MAX_TO_FREE) {
			if (ss->ss_to_free.bq_front == 0 && ss->ss_to_free.bq_back == 0)
				xt_bq_set_size(self, &ss->ss_to_free, XT_TN_MAX_TO_FREE);
		}

		if (ss->ss_flush_pending) {
			time_t then, now;

			/* If nothing happens in the next 2.4 seconds, then
			 * we will flush the log.
			 */
			then = time(NULL);
			xn_sw_wait_for_xact(self, db, 240);

			/* Flush pending means we have written something to the log.
			 *
			 * if so we flush the log so that the writer will also do
			 * its work!
			 *
			 * This will lead to the freeer continuing if it is waiting.
			 */
			now = time(NULL);
			if (now >= then + 2) {
				if (!self->st_xact_buf.xbuf_flush_log(self, TRUE))
					xt_throw(self);
				ss->ss_flush_pending = FALSE;
			}
		}
		else
			/* If no flush is pending, we can wait much longer: */
			xn_sw_wait_for_xact(self, db, 12000);
	}

	if (ss->ss_flush_pending) {
		self->st_xact_buf.xbuf_flush_log(self, TRUE);
		ss->ss_flush_pending = FALSE;
	}

	freer_(); // xn_free_sw_state(ss)
}

static void *xn_sw_run_thread(XTThreadPtr self)
{
	XTDatabaseHPtr	db = (XTDatabaseHPtr) self->t_data;
	int				count;
	void			*mysql_thread;

	mysql_thread = myxt_create_thread();

	while (!self->t_quit) {
		try_(a) {
			/*
			 * The garbage collector requires that the database
			 * is in use because.
			 */
			xt_use_database(self, db, XT_FOR_SWEEPER);

			/* This action is both safe and required:
			 *
			 * safe: releasing the database is safe because as
			 * long as this thread is running the database
			 * reference is valid, and this reference cannot
			 * be the only one to the database because
			 * otherwize this thread would not be running.
			 *
			 * required: releasing the database is necessary
			 * otherwise we cannot close the database
			 * correctly because we only shutdown this
			 * thread when the database is closed and we
			 * only close the database when all references
			 * are removed.
			 */
			xt_heap_release(self, self->st_database);

			xn_sw_main(self);
		}
		catch_(a) {
			/* This error is "normal"! */
			if (self->t_exception.e_xt_err != XT_ERR_NO_DICTIONARY &&
				!(self->t_exception.e_xt_err == XT_SIGNAL_CAUGHT &&
				self->t_exception.e_sys_err == SIGTERM))
				xt_log_and_clear_exception(self);
		}
		cont_(a);

		/* Avoid releasing the database (done above) */
		self->st_database = NULL;
		xt_unuse_database(self, self);

		/* After an exception, pause before trying again... */
		/* Number of seconds */
#ifdef DEBUG
		count = 10;
#else
		count = 2*60;
#endif
		while (!self->t_quit && count > 0) {
			sleep(1);
			count--;
		}
	}

	myxt_destroy_thread(mysql_thread);
	return NULL;
}

static void xn_sw_free_thread(XTThreadPtr self, void *data)
{
	XTDatabaseHPtr db = (XTDatabaseHPtr) data;

	if (db->db_sw_thread) {
		xt_lock_mutex(self, &db->db_xn_wait_lock);
		pushr_(xt_unlock_mutex, &db->db_xn_wait_lock);
		db->db_sw_thread = NULL;
		freer_(); // xt_unlock_mutex(&db->db_xn_wait_lock)
	}
}

/* Wait for a transaction to quit: */
static void xn_sw_wait_for_xact(XTThreadPtr self, XTDatabaseHPtr db, u_int hsecs)
{
	xt_lock_mutex(self, &db->db_xn_wait_lock);
	pushr_(xt_unlock_mutex, &db->db_xn_wait_lock);
	db->db_sw_idle = TRUE;
	if (!self->t_quit && !db->db_sw_wakeup)
		xt_timed_wait_cond(self, &db->db_xn_wait_cond, &db->db_xn_wait_lock, hsecs * 10);
	db->db_sw_idle = FALSE;
	db->db_sw_check_count++;
	freer_(); // xt_unlock_mutex(&db->db_xn_wait_lock)
}

xtPublic void xt_start_sweeper(XTThreadPtr self, XTDatabaseHPtr db)
{
	char name[PATH_MAX];

	sprintf(name, "SW-%s", xt_last_directory_of_path(db->db_main_path));
	xt_remove_dir_char(name);
	db->db_sw_thread = xt_new_thread(self, name, TRUE);
	xt_set_thread_data(db->db_sw_thread, db, xn_sw_free_thread);
	xt_run_thread(self, db->db_sw_thread, xn_sw_run_thread);
}

xtPublic void xt_wait_for_sweeper(XTThreadPtr self, XTDatabaseHPtr db)
{
	time_t	then, now;
	xtBool	message = FALSE;

	if (db->db_sw_thread) {
		then = time(NULL);
		while (!xt_xn_is_before(xt_xn_get_curr_id(db), db->db_xn_to_clean_id)) { // was db->db_xn_to_clean_id <= xt_xn_get_curr_id(db)
			xt_lock_mutex(self, &db->db_xn_wait_lock);
			pushr_(xt_unlock_mutex, &db->db_xn_wait_lock);
			if (db->db_sw_idle) {
				if (!xt_broadcast_cond(NULL, &db->db_xn_wait_cond)) {
					xt_log_and_clear_exception_ns();
					break;
				}
			}
			freer_(); // xt_unlock_mutex(&db->db_xn_wait_lock)
			xt_sleep_100th_second(1);
			now = time(NULL);
			if (now >=  now+10) {
				xt_logf(XT_NT_INFO, "Aborting wait for '%d' sweeper\n", db->db_name);
				break;
			}
			if (now >=  now+2) {
				if (!message) {
					message = TRUE;
					xt_logf(XT_NT_INFO, "Waiting for '%d' sweeper...\n", db->db_name);
				}
			}
		}
	}
}

xtPublic void xt_stop_sweeper(XTThreadPtr self, XTDatabaseHPtr db)
{
	XTThreadPtr thr_sw;

	if (db->db_sw_thread) {
		xt_lock_mutex(self, &db->db_xn_wait_lock);
		pushr_(xt_unlock_mutex, &db->db_xn_wait_lock);

		/* This pointer is safe as long as you have the transaction lock. */
		if ((thr_sw = db->db_sw_thread)) {
			pthread_t thread;

			thread = thr_sw->t_pthread;

			/* Make sure the thread quits when woken up. */
			xt_terminate_thread(self, thr_sw);

			xt_xn_wakeup_transactions(db, TRUE);
	
			freer_(); // xt_unlock_mutex(&db->db_xn_wait_lock)

			/*
			 * GOTCHA: This is a wierd thing but the SIGTERM directed
			 * at a particular thread (in this case the sweeper) was
			 * being caught by a different thread and killing the server
			 * sometimes. Disconcerting.
			 * (this may only be a problem on Mac OS X)
			xt_kill_thread(thread);
			 */
			xt_wait_for_thread(thread);
	
			/* PMC - This should not be necessary to set the signal here, but in the
			 * debugger the handler is not called!!?
			thr_sw->t_delayed_signal = SIGTERM;
			xt_kill_thread(thread);
			 */
			db->db_sw_thread = NULL;
		}
		else
			freer_(); // xt_unlock_mutex(&db->db_xn_wait_lock)
	}
}

#ifdef XXXXXXXXXXXX
/* Return TRUE if records written by the given transaction are
 * visible to the thread (i.e. the transaction of the thread).
 */ 
xtPublic xtBool xt_xn_visible(XTOpenTablePtr ot, xtXactID xn_id, xtRecordID rec_id, xtBool *mine)
{
	register XTThreadPtr	self = ot->ot_thread;
	register XTXactDataPtr	xact;

	/* NOTE: If a transaction is not in RAM, then it is considered aborted.
	 * This means that we can only remove a transaction from memory when
	 * all transactions that were running when we started cleanup have
	 * completed.
	 * Only these transactions may have read something that has been
	 * changed by the sweeper in the meantime.
	 * For example a transaction may fill its buffer when doing a
	 * sequential read. A record may be cleaned by the sweeper that
	 * is in this buffer, but the reader already has a old copy of
	 * the data. Then the sweeper removes the transaction
	 * from RAM. The reader will then consider this record
	 * invalid.
	 */ 
	if (xt_xn_is_before(xn_id, self->st_database->db_xn_min_ram_id)) {
		/* This record is not clean, and the transaction is not in
		 * RAM. This means it has be missed, so clean it up.
		 */
		xn_add_cu_record(ot->ot_table->tab_db, ot->ot_table->tab_id, rec_id);
		return FALSE;
	}
	if (xn_id == self->st_xact_data->xd_start_xn_id) {
		if (mine)
			*mine = TRUE;
		return TRUE;
	}
	if (xt_xn_is_before(self->st_xact_data->xd_start_xn_id, xn_id))
		/* This record is written after the this transaction
		 * started (is not visible).
		 */
		return FALSE;
	if (!(xact = xt_xn_get_xact(self->st_database, xn_id))) {
		xn_add_cu_record(ot->ot_table->tab_db, ot->ot_table->tab_id, rec_id);
		return FALSE;
	}
	if (!(xact->xd_flags & XT_XN_XAC_ENDED) || !xt_xn_is_before(xact->xd_end_xn_id, self->st_xact_data->xd_start_xn_id)) // was >=
		/* Either this transaction has not yet ended, or this
		 * record was written by a transaction that ended
		 * after the reading transaction started!
		 * So this record is not visible!
		 */ 
		return FALSE;
	/* Visible if the transaction was committed: */
	return xact->xd_flags & XT_XN_XAC_COMMITTED;
}

/*
 * Return TRUE if the record has been commited.
 */
xtPublic xtBool xt_xn_committed(XTOpenTablePtr ot, xtXactID xn_id, xtRecordID rec_id, xtBool *mine)
{
	register XTThreadPtr	self = ot->ot_thread;
	register XTXactDataPtr	xact;

	if (xt_xn_is_before(xn_id, self->st_database->db_xn_min_ram_id)) {
		/* This record is not clean, and the transaction is not in
		 * RAM. This means it has be missed, so clean it up.
		 */
		xn_add_cu_record(ot->ot_table->tab_db, ot->ot_table->tab_id, rec_id);
		return FALSE;
	}
	if (xn_id == self->st_xact_data->xd_start_xn_id) {
		if (mine)
			*mine = TRUE;
		return TRUE;
	}
	if (!(xact = xt_xn_get_xact(self->st_database, xn_id))) {
		xn_add_cu_record(ot->ot_table->tab_db, ot->ot_table->tab_id, rec_id);
		return FALSE;
	}
	if (!(xact->xd_flags & XT_XN_XAC_ENDED))
		/* Either this transaction has not yet ended, or */ 
		return FALSE;
	/* TRUE if the record was committed: */
	return xact->xd_flags & XT_XN_XAC_COMMITTED;
}

/*
 * Return TRUE of the transaction is committed, or may be
 * committed in the future.
 *
 * if used, 'wait' must be initialized to FALSE!
 *
 * It will be set to TRUE if the transaction has not yet ended.
 * Return FALSE of the transaction was aborted.
 *
 * Return TRUE if the transaction was committed 
 */
xtPublic xtBool xt_xn_maybe_committed(XTOpenTablePtr ot, xtXactID xn_id, xtRecordID rec_id, xtBool *mine, xtBool *wait)
{
	register XTThreadPtr	self = ot->ot_thread;
	register XTXactDataPtr	xact;

	if (xt_xn_is_before(xn_id, self->st_database->db_xn_min_ram_id)) {
		/* Not in RAM, rollback done: */
		xn_add_cu_record(ot->ot_table->tab_db, ot->ot_table->tab_id, rec_id);
		return FALSE;
	}
	if (xn_id == self->st_xact_data->xd_start_xn_id) {
		if (mine)
			*mine = TRUE;
		return TRUE;
	}
	if (!(xact = xt_xn_get_xact(self->st_database, xn_id))) {
		/* Not in RAM, rollback done: */
		xn_add_cu_record(ot->ot_table->tab_db, ot->ot_table->tab_id, rec_id);
		return FALSE;
	}
	if (!(xact->xd_flags & XT_XN_XAC_ENDED)) {
		/* Transaction not ended, may be visible. */
		if (wait)
			*wait = TRUE;
		return TRUE;
	}
	/* Visible if the transaction was committed: */
	return xact->xd_flags & XT_XN_XAC_COMMITTED;
}

#endif

